/* $Id: AddTerm.c,v 1.16 2000/01/11 01:28:45 frog Exp $ */

#include	<stdio.h>
#include	<X11/Xlib.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Dialog.h>
#include	<Haw/TextOut.h>
#include	<Haw/Preedit.h>
#include	<Haw/Command.h>
#include	<Haw/Util.h>

#include	"hanim.h"
#include	"resource.h"
#include	"hcodec.h"
#include	"hanjadic.h"
#include	"buffer.h"
#include	"XuiRootWin.h"

extern void hanim_error(Buffer *, Widget, String);
extern void error_cancel(Widget, XtPointer, XtPointer);
extern int check_action_args(String *, int, char *);
extern Pixel light_color(Colormap, Pixel);

Widget	AddTermText[2];
Widget	AddTerm;

static unsigned int	code;

void addterm_cancel(Widget w, XtPointer client, XtPointer call)
{
    Buffer *b = (Buffer *)client;

    if ( b->state & ERROR )
        error_cancel( w, client, call );

    XtDestroyWidget( b->shell_widget );
    pop_buffer( b );
    b->shell_widget = NULL;
    b->state = HANJA;
}

void addterm_commit(Widget w, XtPointer client, XtPointer call)
{
    Buffer *b = (Buffer *)client;
    hchar_t *hanja;
    int result;

    hanja = HawTextOutGetStringHcs( AddTermText[0] );
    result = hjdic_addterm( b->hangul, hanja, 0 );
    XFree( hanja );

    if ( !result ) {
        addterm_cancel( w, client, call );
        return;
    }

    switch ( result ) {
    case - 1:
        hanim_error( b, w, resources.died_error );
        break;
    case HANJA_WORD_EXIST:
        hanim_error( b, w, resources.add_error1 );
        break;
    case HANJA_BAD_HANGUL_STR:
        hanim_error( b, w, resources.add_error2 );
        break;
    case HANJA_BAD_HANJA_STR:
        hanim_error( b, w, resources.add_error3 );
        break;
    case HANJA_DICT_FULL:
        hanim_error( b, w, resources.add_error5 );
        break;
    case HANJA_DICT_LOCKED:
        hanim_error( b, w, resources.add_error6 );
        break;
    default:
        hanim_error( b, w, resources.add_error7 );
        break;
    }
}

void addterm(Widget w, XEvent *event, String *params, Cardinal *nn)
{
    Buffer	*b;
    int	n;
    Arg	args[20];
    Pixel	fg, bg, lbg;
    Widget	form, hangul, hanja, part, cancel;
    Dimension	width, height;
    hchar_t	ws[512];

    b = set_buffer( w, event->type );
    if ( b == NULL || !(b->state & HANJA) ||
         (b->state & ERROR) || !check_action_args( params, *nn, "add-term" ) )
        return;

    XtVaGetValues( w, XtNforeground, &fg, XtNbackground, &bg,
                   NULL );
    lbg = light_color( 0, bg );

    n = 0;
    XtSetArg( args[n], XtNtitle, "AddTermDialog" ); n++;
    XtSetArg( args[n], XtNiconic, False ); n++;
    XtSetArg( args[n], XtNmappedWhenManaged, False ); n++;
    XtSetArg( args[n], XtNgeometry, "+0+0" ); n++;
    b->shell_widget = XtCreatePopupShell( "AddTerm",
                                          topLevelShellWidgetClass, TopLevel,
                                          args, n );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    form = XtCreateManagedWidget( "AddForm", formWidgetClass, b->shell_widget,
                                  args, n );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainLeft ); n++;
    hanja = XtCreateManagedWidget( "Term", hLabelWidgetClass, form, args, n );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, lbg ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNfromHoriz, hanja ); n++;
    XtSetArg( args[n], HawNdisplayCaret, False ); n++;
    AddTermText[0] = HawCreateTextOut( form, "TermText", args, n );
    XtManageChild( AddTermText[0] );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainLeft ); n++;
    XtSetArg( args[n], XtNfromVert, AddTermText[0] ); n++;
    hangul = XtCreateManagedWidget( "KeyWord", hLabelWidgetClass, form, args, n );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, lbg ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNfromHoriz, hangul ); n++;
    XtSetArg( args[n], XtNfromVert, AddTermText[0] ); n++;
    AddTermText[1] = HawCreateTextOut( form, "TermText", args, n );
    XtManageChild( AddTermText[1] );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainLeft ); n++;
    XtSetArg( args[n], XtNfromVert, AddTermText[1] ); n++;
    part = XtCreateManagedWidget( "Part", hLabelWidgetClass, form, args, n );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, lbg ); n++;
    XtSetArg( args[n], XtNborderColor, fg ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainLeft ); n++;
    AddTerm = XtCreateManagedWidget( "Ok", hCommandWidgetClass, form, args, n );
    XtAddCallback( AddTerm, XtNcallback, (XtCallbackProc)addterm_commit,
                   (XtPointer)b );

    XtSetArg( args[n], XtNfromHoriz, AddTerm ); n++;
    cancel = XtCreateManagedWidget( "Cancel", hCommandWidgetClass,
                                    form, args, n );
    XtAddCallback( cancel, XtNcallback, (XtCallbackProc)addterm_cancel,
                   (XtPointer)b );

    XtRealizeWidget( b->shell_widget );
    HawInstallTranslations( AddTermText[1], b->preedit_widget );
    HawInstallTranslations( form, b->preedit_widget );
    XtAddGrab( b->shell_widget, False, False );

    XtVaGetValues( b->shell_widget, XtNwidth, &width, XtNheight, &height, NULL );
    XtVaSetValues( b->shell_widget, XtNminWidth, width, XtNminHeight, height,
                   XtNmaxHeight, height, NULL );
    if ( b->xim_style & (XIMPreeditPosition | XIMPreeditArea) )
        HawPreeditMoveWidget( b->preedit_widget,
                              (Position)b->hangulC[b->point], b->shell_widget );
    else
        PEMoveWidget( b->shell_widget );

    code = 0;
    XtVaSetValues( AddTerm, XtNsensitive, False, NULL );

    push_buffer( b );
    buffer_initialize( b );
    b->state = ADD;
    b->xim_style = XIMPreeditNothing | XIMStatusNothing;

    HawTextOutSetStringHcs( AddTermText[0], ws, HawHIGHLIGHT_NORMAL );
    HawTextOutSetString( AddTermText[1], NULL, HawHIGHLIGHT_NORMAL );
}
