/* $Id: CodeTable.c,v 1.28 2000/01/11 01:28:45 frog Exp $ */

#include	<stdio.h>
#include	<stdlib.h>
#include	<X11/Xlib.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Form.h>
#include	<Haw/HanList.h>
#include	<Haw/Label.h>
#include	<Haw/Preedit.h>
#include	<Haw/TextOut.h>
#include	<Haw/Util.h>

#include	"hanim.h"
#include	"resource.h"
#include	"buffer.h"
#include	"hcodec.h"
#include	"XuiDialog.h"
#include	"XuiRootWin.h"

#define		N_KS_HIGH	(96 - 7)
#define		N_KS_LOW	96

extern Pixel	light_color( Colormap, Pixel );
extern int	check_action_args( String *, int, char * );

/** @name ڵǥ Է κ */
/*@{*/

static struct EscapeCode {
    unsigned short	start;
    unsigned short	end;
} ks5601escape[] = {
    { 0x2266, 0x227e }, { 0x252b, 0x252f }, { 0x253a, 0x2540 },
    { 0x2559, 0x2560 }, { 0x2579, 0x257e }, { 0x2665, 0x267e },
    { 0x2770, 0x277e }, { 0x2825, 0x2825 }, { 0x2827, 0x2827 },
    { 0x2830, 0x2830 }, { 0x2a74, 0x2a7e }, { 0x2b77, 0x2b7e },
    { 0x2c42, 0x2c50 }, { 0x2c72, 0x2c7e }, { 0x0000, 0x0000 }
};

static void rewrite_codes( Buffer *b )
{
    char	ss[256];
    Widget	w;

    if ( !b->shell_widget ||
         !(w = XtNameToWidget(b->shell_widget, "*CodeLabel")) )
        return;

#if 1
    sprintf( ss, "%s %04X  %s %04X  %s %04X",
             resources.ksc5601, b->ksc5601 & 0x7f7f,
             resources.euckr, b->ksc5601,
             resources.kssm, ksc2kssmc(b->ksc5601) );
#else
    sprintf( ss, "%s %04X %s %04X %s %04X %s %04X %s %04X",
             resources.ksc5601, b->ksc5601 & 0x7f7f,
             resources.euckr, b->ksc5601,
             resources.kssm, ksc2kssmc(b->ksc5601),
             resources.ucs2, ksc2uc(b->ksc5601),
           );
#endif
    XtVaSetValues( w, XtNlabel, ss, NULL );
}

static unsigned short
nitem2ksc( unsigned short num, unsigned char *hi, unsigned char *low )
{
    *hi = num / N_KS_LOW + 0x21;
    *low = (num % N_KS_LOW) + 0x20;
    if ( *hi >= 0x46 ) /* 0x49 - 3 */
        *hi += 4;
    else if ( *hi >= 0x2d )
        *hi += 3;
    return( ((*hi << 8) | *low) | 0x8080 );
}

static unsigned short ksc2nitem( unsigned short code )
{
    unsigned char	hi, low;
    hi = (code >> 8) & 0x7f;
    low = code & 0x7f;
    if ( hi >= 0x49)
        hi -= 4;
    else if ( hi >= 0x2d )
        hi -= 3;
    hi -= 0x21; low -= 0x20;
    return( hi * N_KS_LOW + low );
}

/// ڵǥ Է Ѵ.
void code_cancel( Widget w, XtPointer client, XtPointer call )
{
    Buffer	*b = (Buffer *)client;

    XtDestroyWidget( b->shell_widget );
    b->shell_widget = NULL;
    b->state &= ~CODE;
}

/// ڵǥ Է Ѵ.
void code_commit( Buffer *b, hchar_t code )
{
    unsigned int	p;
    hchar_t	hh[2];
    Position	pos;

    save_buffer( b );
    if ( !insert_buffer( b, ' ', '\0', b->conversion_mode ) )
        return;

    p = b->point;
    hh[0] = code; hh[1] = '\0';
    if ( !replace_hangul( b, hh ) ) {
        return;
    }

    pos = b->hangulC[p];
    if ( b->xim_style & XIMPreeditNothing ) {
        if ( b->state & EDIT ) {
            HawTextOutInsertHcs( PEText, pos, hh ); // zzz
            HawTextOutSetCursorPosition( PEText, pos + 1 ); // zzz
        } else {
            HawTextOutInsertHcs( b->preedit_widget, pos, hh );
            HawTextOutSetCursorPosition( b->preedit_widget, pos + 1 );
        }
    } else if ( b->xim_style & (XIMPreeditPosition | XIMPreeditArea) ) {
        HawPreeditInsertHcs( b->preedit_widget, pos, hh );
        if ( b->state & EDIT )
            HawPreeditSetHighlight( b->preedit_widget, pos, pos + 1,
                                    HawHIGHLIGHT_SELECTED );
    }

    b->conversion_flag[p] = True;
    b->cursor += 1;
    b->point++;
    b->end_point++;
}

static void activate( Widget w, XtPointer client, XtPointer call )
{
    HawListCallback	lc = (HawListCallback)call;
    Buffer	*b = (Buffer *)client;
    unsigned char	hi, low;

    code_commit( b, ksc2hc( nitem2ksc( lc->item, &hi, &low) ) );
}

static void browse_selection( Widget w, XtPointer client, XtPointer call )
{
    HawListCallback	lc = (HawListCallback)call;
    Buffer	*b = (Buffer *)client;
    unsigned char	hi, low;

    b->ksc5601 = nitem2ksc( lc->item, &hi, &low );
    rewrite_codes( b );
}

static Boolean isescape( State state, unsigned short code )
{
    int	ii;
    unsigned char	low;
    struct EscapeCode	*escape;

    low = code & 0xff;
    if ( low > 0x7e || low < 0x21 )
        return( True );
    escape = ks5601escape;
    for ( ii = 0; escape[ii].start != 0 && code >= escape[ii].start;
          ii++ )
        if ( code <= escape[ii].end )
            return( True );
    return( False );
}

#define NCOLUMN	16

static void CreateCodeSelectWidget( Buffer *b )
{
    Widget	form, codeLabel, widget;
    int	n;
    Arg	args[20];
    unsigned char	hi, low;
    unsigned long	hcode;
    int	selected_item;
    Pixel	lbg;
    unsigned char	*item;
    Dimension	width, height;
    Dimension	vbar_width, base_width;
    short	num_rows;
    int	horiz_distance;

    lbg = light_color( None, default_background );

    /* Dialog Shell */
    n = 0;
    XtSetArg( args[n], XtNtitle, "Code Table" ); n++;
    XtSetArg( args[n], XtNiconic, False ); n++;
    XtSetArg( args[n], XtNgeometry, "+0+0" ); n++;
    XtSetArg( args[n], XtNmappedWhenManaged, False ); n++;
    XtSetArg( args[n], XtNallowShellResize, True ); n++;
    b->shell_widget =
        XtCreatePopupShell("CodeTable",
                           topLevelShellWidgetClass, TopLevel,
                           args, n );

    /* Form which contains push buttons */
    form = XtCreateManagedWidget( "Form", formWidgetClass,
                                  b->shell_widget, NULL, 0 );

    /* Code label which shows hexa value of button */
    n = 0;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    codeLabel = XtCreateManagedWidget( "CodeLabel", hLabelWidgetClass,
                                       form, args, n );
    rewrite_codes( b );

    /* Create code selection browse widget */
    /* ϼ 0x2100 0x7dff ̿ ִ. */
    item = (unsigned char *)XtMalloc( N_KS_LOW * N_KS_HIGH * 2 );
    for ( n = 0, hi = 0x21; hi <= 0x7d; hi++ ) {
        if ( 0x2d == hi)
            hi = 0x30;
        else if ( 0x49 == hi )
            hi = 0x4a;
        for ( low = 0x20; low <= 0x7f; low++, n++ ) {
            hcode = (hi << 8) | low;
            if ( isescape( b->state, (unsigned short)hcode ) )
                hcode = (unsigned long)NULL;
            else
                hcode = hcode | 0x8080;
            item[n * 2 + 0] = (unsigned char)(hcode >> 8);
            item[n * 2 + 1] = (unsigned char)(hcode & 0xff);
        }
    }

    n = 0;
    XtSetArg( args[n], XtNfromVert, codeLabel ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNbottom, XawChainBottom ); n++;
    XtSetArg( args[n], HawNitemBackground, lbg ); n++;
    XtSetArg( args[n], HawNitems, item ); n++;
    XtSetArg( args[n], HawNhighlightAlways, True ); n++;

    XtSetArg( args[n], HawNitemLength, N_KS_LOW * N_KS_HIGH * 2 ); n++;
    XtSetArg( args[n], HawNitemCount, N_KS_LOW * N_KS_HIGH ); n++;

    selected_item = ksc2nitem( b->ksc5601 );
    XtSetArg( args[n], HawNselectedItem, selected_item ); n++;
    XtSetArg( args[n], HawNnumRows, NCOLUMN ); n++;
    XtSetArg( args[n], HawNvisibleHorizontalItemCount, NCOLUMN ); n++;
    XtSetArg( args[n], HawNvisibleVerticalItemCount, 6 ); n++;
    XtSetArg( args[n], XtNallowShellResize, True ); n++;
    b->list_widget = HawCreateScrolledList( form, "List", args, n );
    XtAddCallback( b->list_widget, HawNsetCallback, (XtCallbackProc)activate,
                   (XtPointer)b );
    XtAddCallback( b->list_widget, HawNbrowseSelectionCallback,
                   (XtCallbackProc)browse_selection, (XtPointer)b );
    XtManageChild( b->list_widget );
    XtRealizeWidget( b->shell_widget );

    /*   𸣰, Xaw Viewport  ϸ,
       Ʈ ڽ    ʾҴµ width
       ٲپ ϰ   .
       scrollbar ̰ ݿ ʱ  δ. */
    vbar_width = 0;
    widget = XtNameToWidget( XtParent(b->list_widget), "vertical" );
    if ( widget )
        XtVaGetValues( widget, XtNwidth, &vbar_width, NULL );

    XtVaGetValues( b->list_widget,
                   HawNnumRows, &num_rows,
                   HawNitemBaseWidth, &base_width,
                   XtNhorizDistance, &horiz_distance,
                   NULL );
    width = 10 + vbar_width + base_width * num_rows + horiz_distance;

    XtVaGetValues( b->shell_widget, XtNheight, &height, NULL );
    XtVaSetValues( b->shell_widget,
                   XtNwidth, width, XtNheight, height,
                   XtNminWidth, width, XtNminHeight, height,
                   XtNmaxWidth, width, XtNmaxHeight, height,
                   NULL );

    if ( b->xim_style & (XIMPreeditPosition | XIMPreeditArea) )
        HawPreeditMoveWidget( b->preedit_widget, b->hangulC[b->point],
                              b->shell_widget );
    else
        PEMoveWidget( b->shell_widget );

    HawInstallTranslations( b->list_widget, b->preedit_widget );
    HawInstallTranslations( form, b->preedit_widget );

    AddWMDelCallback(b->shell_widget, code_cancel, b);
    XtAddGrab( b->shell_widget, False, False );

    XFree( item );
}

/// ڵǥ Է Ѵ.
void
codetable( Widget w, XEvent *event, String *params, Cardinal *n )
{
    Buffer *b;

    if ( (b = set_buffer( w, event->type )) == NULL )
        return;
    if ( (b->state & ERROR) || b->shell_widget )
        return;
    if ( !check_action_args( params, *n, "codetable" ) )
        return;
    if ( COMB_STATE(b->hautomata) ) {
        complete_hangul( b );
        clear_hangul_automata( &b->hautomata );
    }
    if ( b->state & (PREEDIT | EDIT) ) {
        b->state |= CODE;
        CreateCodeSelectWidget( b );
    }
}

/*@}*/
