/* $Id: DelTerm.c,v 1.16 2000/01/11 01:28:45 frog Exp $ */

#include	<stdio.h>
#include	<X11/Xlib.h>
#include	<X11/Xlibint.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Dialog.h>
#include	<Haw/Command.h>
#include	<Haw/HanList.h>
#include	<Haw/Util.h>
#include	"hanim.h"
#include	"resource.h"
#include	"buffer.h"
#include	"hcodec.h"
#include	"hanjadic.h"

extern void	hanim_error(Buffer *, Widget, String);
extern Pixel	light_color(Colormap, Pixel);
extern int	check_action_args(String *, int, char *);

Widget	delterm_shell	= NULL;
Widget	question_shell	= NULL;

static Widget	DeleteDialog, QuestionDialog;

#define	ITEM_LENGTH	(24+24+12)

typedef struct {
    hchar_t	hangul;
    hchar_t	hanja;
    int	code;
} del_term;

static del_term	*items;
static int	item_count;

void
delterm_cancel( Widget w, XtPointer client_data, XtPointer call_data )
{
    Buffer	*b = (Buffer *)client_data;

    if ( XtIsManaged(QuestionDialog) ) {
        XtUnmapWidget( question_shell );
        XtUnmanageChild( QuestionDialog );
    } else {
        XtUnmapWidget( delterm_shell );
        XtDestroyWidget( delterm_shell );
        XtDestroyWidget( question_shell );
        delterm_shell = NULL;
        b->state &= ~DELETE;
#if 0
        for ( n = 0; n < item_count; n++ ) {
            XFree(items[n].hangul);
            XFree(items[n].hanja);
        }
#endif
        XFree(items);
        items = NULL;
        if ( b->state & PREEDIT )
            hjdic_close();
    }
}


void
delterm_commit( Widget w, XtPointer client_data, XtPointer call_data )
{
    Buffer	*b = (Buffer *)client_data;
    int	result;
    int	item_position;

    if ( XtIsManaged(QuestionDialog) ) {
        XtVaGetValues( b->list_widget, HawNselectedItem, &item_position, NULL );
        if ( item_position == -1 )
            return;
        XtUnmapWidget( question_shell );
        XtUnmanageChild( QuestionDialog );

#if 0
        result = hjdic_delterm( items[item_position].hangul,
                                items[item_position].hanja,
                                items[item_position].code );
#endif

        switch ( result ) {
        case 0:
            delterm_cancel( w, client_data, call_data );
            break;
        case - 1:
            hanim_error( b, DeleteDialog, resources.died_error );
            break;
        case HANJA_WORD_NOT_EXIST:
            hanim_error( b, DeleteDialog, resources.delete_error1 );
            break;
        case HANJA_DICT_LOCKED:
            hanim_error( b, DeleteDialog, resources.delete_error2 );
            break;
        case HANJA_BAD_HANGUL_STR:
            hanim_error( b, DeleteDialog, resources.delete_error3 );
            break;
        case HANJA_BAD_HANJA_STR:
            hanim_error( b, DeleteDialog, resources.delete_error4 );
            break;
        default:
            hanim_error( b, DeleteDialog, resources.delete_error5 );
            break;
        }
    } else {
        XtVaGetValues( b->list_widget, HawNselectedItem, &item_position, NULL );
        if ( item_position != -1 ) {
            XtManageChild( QuestionDialog );
            XtMapWidget( question_shell );
        }
    }
}


void delterm( Widget w, XEvent *event, String *params, Cardinal *nn )
{
#if 0
    int	flag;
    del_term	item;
    int	item_count;
#endif
    Buffer	*b;
    String	string = NULL;
    Arg	args[20];
    int	n;
    Widget	form, icon, label, ok, cancel;
    Position	x, y;
    Dimension	width, height, hl, base_height;
    Pixel	fg, bg, lbg;
    Screen	*screen;

    if ( (b = set_buffer( w, event->type )) == NULL )
        return;
    if ( (b->state & ~PREEDIT) || delterm_shell ||
         !check_action_args( params, *nn, "delete-term" ) )
        return;

    if ( b->state & PREEDIT ) {
        switch ( hjdic_open( resources.hjdic ) ) {
        case HANJA_CANNOT_OPEN_MDICT:
            hanim_error( b, w, resources.open_error2 );
            return;
        case HANJA_CANNOT_OPEN_UDICT:
            hanim_error( b, w, resources.open_error3 );
            return;
        case HANJA_CANNOT_OPEN_STUDY:
            hanim_error( b, w, resources.open_error4 );
            return;
        case HANJA_CANNOT_MAKE_UDIR:
            hanim_error( b, w, resources.open_error5 );
            return;
        case HANJA_CANNOT_MAKE_UDICT:
            hanim_error( b, w, resources.open_error6 );
            return;
        case HANJA_CANNOT_MAKE_STUDY:
            hanim_error( b, w, resources.open_error7 );
            return;
        }
    }

    item_count = 0;
    items = NULL;
#if 0
    flag = hjdic_getdict( &item );
    while ( flag == 0 ) {
        wcstombs( s1, item.hangul, MB_CUR_MAX * wslen(item.hangul) );
        wcstombs( s2, item.hanja, MB_CUR_MAX * wslen(item.hanja) );
        sprintf( buf, "%-24s%-24s", s1, s2 );

        if ( items == NULL )
            items = (struct del_term *)XtMalloc( sizeof(struct del_term) );
        else
            items = (struct del_term *)XtRealloc( items, sizeof(struct del_term)
                                                  * (item_count + 1) );
        items[item_count].hangul = item.hangul;
        items[item_count].hanja = item.hanja;
        items[item_count].code = item.code;

        if ( string == NULL )
            string = XtMalloc( ITEM_LENGTH + 1 );
        else
            string = XtRealloc( string, ITEM_LENGTH * (item_count + 1) + 1 );

        memset( &string[ITEM_LENGTH*item_count], 0, ITEM_LENGTH );
        strcpy( &string[ITEM_LENGTH*item_count], buf );
        strcat( &string[ITEM_LENGTH*item_count], item.code );
        item_count++;

        flag = hjdic_nextdict_wcs( &item );
    }
    if ( item_count == 0 ) {
        if ( b->state & PREEDIT )
            hjdic_close();
        hanim_error( b, w, resources.delete_error6 );
        return;
    }
#endif

    XtVaGetValues( w, XtNforeground, &fg, XtNbackground, &bg,
                   XtNscreen, &screen, NULL );
    lbg = light_color( 0, bg );

    n = 0;
    XtSetArg( args[n], XtNtitle, "Delete Term Dialog" ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNiconic, False ); n++;
    XtSetArg( args[n], XtNgeometry, "+0+0" ); n++;
    XtSetArg( args[n], XtNmappedWhenManaged, False ); n++;
    XtSetArg( args[n], XtNallowShellResize, True ); n++;
    delterm_shell = XtCreatePopupShell( "DeleteTerm",
                                        topLevelShellWidgetClass, TopLevel,
                                        args, n );

    n = 0;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    form = XtCreateManagedWidget( "DeleteForm", formWidgetClass, delterm_shell,
                                  args, n );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainRight ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    label = XtCreateManagedWidget( "Label", hLabelWidgetClass, form, args, n );

    n = 0;
    XtSetArg( args[n], XtNborderColor, fg ); n++;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNfromVert, label ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNbottom, XawChainBottom ); n++;
    XtSetArg( args[n], XtNjustify, XtJustifyLeft ); n++;
    XtSetArg( args[n], HawNitemBackground, lbg ); n++;
    XtSetArg( args[n], HawNitems, string ); n++;
    XtSetArg( args[n], HawNhighlightAlways, True ); n++;
    XtSetArg( args[n], HawNitemLength, ITEM_LENGTH*item_count ); n++;
    XtSetArg( args[n], HawNitemCount, item_count ); n++;
    XtSetArg( args[n], HawNselectedItem, 0 ); n++;
    XtSetArg( args[n], HawNnumRows, 1 ); n++;
    XtSetArg( args[n], HawNvisibleHorizontalItemCount, 1 ); n++;
    XtSetArg( args[n], XtNresizable, True ); n++;
    b->list_widget = HawCreateScrolledList( form, "List", args, n );
    XtAddCallback( b->list_widget, HawNsetCallback,
                   (XtCallbackProc)delterm_commit, (XtPointer)b );
    XtManageChild( b->list_widget );

    XtRealizeWidget( delterm_shell );
    HawInstallTranslations( b->list_widget, b->preedit_widget );
    HawInstallTranslations( form, b->preedit_widget );
    XtAddGrab( delterm_shell, False, False );

    XtVaGetValues( delterm_shell, XtNwidth, &width, XtNheight, &height, NULL );
    XtVaGetValues( b->list_widget, HawNitemBaseHeight, &base_height,
                   HawNvisibleVerticalItemCount, &item_count,
                   XtNheight, &hl, NULL );
    XtVaSetValues( delterm_shell, XtNminWidth, width, XtNminHeight, height,
                   XtNmaxWidth, width,
                   XtNmaxHeight, height + hl - base_height * item_count, NULL );
    x = ( WidthOfScreen(screen) - width ) / 2;
    y = ( HeightOfScreen(screen) - height ) / 2;
    XtVaSetValues(delterm_shell, XtNx, XtNy, NULL);
    XtMapWidget( delterm_shell );

    b->state |= DELETE;

    n = 0;
    XtSetArg( args[n], XtNtitle, "Question" ); n++;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNiconic, False ); n++;
    XtSetArg( args[n], XtNgeometry, "+0+0" ); n++;
    XtSetArg( args[n], XtNmappedWhenManaged, False ); n++;
    question_shell = XtCreatePopupShell( "Question",
                                         topLevelShellWidgetClass, TopLevel,
                                         args, n );

    n = 0;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    QuestionDialog = XtCreateManagedWidget( "QuestionDialog", dialogWidgetClass,
                                            question_shell, args, n );

    label = XtNameToWidget( QuestionDialog, "label" );
    XtVaSetValues( label, XtNforeground, fg, XtNbackground, bg, NULL );
    icon = XtNameToWidget( QuestionDialog, "icon" );
    if ( icon )
        XtVaSetValues( icon, XtNforeground, fg, XtNbackground, bg, NULL );

    n = 0;
    XtSetArg( args[n], XtNforeground, fg ); n++;
    XtSetArg( args[n], XtNbackground, lbg ); n++;
    XtSetArg( args[n], XtNborderColor, fg ); n++;
    ok = XtCreateManagedWidget( "Ok", hCommandWidgetClass, QuestionDialog,
                                args, n );
    XtAddCallback( ok, XtNcallback, (XtCallbackProc)delterm_commit, (XtPointer)b );

    cancel = XtCreateManagedWidget( "Cancel", hCommandWidgetClass,
                                    QuestionDialog, args, n );
    XtAddCallback( cancel, XtNcallback, (XtCallbackProc)delterm_cancel,
                   (XtPointer)b );

    XtRealizeWidget( question_shell );
    HawInstallTranslations( QuestionDialog, b->preedit_widget );
    XtAddGrab( QuestionDialog, False, False );

    XtVaGetValues( question_shell, XtNwidth, &width, XtNheight, &height, NULL );
    XtVaSetValues( question_shell, XtNminWidth, width, XtNminHeight, height,
                   XtNmaxWidth, width, XtNmaxHeight, height, NULL );
    x = ( WidthOfScreen(screen) - width ) / 2;
    y = ( HeightOfScreen(screen) - height ) / 2;
    XtVaSetValues(question_shell, XtNx, x, XtNy, y, NULL);

    XtUnmanageChild( QuestionDialog );
}

