/* $Id: Label.c,v 1.4 1998/03/11 01:41:03 frog Exp $ */

/**
 * Label.c - multi-byte label widget
 **/

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xos.h>
#include <X11/Xaw/XawInit.h>
#include <Haw/LabelP.h>
#include <X11/Xmu/Converters.h>
#include <X11/Xmu/Drawing.h>
#include <stdio.h>
#include <ctype.h>

#define streq(a,b) (strcmp( (a), (b) ) == 0)

#define MULTI_LINE_LABEL 32767

#ifdef CRAY
#define WORD64
#endif

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

/* Private Data */

#define offset(field) XtOffsetOf(HLabelRec, field)
static XtResource resources[] = {
  {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
   offset(label.foreground), XtRString, XtDefaultForeground},
  {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
   offset(label.font),XtRString, XtDefaultFont},
  {XtNfontSet, XtCFontSet, XtRFontSet, sizeof(XFontSet),
   offset(label.font_set),XtRString, XtDefaultFontSet},
  {XtNlabel,  XtCLabel, XtRString, sizeof(String),
   offset(label.label), XtRString, NULL},
  {XtNjustify, XtCJustify, XtRJustify, sizeof(XtJustify),
   offset(label.justify), XtRImmediate, (XtPointer)XtJustifyCenter},
  {XtNinternalWidth, XtCWidth, XtRDimension,  sizeof(Dimension),
   offset(label.internal_width), XtRImmediate, (XtPointer)4},
  {XtNinternalHeight, XtCHeight, XtRDimension, sizeof(Dimension),
   offset(label.internal_height), XtRImmediate, (XtPointer)2},
  {XtNleftBitmap, XtCLeftBitmap, XtRBitmap, sizeof(Pixmap),
   offset(label.left_bitmap), XtRImmediate, (XtPointer) None},
  {XtNbitmap, XtCPixmap, XtRBitmap, sizeof(Pixmap),
   offset(label.pixmap), XtRImmediate, (XtPointer)None},
  {XtNresize, XtCResize, XtRBoolean, sizeof(Boolean),
   offset(label.resize), XtRImmediate, (XtPointer)True},
};
#undef offset

static void Initialize();
static void Resize();
static void Redisplay();
static Boolean SetValues();
static void ClassInitialize();
static void Destroy();
static XtGeometryResult QueryGeometry();

HLabelClassRec hLabelClassRec = {
  {
    /* core_class fields */	
#define superclass		(&simpleClassRec)
    (WidgetClass) superclass,	/* superclass	  	*/
    "HawLabel",			/* class_name	  	*/
    sizeof(HLabelRec),		/* widget_size	  	*/
    ClassInitialize,		/* class_initialize   	*/
    NULL,			/* class_part_initialize*/
    FALSE,			/* class_inited       	*/
    Initialize,			/* initialize	  	*/
    NULL,			/* initialize_hook	*/
    XtInheritRealize,		/* realize		*/
    NULL,			/* actions		*/
    0,				/* num_actions	  	*/
    resources,			/* resources	  	*/
    XtNumber(resources),	/* num_resources	*/
    NULLQUARK,			/* xrm_class	  	*/
    TRUE,			/* compress_motion	*/
    TRUE,			/* compress_exposure  	*/
    TRUE,			/* compress_enterleave	*/
    FALSE,			/* visible_interest	*/
    Destroy,			/* destroy		*/
    Resize,			/* resize		*/
    Redisplay,			/* expose		*/
    SetValues,			/* set_values	  	*/
    NULL,			/* set_values_hook	*/
    XtInheritSetValuesAlmost,	/* set_values_almost	*/
    NULL,			/* get_values_hook	*/
    NULL,			/* accept_focus	 	*/
    XtVersion,			/* version		*/
    NULL,			/* callback_private   	*/
    NULL,			/* tm_table		*/
    QueryGeometry,		/* query_geometry	*/
    XtInheritDisplayAccelerator,/* display_accelerator	*/
    NULL			/* extension		*/
  },
  /* Simple class fields initialization */
  {
    XtInheritChangeSensitive	/* change_sensitive	*/
  }
};
WidgetClass hLabelWidgetClass = (WidgetClass)&hLabelClassRec;
/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void ClassInitialize()
{
  XawInitializeWidgetSet();
  XtAddConverter( XtRString, XtRJustify, XmuCvtStringToJustify, NULL, 0 );
}

/*
 * Calculate width and height of displayed text in pixels
 */
static void SetMbTextWidthAndHeight(HLabelWidget lw)
{
  register XFontSet	fs = lw->label.font_set;
  XFontSetExtents	*fse;
  char *nl;

  if (lw->label.pixmap != None) {
    Window root;
    int x, y;
    unsigned int width, height, bw, depth;
    if (XGetGeometry(XtDisplay(lw), lw->label.pixmap, &root, &x, &y,
		     &width, &height, &bw, &depth)) {
      lw->label.label_height = height;
      lw->label.label_width = width;
      lw->label.label_len = depth;
      return;
    }
  }

  fse = XExtentsOfFontSet( fs );
  lw->label.label_height = fse->max_logical_extent.height;
  lw->label.line_height = lw->label.label_height;
  lw->label.ascent = -fse->max_logical_extent.y;
  if (lw->label.label == NULL) {
    lw->label.label_len = 0;
    lw->label.label_width = 0;
  }
  else if ((nl = strchr(lw->label.label, '\n')) != NULL) {
    char *label;
    lw->label.label_len = MULTI_LINE_LABEL;
    lw->label.label_width = 0;
    for (label = lw->label.label; nl != NULL; nl = strchr(label, '\n')) {
      int width;

      width = XmbTextEscapement(fs, label, (int)(nl - label));
      if (width > (int)lw->label.label_width)
	lw->label.label_width = width;
      label = nl + 1;
      if (*label)
	lw->label.label_height += fse->max_logical_extent.height;
    }
    if (*label) {
      int width;

      width = XmbTextEscapement(fs, label, strlen(label));
      if (width > (int) lw->label.label_width)
	lw->label.label_width = width;
    }
  } else {
    lw->label.label_len = strlen(lw->label.label);
    lw->label.label_width =
      XmbTextEscapement(fs, lw->label.label, (int) lw->label.label_len);
  }
}

static void GetnormalGC(HLabelWidget lw)
{
  XGCValues	values;

  values.foreground	= lw->label.foreground;
  values.background	= lw->core.background_pixel;
  values.font		= lw->label.font->fid;
  values.graphics_exposures = False;

  lw->label.normal_GC	=
    XtGetGC((Widget)lw, (unsigned) GCForeground | GCBackground |
	    GCFont | GCGraphicsExposures, &values);
}

static void GetgrayGC(HLabelWidget lw)
{
  XGCValues	values;
  unsigned long	gcmask;

  values.foreground = lw->label.foreground;
  values.background = lw->core.background_pixel;
  values.font	    = lw->label.font->fid;
  values.fill_style = FillTiled;
#if 1
  values.tile       = XmuCreateStippledPixmap(XtScreen((Widget)lw),
					      lw->label.foreground, 
					      lw->core.background_pixel,
					      lw->core.depth);
  gcmask	    = GCFont|GCTile|GCFillStyle|GCGraphicsExposures,
#else
  values.tile	    = None;
  gcmask	    = GCFont|GCGraphicsExposures,
#endif
  values.graphics_exposures = False;

  lw->label.stipple = values.tile;
  lw->label.gray_GC = XtGetGC((Widget)lw, gcmask, &values);
}

static void compute_bitmap_offsets (HLabelWidget lw)
{
  /*
   * label will be displayed at (internal_width, internal_height + lbm_y)
   */
  if (lw->label.lbm_height != 0) {
    lw->label.lbm_y = (((int) lw->core.height) -
		       ((int) lw->label.internal_height * 2) -
		       ((int) lw->label.lbm_height)) / 2;
  } else {
    lw->label.lbm_y = 0;
  }
}


static void set_bitmap_info (HLabelWidget lw)
{
  Window root;
  int x, y;
  unsigned int bw, depth;

  if (!(lw->label.left_bitmap &&
	XGetGeometry (XtDisplay(lw), lw->label.left_bitmap, &root, &x, &y,
		      &lw->label.lbm_width, &lw->label.lbm_height,
		      &bw, &depth))) {
    lw->label.lbm_width = lw->label.lbm_height = 0;
  }
  compute_bitmap_offsets (lw);
}

/* ARGSUSED */
static void Initialize(Widget request, Widget new)
{
  HLabelWidget lw = (HLabelWidget) new;

  if (lw->label.label == NULL) 
    lw->label.label = XtNewString(lw->core.name);
  else {
    lw->label.label = XtNewString(lw->label.label);
  }

  GetnormalGC(lw);
  GetgrayGC(lw);

  SetMbTextWidthAndHeight(lw);

  if (lw->core.height == 0)
    lw->core.height = lw->label.label_height + 2*lw->label.internal_height;

  set_bitmap_info (lw);		/* need core.height */

  if (lw->core.width == 0)	/* need label.lbm_width */
    lw->core.width = (lw->label.label_width + 2 * lw->label.internal_width
		      + LEFT_OFFSET(lw));

  lw->label.label_x = lw->label.label_y = 0;
  (*XtClass(new)->core_class.resize) ((Widget)lw);

} /* Initialize */

/* Repaint the widget window */

/* ARGSUSED */
static void Redisplay(Widget w, XEvent *event, Region region)
{
  HLabelWidget lw = (HLabelWidget) w;
  GC gc;

  if (region != NULL) {
    int x = lw->label.label_x;
    unsigned int width = lw->label.label_width;
    if (lw->label.lbm_width) {
      if (lw->label.label_x > (x = lw->label.internal_width))
	width += lw->label.label_x - x;
    }
    if (XRectInRegion(region, x, lw->label.label_y,
		      width, lw->label.label_height) == RectangleOut)
      return;
  }

  gc = XtIsSensitive((Widget)lw) ? lw->label.normal_GC : lw->label.gray_GC;
#ifdef notdef
  if (region != NULL) XSetRegion(XtDisplay(w), gc, region);
#endif /*notdef*/
  if (lw->label.pixmap == None) {
    int len = lw->label.label_len;
    char *label = lw->label.label;
    Position y = lw->label.label_y + lw->label.ascent;

    /* display left bitmap */
    if (lw->label.left_bitmap && lw->label.lbm_width != 0) {
      XCopyPlane (XtDisplay(w), lw->label.left_bitmap, XtWindow(w), gc,
		  0, 0, lw->label.lbm_width, lw->label.lbm_height,
		  (int) lw->label.internal_width, 
		  (int) lw->label.internal_height + lw->label.lbm_y, 
		  (unsigned long) 1L);
    }
    
    if (len == MULTI_LINE_LABEL) {
      char *nl;
      while ((nl = strchr(label, '\n')) != NULL) {
	XmbDrawString(XtDisplay(w), XtWindow(w), lw->label.font_set,
		      gc, lw->label.label_x, y, label, (int)(nl - label));
	y += lw->label.line_height;
	label = nl + 1;
      }
      len = strlen(label);
    }
    if( len )
      XmbDrawString(XtDisplay(w), XtWindow(w), lw->label.font_set,
		    gc, lw->label.label_x, y, label, len);
  } else if (lw->label.label_len == 1) { /* depth */
    XCopyPlane(XtDisplay(w), lw->label.pixmap, XtWindow(w), gc,
	       0, 0, lw->label.label_width, lw->label.label_height,
	       lw->label.label_x, lw->label.label_y, 1L);
  } else {
    XCopyArea(XtDisplay(w), lw->label.pixmap, XtWindow(w), gc,
	      0, 0, lw->label.label_width, lw->label.label_height,
	      lw->label.label_x, lw->label.label_y);
  }
#ifdef notdef
  if (region != NULL)
    XSetClipMask(XtDisplay(w), gc, (Pixmap)None);
#endif /* notdef */
}

static void _Reposition(HLabelWidget lw, Dimension width, Dimension height,
			Position *dx, Position *dy)
{
  Position newPos;
  Position leftedge = lw->label.internal_width + LEFT_OFFSET(lw);

  switch (lw->label.justify) {
  case XtJustifyLeft   :
    newPos = leftedge;
    break;
  case XtJustifyRight  :
    newPos = width -
      (lw->label.label_width + lw->label.internal_width);
    break;
  case XtJustifyCenter :
  default:
    newPos = (int)(width - lw->label.label_width) / 2;
    break;
  }
  if (newPos < (Position)leftedge)
    newPos = leftedge;
  *dx = newPos - lw->label.label_x;
  lw->label.label_x = newPos;
  *dy = (newPos = (int)(height - lw->label.label_height) / 2)
    - lw->label.label_y;
  lw->label.label_y = newPos;
  return;
}

static void Resize(Widget w)
{
  HLabelWidget lw = (HLabelWidget)w;
  Position dx, dy;
  _Reposition(lw, w->core.width, w->core.height, &dx, &dy);
  compute_bitmap_offsets (lw);
}

/*
 * Set specified arguments into widget
 */
#define PIXMAP 0
#define WIDTH 1
#define HEIGHT 2
#define NUM_CHECKS 3

static Boolean SetValues(Widget current, Widget request, Widget new,
			 ArgList args, Cardinal *num_args)
{
  HLabelWidget curlw = (HLabelWidget) current;
  HLabelWidget reqlw = (HLabelWidget) request;
  HLabelWidget newlw = (HLabelWidget) new;
  int i;
  Boolean was_resized = False, redisplay = False, checks[NUM_CHECKS];

  for (i = 0; i < NUM_CHECKS; i++)
    checks[i] = FALSE;

  for (i = 0; i < *num_args; i++) {
    if (streq(XtNbitmap, args[i].name))
      checks[PIXMAP] = TRUE;
    if (streq(XtNwidth, args[i].name))
      checks[WIDTH] = TRUE;
    if (streq(XtNheight, args[i].name))
      checks[HEIGHT] = TRUE;
  }

  if (newlw->label.label == NULL) {
    newlw->label.label = newlw->core.name;
  }

  /*
   * resize on bitmap change
   */
  if (curlw->label.left_bitmap != newlw->label.left_bitmap)
    was_resized = True;

  if (curlw->label.label != newlw->label.label) {
    if (curlw->label.label != curlw->core.name)
      XtFree( (char *)curlw->label.label );

    if (newlw->label.label != newlw->core.name) {
      newlw->label.label = XtNewString( newlw->label.label );
    }
    was_resized = True;
  }

  if (was_resized || (curlw->label.font != newlw->label.font) ||
      (curlw->label.font_set != newlw->label.font_set) ||
      (curlw->label.justify != newlw->label.justify) || checks[PIXMAP]) {

    SetMbTextWidthAndHeight(newlw);
    was_resized = True;
  }

  /* recalculate the window size if something has changed. */
  if (newlw->label.resize && was_resized) {
    if ((curlw->core.height == reqlw->core.height) && !checks[HEIGHT])
      newlw->core.height = (newlw->label.label_height + 
			    2 * newlw->label.internal_height);

    set_bitmap_info (newlw);

    if ((curlw->core.width == reqlw->core.width) && !checks[WIDTH])
      newlw->core.width = (newlw->label.label_width +
			   LEFT_OFFSET(newlw) +
			   2 * newlw->label.internal_width);
  }

  if (curlw->label.foreground != newlw->label.foreground
      || curlw->core.background_pixel != newlw->core.background_pixel
      || curlw->label.font->fid != newlw->label.font->fid) {

    XtReleaseGC(new, curlw->label.normal_GC);
    XtReleaseGC(new, curlw->label.gray_GC);
    if (curlw->label.stipple != None)
      XmuReleaseStippledPixmap( XtScreen(current), curlw->label.stipple );
    curlw->label.stipple = None;
    GetnormalGC(newlw);
    GetgrayGC(newlw);
    redisplay = True;
  }

  if ((curlw->label.internal_width != newlw->label.internal_width)
      || (curlw->label.internal_height != newlw->label.internal_height)
      || was_resized)
  {
    /* Resize() will be called if geometry changes succeed */
    Position dx, dy;
    _Reposition(newlw, curlw->core.width, curlw->core.height, &dx, &dy);
  }

  return was_resized || redisplay ||
    XtIsSensitive(current) != XtIsSensitive(new);
}

static void Destroy(Widget w)
{
  HLabelWidget lw = (HLabelWidget)w;
  XtFree( lw->label.label );
  XtReleaseGC( w, lw->label.normal_GC );
  XtReleaseGC( w, lw->label.gray_GC);
  if (lw->label.stipple != None)
    XmuReleaseStippledPixmap( XtScreen(w), lw->label.stipple );
}

static XtGeometryResult QueryGeometry(Widget w,
				      XtWidgetGeometry *intended,
				      XtWidgetGeometry *preferred)
{
  register HLabelWidget lw = (HLabelWidget)w;

  preferred->request_mode = CWWidth | CWHeight;
  preferred->width = (lw->label.label_width + 2 * lw->label.internal_width +
		      LEFT_OFFSET(lw));
  preferred->height = lw->label.label_height + 2*lw->label.internal_height;
  if (  ((intended->request_mode & (CWWidth | CWHeight))
	 == (CWWidth | CWHeight)) &&
	intended->width == preferred->width &&
	intended->height == preferred->height)
    return XtGeometryYes;
  else if (preferred->width == w->core.width &&
	   preferred->height == w->core.height)
    return XtGeometryNo;
  else
    return XtGeometryAlmost;
}
