/* $Id: Preedit.c,v 1.21 2000/01/11 01:28:46 frog Exp $ */

#include	<stdio.h>
#include	<assert.h>
#include	<X11/Xlib.h>
#include	<X11/Xlibint.h>
#include	<X11/keysym.h>
#include	<X11/IntrinsicP.h>
#include	<X11/StringDefs.h>
#include	<Haw/PreeditP.h>
#include	<X11/Shell.h>

#include	"hcodec.h"
#include	"hstring.h"
#include	"Util.h"
#ifdef DEBUG
#  include	"resource.h"
#endif

#define MAXINT		2147483647
#define ALLOC_SIZE	32

static XtResource local_resources[] =
{
  {
    HawNtopLine, HawCTopLine, XtRInt,
    sizeof(int),
    XtOffset(HawPreeditWidget, preedit.top_line),
    XtRInt, 0 },
  {
    HawNfocusWindow, HawCFocusWindow, XtRWindow,
    sizeof(Window),
    XtOffset(HawPreeditWidget, preedit.focus_window),
    XtRWindow, NULL },
  {
    HawNlineSpacing, HawCLineSpacing, XtRDimension,
    sizeof(Dimension),
    XtOffset(HawPreeditWidget, preedit.line_spacing),
    XtRInt, 0 },
  {
    HawNareaHeight, HawCAreaHeight, XtRDimension,
    sizeof(Dimension),
    XtOffset(HawPreeditWidget, preedit.height),
    XtRInt, 0 },
  {
    HawNareaWidth, HawCAreaWidth, XtRDimension,
    sizeof(Dimension),
    XtOffset(HawPreeditWidget, preedit.width),
    XtRInt, 0 },
  {
    HawNareaX, HawCAreaX, XtRPosition,
    sizeof(Position),
    XtOffset(HawPreeditWidget, preedit.x),
    XtRInt, 0 },
  {
    HawNareaY, HawCAreaY, XtRPosition,
    sizeof(Position),
    XtOffset(HawPreeditWidget, preedit.y),
    XtRInt, 0 },
  {
    HawNareaNeededWidth, HawCAreaNeededWidth, XtRDimension,
    sizeof(Dimension),
    XtOffset(HawPreeditWidget, preedit.area_needed_width),
    XtRInt, 0 },
  {
    HawNareaNeededHeight, HawCAreaNeededHeight, XtRDimension,
    sizeof(Dimension),
    XtOffset(HawPreeditWidget, preedit.area_needed_height),
    XtRInt, 0 },
  {
    HawNspotLocationX, HawCSpotLocationX, XtRPosition,
    sizeof(Position),
    XtOffset(HawPreeditWidget, core.x),
    XtRImmediate, 0 },
  {
    HawNspotLocationY, HawCSpotLocationY, XtRPosition,
    sizeof(Position),
    XtOffset(HawPreeditWidget, core.y),
    XtRImmediate, 0 },
  {
    XtNvalue, XtCValue, XtRString,
    sizeof(String),
    XtOffset(HawPreeditWidget, preedit.value),
    XtRString, NULL },
  {
    HawNhcsValue, HawCHcsValue, HawRHcsValue,
    sizeof(hchar_t *),
    XtOffset(HawPreeditWidget, preedit.hvalue),
    XtRString, NULL },
  {
    HawNcursorPosition, HawCCursorPosition, XtRPosition,
    sizeof(Position),
    XtOffset(HawPreeditWidget, preedit.cursor_position),
    XtRImmediate, (XtPointer)0 },
  {
    HawNcolumns, HawCColumns, XtRShort,
    sizeof(short),
    XtOffset(HawPreeditWidget, preedit.columns),
    XtRImmediate, (XtPointer)1 },
  {
    XtNfontSet, XtCFontSet, XtRFontSet,
    sizeof(XFontSet),
    XtOffset(HawPreeditWidget, preedit.font_set),
    XtRString, XtDefaultFontSet },
  {
    XtNforeground, XtCForeground, XtRPixel,
    sizeof(Pixel),
    XtOffset(HawPreeditWidget, preedit.foreground),
    XtRString, XtDefaultForeground },
  {
    HawNfocus, HawCFocus, XtRBoolean,
    sizeof(Boolean),
    XtOffset(HawPreeditWidget, preedit.has_focus),
    XtRImmediate, (XtPointer)False },
  {
    HawNmaxLength, HawCMaxLength, XtRInt,
    sizeof(int),
    XtOffset(HawPreeditWidget, preedit.max_length),
    XtRImmediate, (XtPointer)MAXINT },
  {
    HawNverifyBell, HawCVerifyBell, XtRBoolean,
    sizeof(Boolean),
    XtOffset(HawPreeditWidget, preedit.verify_bell),
    XtRImmediate, (XtPointer)True },
};

static void ValidateString( HawPreeditWidget w, char *value )
{
  int		len = strlen(value);
  hchar_t	*hs = (hchar_t *)XtMalloc(sizeof(hchar_t)*(len + 1));
  
  len = mbstohcs(hs, value, len );
  if( len != -1 ) {
    XtFree( (char *)TextHcsValue(w) );
    hs[len] = '\0';
    w->preedit.string_length = len;
    w->preedit.alloc_size = sizeof(hchar_t) * ( len + ALLOC_SIZE );
    TextHcsValue(w) =
      hcscpy( (hchar_t *)XtMalloc( w->preedit.alloc_size ), hs );
    XtFree( (char *)hs );
  }
  TextValue(w) = NULL;
}

static void CalculateLine( HawPreeditWidget w, Position position )
{
  int		line;
  Position	*p = w->preedit.line.position;
  Dimension	width;
  hchar_t	*value = w->preedit.hvalue;
  int		ii;
  
  if( w->preedit.string_length == 0 ) {
    w->preedit.line.number = 1;
    w->preedit.line.position[0] = 0;
    w->preedit.line.position[1] = 0;
    return;
  }
  if( w->preedit.max_lines == 1 ) {
    w->preedit.line.number = 1;
    w->preedit.line.position[1] = w->preedit.string_length;
    return;
  }
  
  for( line = w->preedit.line.number - 1; line > 0; line-- )
    if( position >= p[line] )
      break;
  
  if( line == 0 && w->preedit.width < (w->core.x - w->preedit.x) ) {
    p[1] = w->preedit.string_length;
    w->preedit.line.number = 1;
    return;
  }
  do {
    width = (line == 0) ? (w->preedit.width - (w->core.x - w->preedit.x))
      : w->preedit.width;
    /*  Ŀ ׸ ؼ ׸ŭ . */
    width -= w->preedit.cursor_width;
    if( line + 1 >= w->preedit.line.maximum ) {
      w->preedit.line.maximum = line + 2;
      p = w->preedit.line.position
	= (Position *)XtRealloc( (char *)p, (Cardinal)((line + 2) *
						       sizeof(Position)) );
    }
    ii = position - p[line];
    while( HawTextEscapement( TextFontSet(w), value + p[line], ii )
	   <= width ) {
      p[line+1] = p[line] + ii;
      if( (p[line+1]) == w->preedit.string_length )
	break;
      ii++;
    }
    line++;
    position = p[line];
    w->preedit.line.number = line;
  } while( position < w->preedit.string_length );
}

static int FindLine( HawPreeditWidget w, Position position )
{
  Position	*p = w->preedit.line.position;
  int		ii;
  
  if( position >= w->preedit.string_length )
    return( w->preedit.line.number - 1 );
  
  for( ii = 1; ii <= w->preedit.line.number; ii++ )
    if( position < p[ii] )
      break;
  
  return( ii - 1 );
}

static Boolean GetWXYFromPos( HawPreeditWidget w, Position position,
			      Window *window, Position *x, Position *y )
{
  int		line = FindLine( w, position );
  hchar_t	*value = TextHcsValue(w);
  Position	p;

  *window = w->core.window;
  *x = w->preedit.cursor_width;
  *y = w->preedit.ascent;
  
  if( position > w->preedit.string_length )
    return( False );
  if( position == 0 )		// for speed
    return( True );
  
  p = w->preedit.line.position[line];
  if( line == 0 ) {
    *window = w->core.window;
    *x += HawTextEscapement( TextFontSet(w), value, position );
    
    if( w->preedit.max_lines == 1 ) {
      if( w->preedit.attachment == HawCURSOR_LEFT )
	*x -= HawTextEscapement( TextFontSet(w), value,
				 w->preedit.left_cursor_position );
      else
	*x += w->core.width - w->preedit.cursor_width
	  - HawTextEscapement( TextFontSet(w), value,
			       w->preedit.right_cursor_position );
    }
  }
  else if( line == w->preedit.top_line ) {
    *window = w->core.window;
    *x += HawTextEscapement( TextFontSet(w), &value[p], position - p )
      - w->core.x + w->preedit.x;
  }
  return( True );
}

static void PaintCursor( HawPreeditWidget w )
{
  Position	x, y;
  Window	window;

  if( !XtIsRealized((Widget)w) )
    return;

  GetWXYFromPos( w, TextCursorPosition(w), &window, &x, &y );
  x -= w->preedit.cursor_width;
  y -= w->preedit.cursor_height;
  XCopyArea( XtDisplay(w), w->preedit.cursor, window,
      w->preedit.xor_gc, 0, 0,
      w->preedit.cursor_width, w->preedit.cursor_height, x, y );
}

static void PaintDestination( HawPreeditWidget w )
{
  Position	x, y;
  Window	window;

  if( !XtIsRealized( (Widget)w ) )
    return;

  GetWXYFromPos( w, TextCursorPosition(w), &window, &x, &y );
  x -= w->preedit.cursor_width -
    ((w->preedit.cursor_width - CARET_WIDTH) >> 1);
  y -= CARET_HEIGHT - w->preedit.descent;
  XCopyArea( XtDisplay(w), w->preedit.dest_cursor, window,
	     w->preedit.xor_gc, 0, 0,
	     CARET_WIDTH, CARET_HEIGHT, x, y );
  w->preedit.dest_on = !w->preedit.dest_on;
}

static void DrawInsertionPoint( HawPreeditWidget w, Boolean turn_on )
{
  if( w->preedit.cursor_on == turn_on || !XtIsRealized((Widget)w) )
    return;
  w->preedit.cursor_on = turn_on;
  PaintCursor( w );
}

static void LoadGCs(HawPreeditWidget w, Pixel background, Pixel foreground)
{
  Display		*display = XtDisplay(w);
  Window		window = XtWindow(w);
  XGCValues		values;
  unsigned long	mask;

  mask = 0;
  values.function = GXcopy;		mask |= GCFunction;
  values.foreground = foreground;	mask |= GCForeground;
  values.background = background;	mask |= GCBackground;
  values.fill_style = FillSolid;	mask |= GCFillStyle;
  values.line_width = (w->preedit.ascent > 19) ? 2 : 1;
  mask |= GCLineWidth;
  values.graphics_exposures = False;	mask |= GCGraphicsExposures;
  if( w->preedit.gc != NULL )
    XChangeGC( display, w->preedit.gc, mask, &values );
  else
    w->preedit.gc = XCreateGC( display, window, mask, &values );

  values.foreground = background;	mask |= GCForeground;
  values.background = foreground;	mask |= GCBackground;
  if( w->preedit.inv_gc != NULL )
    XChangeGC( display, w->preedit.inv_gc, mask, &values );
  else
    w->preedit.inv_gc = XCreateGC( display, window, mask, &values );

  values.function = GXxor;
  values.graphics_exposures = FALSE;
  values.foreground = foreground ^ background;
  values.background = 0;
  if( w->preedit.xor_gc != NULL )
    XtReleaseGC( (Widget)w, w->preedit.xor_gc );
  w->preedit.xor_gc = XtGetGC( (Widget)w, mask, &values );
}

static void
MakeIBeamCursor( HawPreeditWidget w, Pixmap *cursor, int line_width )
{
  Screen	*screen = XtScreen(w);
  Display	*display = XtDisplay(w);
  Window	window = XtWindow(w);
  XSegment	segments[3];
  XtGCMask	mask;
  XGCValues	values;
  GC		gc;

  mask = GCForeground | GCBackground |GCLineWidth;
  values.foreground = 0;
  values.background = 0;
  values.line_width = line_width;
  gc = XCreateGC( display, window, mask, &values );

  *cursor = XCreatePixmap( display, window,
			   w->preedit.cursor_width,
			   w->preedit.cursor_height,
			   screen->root_depth );

  XFillRectangle( display, *cursor, gc, 0, 0,
		  w->preedit.cursor_width,
		  w->preedit.cursor_height );
  XSetForeground( display, gc,
		  w->preedit.foreground ^ w->core.background_pixel );

  segments[0].x1 = 0;
  segments[0].y1 = line_width - 1;
  segments[0].x2 = w->preedit.cursor_width;
  segments[0].y2 = line_width - 1;

  segments[1].x1 = 0;
  segments[1].y1 = w->preedit.cursor_height - 1;
  segments[1].x2 = w->preedit.cursor_width;
  segments[1].y2 = w->preedit.cursor_height - 1;

  segments[2].x1 = w->preedit.cursor_width / 2;
  segments[2].y1 = line_width;
  segments[2].x2 = w->preedit.cursor_width / 2;
  segments[2].y2 = w->preedit.cursor_height - 1;

  XDrawSegments( display, *cursor, gc, segments, 3 );

  XFreeGC( display, gc );
}

static void MakeDestCursor( HawPreeditWidget w )
{
  Screen *screen = XtScreen(w);

  w->preedit.dest_cursor =
    XCreatePixmapFromBitmapData( XtDisplay(w),
	XtWindow(w),
	(char*)caretBits, CARET_WIDTH, CARET_HEIGHT,
	w->core.background_pixel ^ w->preedit.foreground, 0,
	screen->root_depth );
}

static void MakeCursors( HawPreeditWidget w )
{
  int line_width = 1;

  if( !XtIsRealized((Widget)w) )
    return;

  w->preedit.cursor_width = 3;
  w->preedit.cursor_height = w->preedit.ascent;

  MakeDestCursor( w );

  if( w->preedit.cursor_height > 19 ) {
    w->preedit.cursor_width++;
    line_width = 2;
  }

  MakeIBeamCursor( w, &(w->preedit.cursor), line_width );
}


static void ClassInitialize( void )
{
}

static void Destroy( Widget w )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  Display		*display = XtDisplay(w);

  if( pw->preedit.timer_id )
    XtRemoveTimeOut( pw->preedit.timer_id );
  XFreeGC( display, pw->preedit.gc );
  XFreeGC( display, pw->preedit.inv_gc );
  XtReleaseGC( w, pw->preedit.xor_gc );

  XtFree( (char *)pw->preedit.highlight.list );
  XtFree( (char *)pw->preedit.line.position );
  XtFree( pw->preedit.value );
}

static Cardinal FindHighlight( HawPreeditWidget w, Position position )
{
  HighlightRec	*l = w->preedit.highlight.list;
  Cardinal ii;

  for( ii = w->preedit.highlight.number - 1; ii >= 0; ii-- )
    if( position >= l[ii].position )
      return( ii );
  return( 0 );
}

static void InsertHighlight( HawPreeditWidget w, Position position,
			     HawHighlightMode mode )
{
  HighlightRec	*l = w->preedit.highlight.list;
  int		ii, jj;

  ii = FindHighlight( w, position );
  
  if( l[ii].position == position ) {
    l[ii].mode = mode;
    if( ii > 0 && l[ii-1].mode == mode ) {
      for( jj = ii, w->preedit.highlight.number--;
	   jj < w->preedit.highlight.number; jj++ )
	l[jj] = l[jj+1];
      ii--;
    }
    if( ii + 1 != w->preedit.highlight.number && l[ii+1].mode == mode ) {
      w->preedit.highlight.number--;
      for( ii++; ii < w->preedit.highlight.number; ii++ )
	l[ii] = l[ii+1];
    }
  }
  else  if( l[ii].mode != mode ) {
    if( ii + 1 != w->preedit.highlight.number && l[ii+1].mode == mode ) {
      l[ii+1].position = position;
    }
    else {
      if( w->preedit.highlight.number == w->preedit.highlight.maximum ) {
	w->preedit.highlight.maximum++;
	l = w->preedit.highlight.list = (HighlightRec *)XtRealloc( (char *)l, sizeof(HighlightRec) * w->preedit.highlight.maximum );
      }
      for( jj = w->preedit.highlight.number; jj > ii; jj-- )
	l[jj] = l[jj-1];
      w->preedit.highlight.number++;
      l[ii+1].position = position;
      l[ii+1].mode = mode;
    }
  }
  
  for( ii = 0; ii < w->preedit.highlight.number; ii++ ) {
    if( w->preedit.highlight.list[ii].position > w->preedit.string_length ) {
      w->preedit.highlight.number = ii;
      break;
    }
  }
}

static void PreeditAreaSetHighlight( HawPreeditWidget w,
				     Position left, Position right,
				     HawHighlightMode mode )
{
  HawHighlightMode	rmode;
  int			l, r;

  if( left >= right )
    return;

  rmode = w->preedit.highlight.list[FindHighlight(w, right)].mode;
  InsertHighlight( w, left, mode );
  l = FindHighlight( w, left );
  r = FindHighlight( w, right - 1 );
  while( l < r ) {
    InsertHighlight( w, w->preedit.highlight.list[r].position, mode );
    r = FindHighlight( w, right );
    if( w->preedit.highlight.list[r].mode == mode )
      r--;
  }
  InsertHighlight( w, right, rmode );
}

static void RedisplayText( HawPreeditWidget w, Position start, Position end )
{
  Display	*display = XtDisplay(w);
  hchar_t	*value = w->preedit.hvalue;
  HighlightRec	*list = w->preedit.highlight.list;
  Position	*position = w->preedit.line.position;
  int		x, y, width, height;
  int		ii;
  int		sline = FindLine( w, start );
  int		eline = FindLine( w, end );
  Position	p, _end;
  Window	win;
  GC		gc;
  int		top_line = w->preedit.top_line;

  if( !XtIsRealized((Widget)w) )
    return;
  if( sline < w->preedit.top_line ) {
    sline = w->preedit.top_line;
    start = w->preedit.line.position[sline];
  }
  if( eline >= w->preedit.top_line + w->preedit.max_lines ) {
    eline = w->preedit.top_line + w->preedit.max_lines;
    end = w->preedit.line.position[eline];
  }
  if( sline > eline )
    return;

  DrawInsertionPoint( w, False );

  XSelectInput( display, w->core.window,
		w->preedit.event_mask & ~ExposureMask );
  gc = w->preedit.inv_gc;
  if( sline == top_line ) {
    if( w->preedit.line.number == 1 )
      width = HawTextEscapement( TextFontSet(w), value, 
				 w->preedit.string_length );
    else
      width = w->preedit.width - ( w->core.x - w->preedit.x )
	- w->preedit.cursor_width;
    w->core.width = width + w->preedit.cursor_width;
    if( w->core.width > w->preedit.width - ( w->core.x - w->preedit.x ) )
      w->core.width = w->preedit.width - ( w->core.x - w->preedit.x );
    if( w->core.width > w->preedit.width )
      w->core.width = w->preedit.width;

#ifdef DEBUG
    if( resources.debug & 0x80 ) {
      printf("===== ReplaceText: top_line\n");
      printf("x = %d, y = %d, w = %d, h = %d\n",
	     w->core.x, w->core.y, w->core.width, w->core.height );
    }
#endif
#if 1 // xxxxxx
    {
      int fx, fy;
      Window child;

      XTranslateCoordinates(XtDisplay(w), w->preedit.focus_window,
	      DefaultRootWindow(XtDisplay(w)), 0, 0, &fx, &fy, &child);
      XMoveResizeWindow( display, XtWindow(w),
		fx + w->core.x, fy + w->core.y, w->core.width, w->core.height );
      if (w->preedit.string_length && w->preedit.has_focus)
        XMapRaised(display, XtWindow(w));
      else
        XUnmapWindow(display, XtWindow(w));
    }
#else
    XMoveResizeWindow( display, XtWindow(w), w->core.x, w->core.y,
		       w->core.width, w->core.height );
#endif
#if 0
    XFillRectangle( display, w->core.window, gc,
		    width + (w->preedit.cursor_width / 2), 0,
		    w->preedit.cursor_width, w->core.height );
#else
    XFillRectangle( display, w->core.window, gc,
		    width - (w->preedit.cursor_width / 2), 0,
		    w->preedit.cursor_width, w->core.height );
#endif
  }

  ii = 0;
  while( start != end && ii < w->preedit.highlight.number ) {
    if( list[ii].position <= start &&
	(w->preedit.highlight.number == ii + 1 ||
	 start < list[ii+1].position) ) {
      sline = FindLine( w, start );
#if 0
      x = w->preedit.cursor_width >> 1;
#else
      /* ڰ  ġġ    */
      x = 0;
#endif
      y = w->preedit.ascent;
      if( sline == 0 ) {
	win = w->core.window;
	if( w->preedit.max_lines == 1 ) {
	  start = list[ii].position;
	  if( w->preedit.attachment == HawCURSOR_LEFT )
	    x -= HawTextEscapement( TextFontSet(w), value,
				    w->preedit.left_cursor_position );
	  else
	    x = w->core.width - (w->preedit.cursor_width >> 1)
	      - HawTextEscapement( TextFontSet(w), value,
				   w->preedit.right_cursor_position );
	  x += HawTextEscapement( TextFontSet(w), value, start );
	}
	else
	  x += HawTextEscapement( TextFontSet(w), value, start );
      }
      else if( sline == w->preedit.top_line ) {
	win = w->core.window;
	x += HawTextEscapement( TextFontSet(w), &value[position[sline]],
				start - position[sline] )
	  - w->core.x + w->preedit.x;
      }

      _end = end;
      if( ii+1 < w->preedit.highlight.number &&
	  _end > list[ii+1].position )
	_end = list[ii+1].position;
      if( sline+1 < w->preedit.line.number && _end > position[sline+1] )
	_end = position[sline+1];

      switch( list[ii].mode ) {
      case HawHIGHLIGHT_NORMAL:
      case HawHIGHLIGHT_SECONDARY_SELECTED:
	gc = w->preedit.gc;
	break;

      case HawHIGHLIGHT_SELECTED:
	gc = w->preedit.inv_gc;
	break;

      default:
	gc = w->preedit.gc;
	break;
      }

      HawDrawImageString( display, win, w->preedit.font_set,
			  gc, x, y, value + start, _end - start );

      if( list[ii].mode == HawHIGHLIGHT_SECONDARY_SELECTED )
	XDrawLine( display, win, gc,
		   x, y + w->preedit.descent - 1,
		   x + HawTextEscapement( TextFontSet(w), value + start,
					  _end - start ),
		   y + w->preedit.descent - 1 );
      start = _end;
    }
    else
      ii++;
  }

  if( w->preedit.has_focus )
    DrawInsertionPoint( w, True );
  else
    PaintDestination( w );
  XSelectInput( display, w->core.window, w->preedit.event_mask );
}

static void SetCursorPosition( HawPreeditWidget w, Position position )
{
  Boolean	cursor_on = w->preedit.cursor_on;
  Position	*p = w->preedit.line.position;
  int		old_line, new_line;
  int		n;
  Boolean	redraw = False;
  hchar_t	*value = w->preedit.hvalue;
  Dimension	width;

  if( position < 0 )
    position = 0;

  assert( w->core.x <= w->preedit.x + w->preedit.width );

  if( position > w->preedit.string_length )
    position = w->preedit.string_length;

  if( w->preedit.max_lines == 1 ) {
    if( HawTextEscapement(TextFontSet(w), value, w->preedit.string_length) <
	w->preedit.width - ( w->core.x - w->preedit.x ) ) {
      w->preedit.left_cursor_position = 0;
      w->preedit.right_cursor_position = w->preedit.string_length;
      w->preedit.attachment = HawCURSOR_LEFT;

      XFillRectangle( XtDisplay(w), XtWindow(w), w->preedit.inv_gc,
		      0, 0, w->preedit.cursor_width / 2, w->core.height );
    }
    else if( position < w->preedit.left_cursor_position ) {
      width = w->preedit.width - ( w->core.x - w->preedit.x );
      for( n = position + 1;
	   width > HawTextEscapement( TextFontSet(w), &value[position],
				      n - position ); n++ );
      w->preedit.right_cursor_position = n - 1;
      w->preedit.left_cursor_position = position;
      w->preedit.attachment = HawCURSOR_LEFT;
      redraw = True;
    }
    else if( position > w->preedit.right_cursor_position ) {
      int ii = hcslen(value);

      width = (w->preedit.cursor_width >> 1)
	+ HawTextEscapement( TextFontSet(w), value, position )
	- (w->preedit.width - ( w->core.x - w->preedit.x ));
      for( n = 1; n <= ii && HawTextEscapement( TextFontSet(w), value, n ) < width;
	   n++ );
#if 0
      assert( HawTextEscapement( TextFontSet(w), value, n ) >= width );
      printf("width = %d, hlen = %d\n", width, ii);
      printf("textescape = %d\n", HawTextEscapement( TextFontSet(w), value, 1 )); 
      for( n = 1; n <= ii && HawTextEscapement( TextFontSet(w), value, n ) < width;
	   n++ )
	printf("textescape = %d\n", HawTextEscapement( TextFontSet(w), value, n )); 
      printf("position = %d, left = %d, right = %d\n", 
	     position, w->preedit.left_cursor_position,
	     w->preedit.right_cursor_position );
      printf("core.x = %d, preedit.x = %d, preedit.width = %d\n", 
	     w->core.x, w->preedit.x, w->preedit.width );
      printf("Cursro Position Error\n");
#endif
      w->preedit.left_cursor_position = n;
      w->preedit.right_cursor_position = position;
      w->preedit.attachment = HawCURSOR_RIGHT;
      redraw = True;
    }
    else if( w->preedit.attachment == HawCURSOR_RIGHT &&
	     w->preedit.string_length < w->preedit.right_cursor_position ) {
      width = (w->preedit.cursor_width >> 1)
	+ HawTextEscapement( TextFontSet(w), value,
			     w->preedit.string_length )
	- (w->preedit.width - ( w->core.x - w->preedit.x ));
      for( n = 1; HawTextEscapement( TextFontSet(w), value, n ) < width;
	   n++ );
      w->preedit.left_cursor_position = n;
      w->preedit.right_cursor_position = w->preedit.string_length;
      redraw = True;
    }
  }

  old_line = FindLine( w, TextCursorPosition(w) );
  new_line = FindLine( w, position );
  if( TextCursorPosition(w) < position ) {
    for( n = 1; n <= w->preedit.line.number; n++ ) {
      if( p[n] == TextCursorPosition(w) ) {
	old_line--;
	break;
      }
    }
  }

  if( w->preedit.cursor_on )
    DrawInsertionPoint( w, False );

  TextCursorPosition(w) = position;

  if( old_line > new_line ) {
    if( w->preedit.top_line > new_line ) {
      w->preedit.top_line = new_line;
      redraw = True;
    }
  }
  else if( old_line < new_line ) {
    if( w->preedit.top_line + w->preedit.max_lines <= new_line ) {
      w->preedit.top_line = new_line + 1 - w->preedit.max_lines;
      redraw = True;
    }
  }
  else {
    if( w->preedit.top_line + w->preedit.max_lines >
	w->preedit.line.number &&
	w->preedit.max_lines <= w->preedit.line.number ) {
      w->preedit.top_line = w->preedit.line.number - w->preedit.max_lines;
      redraw = True;
    }
  }
  if( w->preedit.top_line > 0 && new_line == w->preedit.top_line ) {
    w->preedit.top_line--;
    redraw = True;
  }

  if( redraw ) {
    XClearArea( XtDisplay(w), XtWindow(w), 0, 0, 0, 0, True );
  }
  if( cursor_on )
    DrawInsertionPoint( w, True );
}

static Boolean SetValues( Widget current, Widget request, Widget neww,
			  ArgList args, Cardinal *num_args )
{
  HawPreeditWidget	cw = (HawPreeditWidget)current;
  HawPreeditWidget	rw = (HawPreeditWidget)request;
  HawPreeditWidget	nw = (HawPreeditWidget)neww;
  Boolean		changed = False;
  Display		*d = XtDisplay( request );
  XFontStruct		**fs;
  char			**fn;
  int			n;
  Dimension		height;

  if( nw->preedit.x + nw->preedit.width <= nw->core.x )
  {
#if 0
    printf("Error: Preedit\n");
#endif
    nw->core.x = nw->preedit.x + nw->preedit.width;
  }

  if( cw->core.width != nw->core.width )
    nw->core.width = cw->core.width;
  if( cw->core.height != nw->core.height )
    nw->core.height = cw->core.height;

  if( nw->preedit.font_set != cw->preedit.font_set ) {
    if( nw->preedit.font_set != NULL )
      XFreeFontSet( d, cw->preedit.font_set );
    n = XFontsOfFontSet( TextFontSet(nw), &fs, &fn );
    nw->preedit.ascent = 0;
    nw->preedit.descent = 0;
    while( --n >= 0 ) {
      if( nw->preedit.ascent < fs[n]->ascent )
	nw->preedit.ascent = fs[n]->ascent;
      if( nw->preedit.descent < fs[n]->descent )
	nw->preedit.descent = fs[n]->descent;
    }
    MakeCursors( nw );
    XClearWindow( d, XtWindow(cw) );
    if( nw->preedit.line_spacing == cw->preedit.line_spacing )
      nw->preedit.line_spacing = nw->preedit.ascent + nw->preedit.descent;
    changed = True;
  }

#if 0 // xxx
  if( nw->preedit.focus_window != cw->preedit.focus_window ) {
    XReparentWindow( XtDisplay(request), nw->core.window,
		     nw->preedit.focus_window, nw->core.x, nw->core.y );
  }
#endif

  if( nw->core.x != cw->core.x || nw->core.y != cw->core.y ) {
    cw->core.x = nw->core.x;
    cw->core.y = nw->core.y;
    nw->preedit.top_line = 0;
#if 0 /* overTheSpot 쿡 ٷ */
    nw->preedit.max_lines =
      (nw->preedit.height -
       (nw->core.y - nw->preedit.descent - nw->preedit.y))
      / nw->preedit.line_spacing;
#else
    nw->preedit.max_lines = 1;
#endif
    if( nw->preedit.max_lines == 0 )
      nw->preedit.max_lines = 1;
    CalculateLine( nw, 0 );
    changed = True;
  }

  if( nw->preedit.x != cw->preedit.x || nw->preedit.y != cw->preedit.y ||
      nw->preedit.width != cw->preedit.width ||
      nw->preedit.height != cw->preedit.height ||
      nw->preedit.line_spacing != cw->preedit.line_spacing ) {
    if( nw->preedit.line_spacing <= 0 )
      nw->preedit.line_spacing = nw->preedit.ascent + nw->preedit.descent;
    cw->core.height = nw->core.height = nw->preedit.line_spacing;
#if 0 /* overTheSpot 쿡 ٷ */
    nw->preedit.max_lines =
      (nw->preedit.height -
       (nw->core.y - nw->preedit.descent - nw->preedit.y))
      / nw->preedit.line_spacing;
#else
    nw->preedit.max_lines = 1;
#endif
    if( nw->preedit.max_lines == 0 )
      nw->preedit.max_lines = 1;
    CalculateLine( nw, 0 );
    nw->preedit.area_needed_width = nw->preedit.width;
    nw->preedit.area_needed_height = nw->preedit.height;
    SetCursorPosition( nw, 0 );
    SetCursorPosition( nw, nw->preedit.string_length );
    SetCursorPosition( nw, TextCursorPosition(cw) );
    changed = True;
  }

  if( nw->preedit.area_needed_width <= 0 )
    nw->preedit.area_needed_width = nw->preedit.width;
  height = nw->preedit.ascent + nw->preedit.descent;
  if( nw->preedit.area_needed_height < height )
    nw->preedit.area_needed_height = height;

  if( nw->preedit.foreground != cw->preedit.foreground ||
      nw->core.background_pixel != cw->core.background_pixel ) {
    LoadGCs( nw, nw->preedit.foreground, nw->core.background_pixel );
    changed = True;
  }

  if( hcscmp( nw->preedit.hvalue, cw->preedit.hvalue ) ) {
    XtFree( (char *)TextHcsValue(cw) );
    if( nw->preedit.hvalue == NULL ) {
      nw->preedit.string_length = 0;
      nw->preedit.alloc_size = sizeof(hchar_t) * ALLOC_SIZE;
      TextHcsValue(nw) = (hchar_t *)XtMalloc( nw->preedit.alloc_size );
      TextHcsValue(nw)[0] = '\0';
    }
    else {
      nw->preedit.string_length = hcslen( TextHcsValue(nw) );
      nw->preedit.alloc_size = sizeof(hchar_t) *
	( nw->preedit.string_length + ALLOC_SIZE );
      TextHcsValue(nw) =
	hcscpy( (hchar_t *)XtMalloc( nw->preedit.alloc_size ),
	       TextHcsValue(rw) );
    }

    CalculateLine( nw, 0 );
    nw->preedit.highlight.number = 1;
    nw->preedit.highlight.list[0].position = 0;

    HawPreeditSetCursorPosition( neww, nw->preedit.string_length );
    changed = True;
    nw->preedit.value = NULL;
  }
  else if( nw->preedit.value != NULL ) {
    ValidateString( nw, nw->preedit.value );
    CalculateLine( nw, 0 );
    nw->preedit.highlight.number = 1;
    nw->preedit.highlight.list[0].position = 0;

    HawPreeditSetCursorPosition( neww, nw->preedit.string_length );
    changed = True;
    nw->preedit.value = NULL;
  }

  if( nw->preedit.cursor_position != cw->preedit.cursor_position )
    changed = True;

  if( nw->preedit.has_focus != cw->preedit.has_focus ) {
    if( nw->preedit.has_focus ) {
      PaintDestination( nw );
      DrawInsertionPoint( nw, True );
    }
    else {
      DrawInsertionPoint( nw, False );
      PaintDestination( nw );
    }
  }

  return( changed );
}

static Boolean ReplaceText( HawPreeditWidget w, XEvent *event,
    Position replace_prev, Position replace_next,
    hchar_t *insert, int insert_length )
{
  int			replace_length, ii, jj;
  hchar_t		*src, *dst;
  int			delta = 0;
  Position		cursorPos;
  hchar_t		*orig_insert;
  HawHighlightMode	highlight_mode;
  HighlightRec		*l;
  int			number;

  if( !TextEditable(w) ) {
    if( w->preedit.verify_bell )
      XBell( XtDisplay((Widget)w), 0 );
    return( False );
  }

  if( w->preedit.cursor_on )
    DrawInsertionPoint( w, False );

  highlight_mode = w->preedit.highlight.list[FindHighlight( w, replace_prev )].mode;
  orig_insert = insert;
  replace_length = replace_next - replace_prev;

  if( (w->preedit.string_length + insert_length - replace_length) - 
      w->preedit.max_length > 0) { 
    if( w->preedit.verify_bell )
      XBell( XtDisplay( (Widget)w), 0 );
    if( insert != orig_insert )
      XtFree( (char *)insert );
    return( False );
  }

  if( w->preedit.string_length + insert_length - replace_length >=
      (w->preedit.alloc_size / sizeof(hchar_t)) ) {
    w->preedit.alloc_size += max( insert_length + 1, ALLOC_SIZE );
    TextHcsValue(w) = (hchar_t *)XtRealloc( (char *)TextHcsValue(w),
	w->preedit.alloc_size );
  }

  delta = insert_length - replace_length;

  if( replace_length > insert_length ) {
    for( src = TextHcsValue(w) + replace_next, dst = src + delta,
	   ii = (w->preedit.string_length + 1) - replace_next;
	 ii > 0; ++src, ++dst, --ii )
      *dst = *src;
    l = w->preedit.highlight.list;
    for( ii = w->preedit.highlight.number - 1; ii >= 0; ii-- )
      if( replace_next > l[ii].position )
	break;
    jj = ii;
    jj++;
    while( jj < w->preedit.highlight.number ) {
      l[jj].position += delta;
      jj++;
    }
    if( (ii + 1) < w->preedit.highlight.number ) {
      number = 0;
      while( l[ii].position >= l[ii+number+1].position &&
	     (ii+number+1) < w->preedit.highlight.number )
	number++;
      if( number > 0 ) {
	jj = ii;
	while( jj < w->preedit.highlight.number ) {
	  l[jj] = l[jj+number];
	  jj++;
	}
	w->preedit.highlight.number -= number;
	if( ii > 0 && l[ii].mode == l[ii-1].mode ) {
	  while( ii < w->preedit.highlight.number ) {
	    l[ii] = l[ii+1];
	    ii++;
	  }
	  w->preedit.highlight.number--;
	}
      }
    }
  }
  else if( replace_length < insert_length ) {
    for( src = TextHcsValue(w) + w->preedit.string_length,
	   dst = src + delta,
	   ii = (w->preedit.string_length + 1) - replace_next;
	 ii > 0; src--, dst--, ii-- )
      *dst = *src;
    l = w->preedit.highlight.list;
    for( ii = w->preedit.highlight.number - 1; ii > 0; ii-- )
      if( replace_next > l[ii].position )
	break;
    ii++;
    while( ii < w->preedit.highlight.number ) {
      l[ii].position += delta;
      ii++;
    }
    PreeditAreaSetHighlight( w, replace_prev,
	replace_prev + insert_length, highlight_mode );
  }

  if( insert_length != 0 ) {
    for( src = insert, dst = TextHcsValue(w) + replace_prev,
	   ii = insert_length;
	 ii > 0; ++src, ++dst, --ii )
      *dst = *src;
  }

  w->preedit.string_length += delta;
  cursorPos = replace_next + delta;
  CalculateLine( w, replace_prev );
  SetCursorPosition( w, cursorPos );

  if( replace_length == insert_length )
    replace_next = replace_prev + replace_length;
  else
    replace_next = w->preedit.string_length;
  if( FindLine(w, replace_next) + 1 == w->preedit.line.number )
    replace_prev = w->preedit.line.position[FindLine(w, replace_next)];
  RedisplayText( w, replace_prev, replace_next );

  if( insert != orig_insert )
    XtFree( (char *)insert );
  return( True );
}

static void Initialize( Widget request, Widget new,
			ArgList args, Cardinal *num_args )
{
  HawPreeditWidget	rw = (HawPreeditWidget)request;
  HawPreeditWidget	nw = (HawPreeditWidget)new;
  XFontStruct		**fs;
  char			**fn;
  int			n;
  hchar_t		*hcs;

  if( TextCursorPosition(rw) < 0 ) {
    XtWarning( "Invalid cursor position, must be >= 0." );
    TextCursorPosition(nw) = 0;
  }

  if( TextColumns(rw) <= 0 ) {
    XtWarning( "Invalid columns, must be > 0." );
    TextColumns(nw) = 1;
  }

  nw->core.sensitive = False;
  n = XFontsOfFontSet( TextFontSet(nw), &fs, &fn );
  nw->preedit.ascent = 0;
  nw->preedit.descent = 0;
  while( --n >= 0 ) {
    if( nw->preedit.ascent < fs[n]->ascent )
      nw->preedit.ascent = fs[n]->ascent;
    if( nw->preedit.descent < fs[n]->descent )
      nw->preedit.descent = fs[n]->descent;
  }
  if( nw->preedit.line_spacing == 0 )
    nw->preedit.line_spacing = nw->preedit.ascent + nw->preedit.descent;
  nw->core.height = nw->preedit.line_spacing;
  if( nw->core.height > nw->preedit.line_spacing )
    nw->core.height = nw->preedit.line_spacing;

  if( nw->preedit.area_needed_width <= 0 )
    nw->preedit.area_needed_width = nw->preedit.width;
  nw->preedit.area_needed_height = nw->preedit.ascent + nw->preedit.descent;

  nw->preedit.cursor_on = True;
  nw->preedit.dest_on = False;

  nw->preedit.gc = NULL;
  nw->preedit.inv_gc = NULL;
  nw->preedit.xor_gc = NULL;

  if( TextHcsValue(nw) == NULL ) {
    if( TextValue(nw) == NULL ) {
      nw->preedit.string_length = 0;
      nw->preedit.alloc_size = sizeof(hchar_t) * ALLOC_SIZE;
      TextHcsValue(nw) = (hchar_t *)XtMalloc( nw->preedit.alloc_size );
      TextHcsValue(nw)[0] = '\0';
    }
    else
      ValidateString( nw, TextValue(nw) );
  }
  else {
    hcs = TextHcsValue(nw);
    nw->preedit.string_length = hcslen( hcs );
    nw->preedit.alloc_size = sizeof(hchar_t) *
      ( nw->preedit.string_length + ALLOC_SIZE );
    TextHcsValue(nw) = hcscpy( (hchar_t *)XtMalloc( nw->preedit.alloc_size ),
			      hcs );
    TextValue(nw) = NULL;
  }

  if( TextMaxLength(nw) < MAXINT ) {
    if( nw->preedit.string_length > TextMaxLength(nw) ) {
      nw->preedit.string_length = TextMaxLength(nw);
      TextHcsValue(nw)[nw->preedit.string_length] = '\0';
    }
  }

  nw->preedit.cursor_height = nw->preedit.cursor_width = 0;

  nw->preedit.highlight.number = nw->preedit.highlight.maximum = 1;
  nw->preedit.highlight.list
    = (HighlightRec *)XtMalloc( (unsigned)sizeof(HighlightRec) );
  nw->preedit.highlight.list[0].position = 0;
  nw->preedit.highlight.list[0].mode = HawHIGHLIGHT_SECONDARY_SELECTED;

  nw->preedit.line.number = 1;
  nw->preedit.line.maximum = 2;
  nw->preedit.line.position
    = (Position *)XtMalloc( (unsigned)sizeof(Position) * 2 );
  nw->preedit.line.position[0] = 0;
  nw->preedit.line.position[1] = 0;

  nw->preedit.timer_id = None;
  nw->preedit.cursor_width = 3;

  nw->core.x = nw->preedit.x;
  nw->core.y = nw->preedit.y;
  nw->core.width = nw->preedit.width;

  TextCursorPosition(nw) = 0;
  TextColumns(nw) = 1;
}

static void Realize( Widget widget, Mask *vmask, XSetWindowAttributes *attr )
{
  HawPreeditWidget	pw = (HawPreeditWidget)widget;
  Display		*d = XtDisplay(widget);
  XSetWindowAttributes	at;

  attr->border_pixel = pw->preedit.foreground;
  *vmask |= CWBitGravity | CWEventMask;
  attr->bit_gravity = NorthWestGravity;
  attr->event_mask |= KeyPressMask;
  pw->preedit.event_mask = attr->event_mask;

  pw->core.window = XCreateWindow(d,
      pw->preedit.focus_window,
      pw->core.x, widget->core.y,
      pw->core.width, widget->core.height,
      pw->core.border_width,
      pw->core.depth, InputOutput,
      CopyFromParent, *vmask, attr);

  LoadGCs( pw, pw->core.background_pixel, pw->preedit.foreground );
  MakeCursors( pw );
}

static void Resize(Widget w)
{
  HawPreeditWidget pw = (HawPreeditWidget)w;
  
  pw->core.height = pw->preedit.line_spacing;
  RedisplayText( (HawPreeditWidget)w, TextCursorPosition(pw),
      pw->preedit.string_length );
}

static void Redisplay( Widget w, XEvent *event, Region region )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  Position		x, y;
  Window		window;

  if( !pw->core.sensitive )
    return;

  if( pw->preedit.cursor_on || pw->preedit.dest_on ) {
    GetWXYFromPos( pw, TextCursorPosition(pw), &window, &x, &y );
    x -= pw->preedit.cursor_width;
    y -= pw->preedit.cursor_height;
    XFillRectangle( XtDisplay(w), window, pw->preedit.inv_gc,
		    x, y, pw->preedit.cursor_width, pw->core.height );
    pw->preedit.cursor_on = False;
    pw->preedit.dest_on = False;
  }
  RedisplayText( pw, 0, pw->preedit.string_length );
}

HawPreeditClassRec hawPreeditClassRec = 
{
  {
    /* superclass	     */ (WidgetClass) &widgetClassRec,
    /* class_name	     */ "HawPreedit",
    /* widget_size	     */ sizeof(HawPreeditRec),
    /* class_initialize	     */ ClassInitialize,
    /* class_part_initialize */ NULL,
    /* class_inited          */ False,
    /* initialize	     */ Initialize,
    /* initialize_hook       */ NULL,
    /* realize		     */ Realize,
    /* actions               */ NULL,
    /* num_actions    	     */ 0,
    /* resources	     */ local_resources,
    /* num_resources         */ XtNumber(local_resources),
    /* xrm_class	     */ NULLQUARK,
    /* compress_motion       */ FALSE,
    /* compress_exposure     */ TRUE,
    /* compress_enterleave   */ FALSE,
    /* visible_interest      */ FALSE,
    /* destroy               */ Destroy,
    /* resize                */ Resize,
    /* expose                */ Redisplay,
    /* set_values	     */ SetValues,
    /* set_values_hook       */ NULL,
    /* set_values_almost     */ XtInheritSetValuesAlmost,
    /* get_values_hook       */ NULL,
    /* accept_focus	     */ NULL,
    /* version               */ XtVersion,
    /* callback private      */ NULL,
    /* tm_table              */ NULL,
    /* query_geometry        */ XtInheritQueryGeometry,
    /* display_accelerator   */ NULL,
    /* extension             */ NULL,
   },
   {
    /* extension	     */ NULL,
   },

};

WidgetClass hawPreeditClass = (WidgetClass)&hawPreeditClassRec;

Widget HawCreatePreedit( Widget parent, char *name, ArgList args,
			 Cardinal argc )
{
  Widget	pe;
  Window	win;
  Position	x, y, w, h;
  int		n;

  x = y = w = h = 0;
  for( n = 0; n < argc; n++ ) {
    if( strcmp( args[n].name, HawNareaX ) == 0 )
      x = args[n].value;
    else if( strcmp( args[n].name, HawNareaY ) == 0 )
      y = args[n].value;
    else if( strcmp( args[n].name, HawNareaWidth ) == 0)
      w = args[n].value ? args[n].value : 10;
    else if( strcmp( args[n].name, HawNareaHeight ) == 0)
      h = args[n].value ? args[n].value : 10;
  }

  pe = XtCreateManagedWidget( name, hawPreeditClass, parent,
      args, argc );
  XtVaSetValues( pe, XtNx, x, XtNy, y,
      XtNwidth, w, XtNheight, h,
      NULL);
  if( (win = ((HawPreeditWidget)pe)->preedit.focus_window) != None )
  {
#if 0 
      XReparentWindow( XtDisplay(pe), XtWindow(pe), win, x, y );
#else
      XWindowAttributes xwa;
      XSetWindowAttributes xswa;
      Window child;

      XGetWindowAttributes(XtDisplay(pe), win, &xwa);

      xswa.override_redirect = True;
      XChangeWindowAttributes(XtDisplay(pe), XtWindow(pe),
	      CWOverrideRedirect, &xswa);
      XTranslateCoordinates(XtDisplay(pe), win, DefaultRootWindow(XtDisplay(pe)),
	      xwa.x, xwa.y, &xwa.x, &xwa.y, &child);
      xwa.x = xwa.y = 0;
      XReparentWindow( XtDisplay(pe), XtWindow(pe),
	     DefaultRootWindow(XtDisplay(pe)), xwa.x + x, xwa.y + y);
#endif
  }

  XtAddGrab( pe, False, False );
  return( pe );
}

void HawPreeditSetHighlight( Widget w, Position left,
    Position right,HawHighlightMode mode )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;

  if( left >= right || right <= 0 )	return;

  if( left < 0 )	left = 0;

  if( right > pw->preedit.string_length )
    right = pw->preedit.string_length;

  PreeditAreaSetHighlight( pw, left, right, mode );

  RedisplayText( pw, left, right );
}

Boolean	HawPreeditReplace( Widget w, Position from_pos,
    Position to_pos, char *value )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  Boolean		save_editable = TextEditable(pw);
  Boolean		result;
  int			len;
  hchar_t		*hcs;

  if( from_pos < 0 )
    from_pos = 0;

  if( to_pos >= pw->preedit.string_length )
    to_pos = pw->preedit.string_length;

  if( value == NULL )
    value = "";
  len = strlen( value );
  hcs = (hchar_t *)XtMalloc( sizeof(hchar_t) * ( len + 1 ) );
  len = mbstohcs( hcs, value, len );

  TextEditable(pw) = True;
  result = ReplaceText( pw, NULL, from_pos, to_pos, hcs, len );
  TextEditable(pw) = save_editable;
  XtFree( (char *)hcs );
  return( result );
}

Boolean HawPreeditReplaceHcs( Widget w, Position from_pos, Position to_pos,
    hchar_t *value )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  Boolean		save_editable = TextEditable(pw);
  Boolean		result;
  hchar_t		*hcs;

  if( from_pos < 0 )
    from_pos = 0;
  
  if( to_pos >= pw->preedit.string_length )
    to_pos = pw->preedit.string_length;
  
  if( value == NULL ) {
    hcs = (hchar_t *)XtMalloc( sizeof(hchar_t) );
    hcs[0] = '\0';
  }
  else
    hcs = value;
  
  TextEditable(pw) = True;
  result = ReplaceText(pw, NULL, from_pos, to_pos, hcs, hcslen(hcs));
  TextEditable(pw) = save_editable;
  if( value == NULL )
    XtFree( (char *)hcs );
  return( result );
}

void HawPreeditInsert( Widget w, Position position, char *value )
{
  HawPreeditReplace( w, position, position, value );
}

void HawPreeditInsertHcs( Widget w, Position position, hchar_t *value )
{
  HawPreeditReplaceHcs( w, position, position, value );
}

void HawPreeditSetEditable( Widget w, Boolean editable )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  
  TextEditable(pw) = editable;
}

void HawPreeditMoveWidget( Widget w, Position position, Widget sh )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  Window		win;
  Position		x, y, newx, newy;
  int			xx, yy;
  unsigned int		width, height, border, depth;
  Display		*display = XtDisplay(w);
  int			screen = DefaultScreen(display);
  Window		root, child;
  XSizeHints		hints;
  long			supplied;
  Dimension		wi, he;

  XGetWMNormalHints( XtDisplay(sh), XtWindow(sh), &hints, &supplied );
  XtVaGetValues( sh, XtNwidth, &wi, XtNheight, &he, NULL );
  hints.flags |= PWinGravity;

  GetWXYFromPos( pw, position, &win, &x, &y );
  y += pw->preedit.descent;
  XGetGeometry( display, win, &root, &xx, &yy, &width, &height,
		&border, &depth );
  XTranslateCoordinates( display, win, root, (int)x, (int)y,
			 &xx, &yy, &child );
  newx = xx;
  newy = yy;
  hints.win_gravity = NorthWestGravity;

  width = DisplayWidth( display, screen );
  height = DisplayHeight( display, screen );
  if( xx + wi > width ) {
    newx -= xx + wi - width;
    hints.win_gravity = NorthEastGravity;
  }
  if( yy + he > height ) {
    newy -= pw->preedit.descent + pw->preedit.ascent + he;
    hints.win_gravity = (hints.win_gravity == NorthWestGravity) ?
      SouthWestGravity : SouthEastGravity;
  }
  XSetWMNormalHints( XtDisplay(sh), XtWindow(sh), &hints );
  XtMoveWidget( sh, newx, newy );
  XtMapWidget( sh );
}

void HawPreeditSetString( Widget w, char *value, HawHighlightMode mode )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;

  if( value == NULL )
    value = "";
  ValidateString( pw, value );
  CalculateLine( pw, 0 );
  pw->preedit.top_line = 0;
  pw->preedit.highlight.number = 1;
  pw->preedit.highlight.list[0].position = 0;
  pw->preedit.highlight.list[0].mode = mode;

  HawPreeditSetCursorPosition( (Widget)pw, pw->preedit.string_length );
  RedisplayText( pw, 0, pw->preedit.string_length );
}

void HawPreeditSetStringHcs( Widget w, hchar_t *value, HawHighlightMode mode )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;

  XtFree( (char *)TextHcsValue(pw) );
  if( value == NULL ) {
    pw->preedit.string_length = 0;
    pw->preedit.alloc_size = sizeof(hchar_t) * ALLOC_SIZE;
    TextHcsValue(pw) = (hchar_t *)XtMalloc( pw->preedit.alloc_size );
    TextHcsValue(pw)[0] = '\0';
  }
  else {
    pw->preedit.string_length = hcslen( value );
    pw->preedit.alloc_size = sizeof(hchar_t) *
      ( pw->preedit.string_length + ALLOC_SIZE );
    TextHcsValue(pw) =
      hcscpy( (hchar_t *)XtMalloc( pw->preedit.alloc_size ), value );
  }

  CalculateLine( pw, 0 );
  pw->preedit.highlight.number = 1;
  pw->preedit.highlight.list[0].position = 0;
  pw->preedit.highlight.list[0].mode = mode;

  HawPreeditSetCursorPosition( w, pw->preedit.string_length );
  RedisplayText( pw, 0, pw->preedit.string_length );
}

void HawPreeditMapWidget( Widget w )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;

  pw->preedit.top_line = 0;
  pw->preedit.max_lines = 1;

  pw->core.sensitive = True;
  TextEditable(pw) = True;
  pw->preedit.cursor_on = pw->preedit.dest_on = False;
#ifdef DEBUG
  if( resources.debug & 0x80 ) {
    printf("===== MapWidget: %d %d (%d, %d)\n",
	pw->core.x, pw->core.y, pw->core.width, pw->core.height );
  }
#endif
#if 0 // xxx
  XMoveResizeWindow( XtDisplay(pw), pw->core.window,
      pw->core.x, pw->core.y, pw->core.width, pw->core.height );
  XMapWindow( XtDisplay(pw), pw->core.window );
#else
    {
      XWindowAttributes xwa;
      Window child;

      XGetWindowAttributes(XtDisplay(pw), pw->preedit.focus_window, &xwa);
      XTranslateCoordinates(XtDisplay(pw), pw->preedit.focus_window,
	      DefaultRootWindow(XtDisplay(w)), xwa.x, xwa.y,
	      &xwa.x, &xwa.y, &child);
      xwa.x = xwa.y = 0;
      XMoveResizeWindow( XtDisplay(pw), XtWindow(pw),
		xwa.x + pw->core.x, xwa.y + pw->core.y,
		pw->core.width, pw->core.height );
    }
    if (pw->preedit.string_length && pw->preedit.has_focus)
      XMapWindow( XtDisplay(pw), pw->core.window );
#endif
}

void HawPreeditUnmapWidget( Widget w )
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;
  
  pw->core.sensitive = False;
  pw->preedit.has_focus = False;
  XUnmapWindow( XtDisplay(pw), pw->core.window );
}

void HawPreeditAutoReplace(Widget w)
{
  HawPreeditWidget	pw = (HawPreeditWidget)w;

  XUnmapWindow( XtDisplay(pw), pw->core.window );
  XSync( XtDisplay(pw), False );

  pw->core.width = pw->preedit.cursor_width;
  pw->core.x -= pw->preedit.x;
  pw->core.x += HawTextEscapement( TextFontSet(pw), TextHcsValue(pw),
				   pw->preedit.string_length );
  //  ġ  ̸ 0  ȴ. 
  if( pw->preedit.width == 0 )
    pw->preedit.width = 1;
  pw->core.y += (pw->core.x / pw->preedit.width) * pw->preedit.line_spacing;
  pw->core.x %= pw->preedit.width;
  pw->core.x += pw->preedit.x;

  HawPreeditSetString( w, "", HawHIGHLIGHT_SECONDARY_SELECTED );
 
#ifdef DEBUG
  if( resources.debug & 0x80 ) {
    printf("===== AutoReplace:\n");
    printf("x = %d, y = %d, w = %d, h = %d\n",
	   pw->core.x, pw->core.y, pw->core.width, pw->core.height );
  }
#endif
  HawPreeditMapWidget(w);
}

void HawPreeditSetCursorPosition( Widget w, Position position )
{
  SetCursorPosition( (HawPreeditWidget)w, position );
}
