/* $Id: StsArea.c,v 1.4 1999/02/09 15:45:13 frog Exp $ */
#include	<ctype.h>
#include	<Haw/StsAreaP.h>
#include	<X11/Shell.h>

static XtResource resources[] =
{
  {
    HawNmarginWidth,
    HawCMarginWidth, XtRDimension, sizeof(Dimension),
    XtOffset(HawStatusAreaWidget, status_area.margin_width),
    XtRImmediate, (XtPointer)5 },
  {
    HawNmarginHeight,
    HawCMarginHeight, XtRDimension, sizeof(Dimension),
    XtOffset(HawStatusAreaWidget, status_area.margin_height),
    XtRImmediate, (XtPointer)2 },
  {
    HawNspacing,
    HawCSpacing, XtRDimension, sizeof(Dimension),
    XtOffset(HawStatusAreaWidget, status_area.spacing),
    XtRImmediate, (XtPointer)10 },
  {
    HawNparentWindow,
    HawCParentWindow, XtRWindow, sizeof(Window),
    XtOffset(HawStatusAreaWidget, status_area.parent_window),
    XtRWindow, NULL	},
  {
    HawNcursor,
    XtCCursor, XtRCursor, sizeof(Cursor),
    XtOffset(HawStatusAreaWidget, status_area.cursor),
    XtRImmediate, (XtPointer)0 },
  {
    HawNareaNeededWidth,
    XtCWidth, XtRDimension, sizeof(Dimension),
    XtOffset(HawStatusAreaWidget, status_area.needed_width),
    XtRImmediate, (XtPointer)0 },
  {
    HawNareaNeededHeight,
    XtCHeight, XtRDimension, sizeof(Dimension),
    XtOffset(HawStatusAreaWidget, status_area.needed_height),
    XtRImmediate, (XtPointer)0 }
};

WidgetClass	hawStatusAreaWidgetClass = (WidgetClass)&hawStatusAreaClassRec;

static void
Initialize( Widget request, Widget new, ArgList args, Cardinal *num_args )
{
  HawStatusAreaWidget	nw = (HawStatusAreaWidget)new;

  if( nw->core.width == 0 )
    nw->core.width = 1;
  if( nw->core.height == 0 )
    nw->core.height = 1;
}


static void LayoutChildren( HawStatusAreaWidget aw )
{
  Widget	widget;
  Dimension	max_height;
  Dimension	widget_height;
  Position	x, y;
  int		ii;

  x = aw->status_area.margin_width;
  max_height = 0;
  for( ii = 0; ii < aw->composite.num_children; ii++ ) {
    widget = aw->composite.children[ii];
    if( XtIsManaged(widget) ) {
      y = ( aw->core.height - widget->core.height ) / 2
	- widget->core.border_width;
      XtMoveWidget( widget, x, y );
      x += widget->core.width + aw->status_area.spacing;

      widget_height = widget->core.height
	+ widget->core.border_width * 2;
      if( widget_height > max_height )
	max_height = widget_height;
    }
  }

  aw->status_area.needed_width = x;
  aw->status_area.needed_height = max_height +
    aw->status_area.margin_height * 2;
}


static void Resize( Widget w )
{
  LayoutChildren( (HawStatusAreaWidget)w );
}


static Boolean
SetValues( HawStatusAreaWidget current,
	   HawStatusAreaWidget request,
	   HawStatusAreaWidget neww,
	   ArgList args, Cardinal *nargs )
{
  Boolean	change = False;
  Dimension	bw;

  if( *nargs == 0 )
    return FALSE;

  if( current->core.height != neww->core.height )
    LayoutChildren( neww );
  if( current->core.x != neww->core.x ||
      current->core.y != neww->core.y  ||
      current->core.width != neww->core.width ||
      current->core.height != neww->core.height )
  {
    XtConfigureWidget( XtParent(neww), neww->core.x, neww->core.y,
		       neww->core.width, neww->core.height, bw );
    XMoveResizeWindow( XtDisplay(neww), XtWindow(neww),
		       neww->core.x, neww->core.y,
		       neww->core.width, neww->core.height );
  }

  if( current->status_area.margin_width !=
      neww->status_area.margin_width ||
      current->status_area.margin_height !=
      neww->status_area.margin_height ||
      current->status_area.spacing != neww->status_area.spacing ) {
    LayoutChildren( neww );
    change = True;
  }

  if( current->status_area.parent_window !=
      neww->status_area.parent_window ) {
    XReparentWindow( XtDisplay(neww), XtWindow(neww),
		     neww->status_area.parent_window,
		     neww->core.x, neww->core.y );
  }

  if( current->status_area.cursor != neww->status_area.cursor )
    XDefineCursor( XtDisplay(neww), XtWindow(neww),
		   neww->status_area.cursor );

  return( change );
}


static XtGeometryResult
QueryGeometry(HawStatusAreaWidget aw,
	      XtWidgetGeometry *proposed,
	      XtWidgetGeometry *desired)
{
  desired->width  = aw->core.width;
  desired->height = aw->core.height;
  desired->request_mode = CWWidth | CWHeight;

  if( (proposed->request_mode & CWWidth)  &&
      (proposed->width >= desired->width)  &&
      (proposed->request_mode & CWHeight)  &&
      (proposed->height >= desired->height) ) {
    return( XtGeometryYes );
  }

  return( XtGeometryAlmost );
}


static XtGeometryResult
GeometryManager(Widget 	w, XtWidgetGeometry *request,
		XtWidgetGeometry *reply)
{
  if( request->request_mode & CWWidth )
    w->core.width = request->width;

  if( request->request_mode & CWHeight )
    w->core.height = request->height;

  if( request->request_mode & CWBorderWidth )
    w->core.border_width = request->border_width;

  if( request->request_mode & (CWWidth | CWHeight | CWBorderWidth) )
    LayoutChildren( (HawStatusAreaWidget)XtParent( w ) );
	    
  return( XtGeometryYes );
}

static void ChangeManaged( HawStatusAreaWidget aw )
{
  LayoutChildren( aw );
}

HawStatusAreaClassRec	hawStatusAreaClassRec =
{
  { /* core_class fields      */
    /* superclass             */ (WidgetClass)&constraintClassRec,
    /* class_name             */ "HawStatusArea",
    /* widget_size            */ sizeof(HawStatusAreaWidgetRec),
    /* class_initialize       */ NULL,
    /* class_part_initialize  */ NULL,
    /* class_inited           */ False,
    /* initialize             */ Initialize,
    /* initialize_hook        */ NULL,
    /* realize                */ XtInheritRealize,
    /* actions                */ NULL,
    /* num_actions            */ 0,
    /* resources              */ resources,
    /* num_resources          */ XtNumber(resources),
    /* xrm_class              */ NULLQUARK,
    /* compress_motion        */ True,
    /* compress_exposure      */ XtExposeCompressMaximal,
    /* compress_enterleave    */ True,
    /* visible_interest       */ False,
    /* destroy                */ NULL,
    /* resize                 */ Resize,
    /* expose                 */ XtInheritExpose,
    /* set_values             */ (XtSetValuesFunc)SetValues,
    /* set_values_hook        */ NULL,
    /* set_values_almost      */ XtInheritSetValuesAlmost,
    /* get_values_hook        */ NULL,
    /* accept_focus           */ NULL,
    /* version                */ XtVersion,
    /* callback_private       */ NULL,
    /* tm_table               */ NULL,
    /* query_geometry         */ (XtGeometryHandler)QueryGeometry,
    /* display_accelerator    */ NULL,
    /* extension              */ NULL,
  },
  { /* composite_class fields */
    /* geometry_handler       */ (XtGeometryHandler)GeometryManager,
    /* change_managed         */ (XtWidgetProc)ChangeManaged,
    /* insert_child           */ XtInheritInsertChild,
    /* delete_child           */ XtInheritDeleteChild,
    /* extension              */ NULL,
  },
  { /* constraint_class fields*/
    /* resources              */ NULL,
    /* num_resources          */ 0,
    /* constraint_size        */ 0,
    /* initialize             */ NULL,
    /* destroy                */ NULL,
    /* set_values             */ NULL,
    /* extension              */ NULL,
  },
  { /* status_area_class fields */
    /* extension              */ NULL,
  }
};


Widget
HawCreateManagedStatusArea( Widget parent, String name, ArgList args,
			    Cardinal argc )
{
  Widget	w, override;
  Window	win;
  Arg		*merged_args;
  char		*ss, *tt;
  int		n;
  Position	x, y;

  ss = XtMalloc( strlen(name) + 3 );	/* name + "OS" + NULL */
  tt = XtMalloc( strlen(name) + 3 );
  strcpy( ss, name );
  strcat( ss, "OS" );
  ss[0] = tolower( ss[0] );
  strcpy( tt, ss );
  tt[0] = toupper( tt[0] );

  x = y = 0;
  merged_args = (ArgList)XtCalloc( argc+1, sizeof(Arg) );
  for( n = 0; n < argc; n++ ) {
    if( strcmp( args[n].name, XtNx ) == 0 )
      x = args[n].value;
    if( strcmp( args[n].name, XtNy ) == 0 )
      y = args[n].value;
    if( strcmp( args[n].name, XtNwidth ) == 0  &&
	args[n].value == 0 )
      args[n].value = 10;
    if( strcmp( args[n].name, XtNheight ) == 0  &&
	args[n].value == 0 )
      args[n].value = 10;
    merged_args[n].name = args[n].name;
    merged_args[n].value = args[n].value;
  }
  XtSetArg( merged_args[n], XtNmappedWhenManaged, False );	n++;

  override = XtAppCreateShell( ss, tt, overrideShellWidgetClass,
			       XtDisplay(parent), merged_args, n );
  XtFree( ss );
  XtFree( tt );
  XtFree( (char *)merged_args );

  w = XtCreateManagedWidget( name, hawStatusAreaWidgetClass, override, args,
			     argc );
  XtRealizeWidget( override );
  if( (win = ((HawStatusAreaWidget)w)->status_area.parent_window) != None )
    XReparentWindow( XtDisplay(w), XtWindow(w), win, x, y );
  XtAddGrab( w, False, False );
  return( w );
}
