/* $Id: TextOut.c,v 1.10 1999/04/17 11:03:35 frog Exp $ */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<X11/Xlib.h>
#include	<X11/Xlibint.h>
#include	<X11/keysym.h>
#include	<X11/IntrinsicP.h>
#include	<X11/StringDefs.h>
#include	<X11/Xaw/XawInit.h>
#include	<Haw/TextOutP.h>
#include	<Haw/HawStrDefs.h>
#include	<Haw/Util.h>

#include	"hcodec.h"
#include	"hstring.h"

#define MAXINT		2147483647
#define ALLOC_SIZE	32

void HawCvtStringToValueHcs( XrmValuePtr args, Cardinal *num_args,
    XrmValue *from_val, XrmValue *to_val )
{
  register char		*in_str = (char *)(from_val->addr);
  register int		len;
  static int		hlen = 0;
  static hchar_t	*hcs;

  to_val->size = sizeof(hchar_t *);
  to_val->addr = (XPointer)&hcs;

#ifndef USE_JOHAB_CODE
  len = mbstohcs( NULL, in_str, strlen(in_str) );
  hcs = (hchar_t *)XtRealloc( (XtPointer)hcs, sizeof(hchar_t) * (len + 1) );
#else
  if( hlen < (len = strlen(in_str)) ) {
    hlen = len;
    hcs = (hchar_t *)XtRealloc( (XtPointer)hcs, sizeof(hchar_t) * (len + 1) );
  }
#endif /* USE_JOHAB_CODE */
  mbstohcs( hcs, in_str, strlen(in_str) );
}


static void ValidateString( HawTextOutWidget w, char *value )
{
  int		len = strlen(value);
  hchar_t	*hs = (hchar_t *)XtMalloc(sizeof(hchar_t)*(len + 1));

  len = mbstohcs( hs, value, len );
  if( len != -1 ) {
    XtFree( (char *)TextHcsValue(w) );
    w->text_out.string_length = len;
    w->text_out.alloc_size = sizeof(hchar_t) * ( len + ALLOC_SIZE );
    TextHcsValue(w) = hcscpy( (hchar_t *)XtMalloc( w->text_out.alloc_size ),
			     hs );
    XtFree( (char *)hs );
  }
  TextValue(w) = NULL;
}


static Boolean
GetXYFromPos( HawTextOutWidget w, Position position, Position *x, Position *y )
{
  hchar_t	*hvalue = TextHcsValue(w);

  *x = w->text_out.cursor_width + w->text_out.left_margin;
  *y = w->text_out.ascent + w->text_out.top_margin;

  if( position > w->text_out.string_length )
    return( False );

  *x += HawTextEscapement( TextFontSet(w), hvalue, position );
  if( w->text_out.attachment == HawCURSOR_LEFT )
    *x -= HawTextEscapement( TextFontSet(w), hvalue,
			     w->text_out.left_cursor_position );
  else
    *x += w->core.width - w->text_out.cursor_width
      - HawTextEscapement( TextFontSet(w), hvalue,
			   w->text_out.right_cursor_position );
  return( True );
}

static void PaintCursor( HawTextOutWidget w )
{
  Position	x, y;

  GetXYFromPos( w, TextCursorPosition(w), &x, &y );
  x -= w->text_out.cursor_width;
  y -= w->text_out.cursor_height;
  XCopyArea( XtDisplay(w), w->text_out.cursor, XtWindow(w),
	     w->text_out.xor_gc, 0, 0,
	     w->text_out.cursor_width, w->text_out.cursor_height, x, y );
  w->text_out.cursor_on = !w->text_out.cursor_on;
}

static void PaintCaret( HawTextOutWidget w )
{
  Position	x, y;

  GetXYFromPos( w, TextCursorPosition(w), &x, &y );
  x -= w->text_out.cursor_width -
    ((w->text_out.cursor_width - CARET_WIDTH) >> 1);
  y -= CARET_HEIGHT - w->text_out.descent;
  XCopyArea( XtDisplay(w), w->text_out.dest_cursor, XtWindow(w),
	     w->text_out.xor_gc, 0, 0,
	     CARET_WIDTH, CARET_HEIGHT, x, y );
  w->text_out.caret_on = !w->text_out.caret_on;
}

static void LoadGCs( HawTextOutWidget w, Pixel bg, Pixel fg )
{
  Display	*display = XtDisplay(w);
  Window	window = XtWindow(w);
  XGCValues	values;
  unsigned long	mask;

  mask = 0;
  values.function = GXcopy;	mask |= GCFunction;
  values.foreground = fg;	mask |= GCForeground;
  values.background = bg;	mask |= GCBackground;
  values.fill_style = FillSolid; mask |= GCFillStyle;
  values.line_width = (w->text_out.ascent > 19) ? 2 : 1;
  mask |= GCLineWidth;
  values.graphics_exposures = False;	mask |= GCGraphicsExposures;
  if( w->text_out.gc != NULL )
    XChangeGC( display, w->text_out.gc, mask, &values );
  else
    w->text_out.gc = XCreateGC( display, window, mask, &values );

  values.foreground = bg;	mask |= GCForeground;
  values.background = fg;	mask |= GCBackground;
  if( w->text_out.inv_gc != NULL )
    XChangeGC( display, w->text_out.inv_gc, mask, &values );
  else
    w->text_out.inv_gc = XCreateGC( display, window, mask, &values );

  values.function = GXxor;
  values.graphics_exposures = FALSE;
  values.foreground = fg ^ bg;
  values.background = 0;
  if( w->text_out.xor_gc != NULL )
    XtReleaseGC( (Widget)w, w->text_out.xor_gc );
  w->text_out.xor_gc = XtGetGC( (Widget)w, mask, &values );
}


static void MakeIBeamCursor( HawTextOutWidget w, int line_width )
{
  Screen	*screen = XtScreen(w);
  Display	*display = XtDisplay(w);
  Window	window = XtWindow(w);
  XSegment	segments[3];
  XtGCMask	mask;
  XGCValues	values;
  GC		gc;

  mask = GCForeground | GCBackground |GCLineWidth;
  values.foreground = 0;
  values.background = 0;
  values.line_width = line_width;
  gc = XCreateGC( display, window, mask, &values );

  w->text_out.cursor = XCreatePixmap( display, window,
				      w->text_out.cursor_width,
				      w->text_out.cursor_height,
				      screen->root_depth );

  XFillRectangle( display, w->text_out.cursor, gc, 0, 0,
		  w->text_out.cursor_width,
		  w->text_out.cursor_height );
  XSetForeground( display, gc,
		  w->text_out.foreground ^ w->core.background_pixel );

  segments[0].x1 = 0;
  segments[0].y1 = line_width - 1;
  segments[0].x2 = w->text_out.cursor_width;
  segments[0].y2 = line_width - 1;

  segments[1].x1 = 0;
  segments[1].y1 = w->text_out.cursor_height - 1;
  segments[1].x2 = w->text_out.cursor_width;
  segments[1].y2 = w->text_out.cursor_height - 1;

  segments[2].x1 = w->text_out.cursor_width / 2;
  segments[2].y1 = line_width;
  segments[2].x2 = w->text_out.cursor_width / 2;
  segments[2].y2 = w->text_out.cursor_height - 1;

  XDrawSegments( display, w->text_out.cursor, gc, segments, 3 );

  XFreeGC( display, gc );
}


static void MakeDestCursor( HawTextOutWidget w )
{
  Display	*dpy = XtDisplay(w);
  Screen	*screen = XtScreen(w);

  w->text_out.dest_cursor =
    XCreatePixmapFromBitmapData( dpy, XtWindow(w),
	(char*)caretBits, CARET_WIDTH, CARET_HEIGHT,
	w->core.background_pixel ^ w->text_out.foreground, 0,
	screen->root_depth );
}

static void MakeCursors( HawTextOutWidget w )
{
  int line_width = 1;

  if( !XtIsRealized((Widget)w) )
    return;

  w->text_out.cursor_width = 3;
  w->text_out.cursor_height = w->text_out.ascent;

  MakeDestCursor( w );

  if( w->text_out.cursor_height > 19 ) {
    w->text_out.cursor_width++;
    line_width = 2;
  }

  MakeIBeamCursor( w, line_width );
}


static void ClassInitialize( void )
{
  XawInitializeWidgetSet();
  XtAddConverter( XtRString, HawRHcsValue, HawCvtStringToValueHcs, NULL, 0 );
}

static void Destroy( Widget w )
{
  HawTextOutWidget	pw = (HawTextOutWidget)w;
  Display		*display = XtDisplay(w);

  if( pw->text_out.timer_id )
    XtRemoveTimeOut( pw->text_out.timer_id );
  XFreeGC( display, pw->text_out.gc );
  XFreeGC( display, pw->text_out.inv_gc );
  XtReleaseGC( w, pw->text_out.xor_gc );
  XtFree( (char *)pw->text_out.highlight.list );
  XtFree( pw->text_out.value );
}

static Boolean SetValues( Widget current, Widget request, Widget new,
    ArgList args, Cardinal *num_args )
{
  HawTextOutWidget	cw = (HawTextOutWidget)current;
  HawTextOutWidget	rw = (HawTextOutWidget)request;
  HawTextOutWidget	nw = (HawTextOutWidget)new;
  Boolean		changed = False;
  Display		*d = XtDisplay( request );
  XFontStruct		**fs;
  char			**fn;
  int			n;
  XFontSetExtents	*fse;

  if( nw->text_out.left_margin != cw->text_out.left_margin )
    nw->text_out.left_margin = cw->text_out.left_margin;
  if( nw->text_out.right_margin != cw->text_out.right_margin )
    nw->text_out.right_margin = cw->text_out.right_margin;
  if( nw->text_out.top_margin != cw->text_out.top_margin )
    nw->text_out.top_margin = cw->text_out.top_margin;
  if( nw->text_out.bottom_margin != cw->text_out.bottom_margin )
    nw->text_out.bottom_margin = cw->text_out.bottom_margin;

  if( nw->text_out.font_set != cw->text_out.font_set ) {
    if( nw->text_out.font_set != NULL )
      XFreeFontSet( d, cw->text_out.font_set );
    n = XFontsOfFontSet( TextFontSet(rw), &fs, &fn );
    nw->text_out.ascent = 0;
    nw->text_out.descent = 0;
    while( --n >= 0 ) {
      if( nw->text_out.ascent < fs[n]->ascent )
	nw->text_out.ascent = fs[n]->ascent;
      if( nw->text_out.descent < fs[n]->descent )
	nw->text_out.descent = fs[n]->descent;
    }
    fse = XExtentsOfFontSet( TextFontSet(nw) );
    nw->text_out.max_extent = fse->max_logical_extent.width;
    cw->core.height =
      nw->core.height = nw->text_out.ascent + nw->text_out.descent;
    MakeCursors( nw );
    changed = True;
  }

  if( hcscmp( nw->text_out.hvalue, cw->text_out.hvalue ) ) {
    XtFree( (char *)TextHcsValue(cw) );
    if( nw->text_out.hvalue == NULL ) {
      nw->text_out.string_length = 0;
      nw->text_out.alloc_size = sizeof(hchar_t) * ALLOC_SIZE;
      TextHcsValue(nw) = (hchar_t *)XtMalloc( nw->text_out.alloc_size );
      TextHcsValue(nw)[0] = '\0';
    }
    else {
      nw->text_out.string_length = hcslen( TextHcsValue(nw) );
      nw->text_out.alloc_size = sizeof(hchar_t) *
	( nw->text_out.string_length + ALLOC_SIZE );
      TextHcsValue(nw) =
	hcscpy( (hchar_t *)XtMalloc( nw->text_out.alloc_size ),
	       TextHcsValue(rw) );
    }

    nw->text_out.highlight.number = 1;
    nw->text_out.highlight.list[0].position = 0;

    HawTextOutSetCursorPosition( new, nw->text_out.string_length );

    changed = True;
    nw->text_out.value = NULL;
  }
  else if( nw->text_out.value != NULL ) {
    ValidateString( nw, nw->text_out.value );
    nw->text_out.highlight.number = 1;
    nw->text_out.highlight.list[0].position = 0;

    HawTextOutSetCursorPosition( new, nw->text_out.string_length );

    changed = True;
    nw->text_out.value = NULL;
  }

  if( nw->text_out.cursor_position != cw->text_out.cursor_position )
    changed = True;

  nw->text_out.columns = nw->core.width / nw->text_out.max_extent;

  return( changed );
}

static void GetValuesHook(Widget w, ArgList args, Cardinal *num_args)
{
  int	n, len;
  char	*mbs;

  for( n = 0; n < *num_args ; n++ ) {
    if( strcmp( args[n].name, XtNvalue ) == 0 ) {
      len = hcslen( TextHcsValue(w) ) * MB_CUR_MAX;
      mbs = XtMalloc(len + 1);
      if( mbs == NULL )
	*((char **)args[n].value) = NULL;
      else {
	len = hcstombs( mbs, TextHcsValue(w), len );
	if( len != -1 )
	  *((char **)args[n].value) = mbs;
	else
	  *((char **)args[n].value) = NULL;
      }
    }
  }
}

static int FindHighlight(HawTextOutWidget w, Position position)
{
  HighlightRec	*l = w->text_out.highlight.list;
  int	ii;

  for( ii = w->text_out.highlight.number - 1; ii >= 0; ii-- )
    if( position >= l[ii].position )
      return( ii );
  return( 0 );
}


static void InsertHighlight( HawTextOutWidget w, Position position,
    HawHighlightMode mode )
{
  HighlightRec	*l = w->text_out.highlight.list;
  int	ii, jj;

  ii = FindHighlight( w, position );

  if( l[ii].position == position ) {
    l[ii].mode = mode;
    if( ii > 0  && l[ii-1].mode == mode ) {
      for( jj = ii, w->text_out.highlight.number--;
	   jj < w->text_out.highlight.number; jj++ )
	l[jj] = l[jj+1];
      ii--;
    }
    if( ii + 1 != w->text_out.highlight.number  &&  l[ii+1].mode == mode ) {
      w->text_out.highlight.number--;
      for( ii++; ii < w->text_out.highlight.number; ii++ )
	l[ii] = l[ii+1];
    }
  }
  else  if( l[ii].mode != mode ) {
    if( ii + 1 != w->text_out.highlight.number  &&  l[ii+1].mode == mode ) {
      l[ii+1].position = position;
    }
    else {
      if( w->text_out.highlight.number == w->text_out.highlight.maximum ) {
	w->text_out.highlight.maximum++;
	l = w->text_out.highlight.list =
	  (HighlightRec *)XtRealloc( (char *)l,
	      sizeof(HighlightRec) * w->text_out.highlight.maximum );
      }
      for( jj = w->text_out.highlight.number; jj > ii; jj-- )
	l[jj] = l[jj-1];
      w->text_out.highlight.number++;
      l[ii+1].position = position;
      l[ii+1].mode = mode;
    }
  }

  for( ii = 0; ii < w->text_out.highlight.number; ii++ ) {
    if( w->text_out.highlight.list[ii].position > w->text_out.string_length ) {
      w->text_out.highlight.number = ii;
      break;
    }
  }
}

static void TextOutAreaSetHighlight( HawTextOutWidget w,
    Position left, Position right, HawHighlightMode mode )
{
  HawHighlightMode	rmode;
  int			l, r;

  if( left >= right )
    return;

  rmode = w->text_out.highlight.list[FindHighlight(w, right)].mode;
  InsertHighlight( w, left, mode );
  l = FindHighlight( w, left );
  r = FindHighlight( w, right - 1 );
  while( l < r ) {
    InsertHighlight( w, w->text_out.highlight.list[r].position, mode );
    r = FindHighlight( w, right );
    if( w->text_out.highlight.list[r].mode == mode )
      r--;
  }
  InsertHighlight( w, right, rmode );
}


static void 
RedisplayText( HawTextOutWidget w, Position start, Position end )
{
  Display	*display = XtDisplay(w);
  hchar_t	*value = w->text_out.hvalue;
  HighlightRec	*list = w->text_out.highlight.list;
  int		x, y, width;
  int		ii;
  Position	_end;
  GC		gc;

  if( !XtIsRealized((Widget)w) )
    return;

  if( w->text_out.cursor_on )
    PaintCursor( w );
  if( w->text_out.caret_on )
    PaintCaret( w );

  XSelectInput( display, w->core.window,
		w->text_out.event_mask & ~ExposureMask );

  width = HawTextEscapement( TextFontSet(w), value,
			     w->text_out.string_length )
    + w->text_out.cursor_width / 2;
  if( width < w->core.width )
    XFillRectangle( display, XtWindow(w), w->text_out.inv_gc, width, 0,
		    w->core.width, w->core.height );

  ii = 0;
  while( start != end  &&  ii < w->text_out.highlight.number ) {
    if( list[ii].position <= start  &&
	(w->text_out.highlight.number == ii + 1  ||
	 start < list[ii+1].position) ) {
      x = HawTextEscapement( TextFontSet(w), value, start )
	+ w->text_out.left_margin;
      y = w->text_out.ascent + w->text_out.top_margin;
      if( w->text_out.attachment == HawCURSOR_LEFT )
	x += (w->text_out.cursor_width >> 1)
	  - HawTextEscapement( TextFontSet(w), value,
			       w->text_out.left_cursor_position );
      else
	x += w->core.width - (w->text_out.cursor_width >> 1)
	  - HawTextEscapement( TextFontSet(w), value,
			       w->text_out.right_cursor_position );

      _end = end;
      if( ii+1 < w->text_out.highlight.number  &&
	  _end > list[ii+1].position )
	_end = list[ii+1].position;

      switch( list[ii].mode ) {
      case HawHIGHLIGHT_NORMAL:
      case HawHIGHLIGHT_SECONDARY_SELECTED:
	gc = w->text_out.gc;
	break;

      case HawHIGHLIGHT_SELECTED:
	gc = w->text_out.inv_gc;
	break;

      default:
	gc = w->text_out.gc;
	break;
      }

      HawDrawImageString( display, XtWindow(w), TextFontSet(w),
			  gc, x, y, value + start, _end - start );

      if( list[ii].mode == HawHIGHLIGHT_SECONDARY_SELECTED )
	XDrawLine( display, XtWindow(w), gc,
		   x, y + w->text_out.descent - 1,
		   x + HawTextEscapement( TextFontSet(w), &value[start],
					  _end - start ),
		   y + w->text_out.descent - 1 );

      start = _end;
    }
    else
      ii++;
  }

  if( w->text_out.enter )
    PaintCursor( w );
  else
    PaintCaret( w );
  XSelectInput( display, XtWindow(w), w->text_out.event_mask );
}


static void SetCursorPosition( HawTextOutWidget w, Position position )
{
  int		n;
  Boolean	redraw = False;
  hchar_t	*value = w->text_out.hvalue;
  Dimension	width;

  if( position < 0 )
    position = 0;

  if( position > w->text_out.string_length )
    position = w->text_out.string_length;

  if( HawTextEscapement( TextFontSet(w), value, w->text_out.string_length ) <
      w->core.width ) {
    w->text_out.left_cursor_position = 0;
    w->text_out.right_cursor_position = w->text_out.string_length;
    w->text_out.attachment = HawCURSOR_LEFT;

    XFillRectangle( XtDisplay(w), XtWindow(w), w->text_out.inv_gc,
		    0, 0, w->text_out.cursor_width / 2, w->core.height );
  }
  else if( position < w->text_out.left_cursor_position ) {
    width = w->core.width;
    for( n = position + 1;
	 width > HawTextEscapement( TextFontSet(w), &value[position],
				    n - position ); n++ );
    w->text_out.right_cursor_position = n - 1;
    w->text_out.left_cursor_position = position;
    w->text_out.attachment = HawCURSOR_LEFT;
    redraw = True;
  }
  else if( position > w->text_out.right_cursor_position ) {
    width = (w->text_out.cursor_width >> 1)
      + HawTextEscapement( TextFontSet(w), value, position )
      - w->core.width;
    for( n = 1; HawTextEscapement( TextFontSet(w), value, n ) < width;
	 n++ );
    w->text_out.left_cursor_position = n;
    w->text_out.right_cursor_position = position;
    w->text_out.attachment = HawCURSOR_RIGHT;
    redraw = True;
  }
  else if( w->text_out.attachment == HawCURSOR_RIGHT  &&
	   w->text_out.string_length < w->text_out.right_cursor_position ) {
    width = (w->text_out.cursor_width >> 1)
      + HawTextEscapement( TextFontSet(w), value,
			   w->text_out.string_length )
      - w->core.width;
    for( n = 1; HawTextEscapement( TextFontSet(w), value, n ) < width;
	 n++ );
    w->text_out.left_cursor_position = n;
    w->text_out.right_cursor_position = w->text_out.string_length;
    redraw = True;
  }

  if( w->text_out.cursor_on )
    PaintCursor( w );
  if( w->text_out.caret_on )
    PaintCaret( w );

  TextCursorPosition(w) = position;

  if( redraw )
    XClearArea( XtDisplay(w), XtWindow(w), 0, 0, 0, 0, True );

  if( w->text_out.enter )
    PaintCursor( w );
  else
    PaintCaret( w );
}


static Boolean
ReplaceText( HawTextOutWidget w, XEvent *event, Position replace_prev,
    Position replace_next, hchar_t *insert, int insert_length )
{
  int			replace_length, ii, jj;
  hchar_t		*src, *dst;
  int			delta = 0;
  Position		cursorPos;
  hchar_t		*orig_insert;
  HawHighlightMode	highlight_mode;
  HighlightRec		*l;
  int			number;

  highlight_mode = w->text_out.highlight.list[FindHighlight( w, replace_prev )].mode;
  orig_insert = insert;
  replace_length = replace_next - replace_prev;

  if( (w->text_out.string_length + insert_length - replace_length) - 
      w->text_out.max_length > 0) { 
    if( w->text_out.verify_bell )
      XBell( XtDisplay( (Widget)w), 0 );
    if( insert != orig_insert )
      XtFree( (char *)insert );
    return( False );
  }

  if( w->text_out.string_length + insert_length - replace_length >=
      (w->text_out.alloc_size / sizeof(hchar_t)) ) {
    w->text_out.alloc_size += max( insert_length + 1, ALLOC_SIZE );
    TextHcsValue(w) = (hchar_t *)XtRealloc( (char *)TextHcsValue(w),
					    w->text_out.alloc_size );
  }

  delta = insert_length - replace_length;

  if( replace_length > insert_length ) {
    for( src = TextHcsValue(w) + replace_next, dst = src + delta,
	   ii = (w->text_out.string_length + 1) - replace_next;
	 ii > 0; ++src, ++dst, --ii )
      *dst = *src;
    l = w->text_out.highlight.list;
    for( ii = w->text_out.highlight.number - 1; ii >= 0; ii-- )
      if( replace_next > l[ii].position )
	break;
    jj = ii;
    jj++;
    while( jj < w->text_out.highlight.number ) {
      l[jj].position += delta;
      jj++;
    }
    if( (ii + 1) < w->text_out.highlight.number ) {
      number = 0;
      while( l[ii].position >= l[ii+number+1].position  &&
	     (ii+number+1) < w->text_out.highlight.number )
	number++;
      if( number > 0 ) {
	jj = ii;
	while( jj < w->text_out.highlight.number ) {
	  l[jj] = l[jj+number];
	  jj++;
	}
	w->text_out.highlight.number -= number;
	if( ii > 0  &&  l[ii].mode == l[ii-1].mode ) {
	  while( ii < w->text_out.highlight.number ) {
	    l[ii] = l[ii+1];
	    ii++;
	  }
	  w->text_out.highlight.number--;
	}
      }
    }
  }
  else if( replace_length < insert_length ) {
    for( src = TextHcsValue(w) + w->text_out.string_length,
	   dst = src + delta,
	   ii = (w->text_out.string_length + 1) - replace_next;
	 ii > 0; src--, dst--, ii-- )
      *dst = *src;
    l = w->text_out.highlight.list;
    for( ii = w->text_out.highlight.number - 1; ii > 0; ii-- )
      if( replace_next > l[ii].position )
	break;
    ii++;
    while( ii < w->text_out.highlight.number ) {
      l[ii].position += delta;
      ii++;
    }
    TextOutAreaSetHighlight( w, replace_prev, replace_prev + insert_length,
			     highlight_mode );
  }

  if( insert_length != 0 ) {
    for( src = insert, dst = TextHcsValue(w) + replace_prev,
	   ii = insert_length;
	 ii > 0; ++src, ++dst, --ii )
      *dst = *src;
  }

  w->text_out.string_length += delta;
  cursorPos = replace_next + delta;
  SetCursorPosition( w, cursorPos );

  if( replace_length == insert_length )
    replace_next = replace_prev + replace_length;
  else
    replace_next = w->text_out.string_length;
  RedisplayText( w, 0, replace_next );

  if( insert != orig_insert )
    XtFree( (char *)insert );
  return( True );
}


static void
Initialize( Widget request, Widget new, ArgList args, Cardinal *num_args )
{
  HawTextOutWidget	rw = (HawTextOutWidget)request;
  HawTextOutWidget	nw = (HawTextOutWidget)new;
  XFontStruct		**fs;
  char			**fn;
  int			n;
  hchar_t		*hcs;
  XFontSetExtents	*fse;

  if( TextCursorPosition(rw) < 0 )
    TextCursorPosition(nw) = 0;

  n = XFontsOfFontSet( TextFontSet(nw), &fs, &fn );
  nw->text_out.ascent = 0;
  nw->text_out.descent = 0;
  while( --n >= 0 ) {
    if( nw->text_out.ascent < fs[n]->ascent )
      nw->text_out.ascent = fs[n]->ascent;
    if( nw->text_out.descent < fs[n]->descent )
      nw->text_out.descent = fs[n]->descent;
  }
  if( nw->core.height < nw->text_out.ascent + nw->text_out.descent )
    nw->core.height = nw->text_out.ascent + nw->text_out.descent
      + nw->text_out.top_margin + nw->text_out.bottom_margin;

  fse = XExtentsOfFontSet( TextFontSet(nw) );
  nw->text_out.max_extent = fse->max_logical_extent.width;
  if( nw->core.width <= 0 ) {
    if( nw->text_out.columns > 0 )
      nw->core.width = nw->text_out.max_extent * nw->text_out.columns
	+ nw->text_out.left_margin
	+ nw->text_out.right_margin;
    else {
      nw->core.width = nw->core.parent->core.width;
      nw->text_out.columns = nw->core.width / nw->text_out.max_extent;
      if( nw->text_out.columns == 0 ) {
	nw->core.width = nw->text_out.max_extent * 20
	  + nw->text_out.left_margin
	  + nw->text_out.right_margin;
	nw->text_out.columns = 20;
      }
    }
  }
  nw->text_out.old_width = nw->core.width;
  nw->text_out.old_height = nw->core.height;

  nw->text_out.cursor_on = nw->text_out.caret_on = False;

  nw->text_out.gc = NULL;
  nw->text_out.inv_gc = NULL;
  nw->text_out.xor_gc = NULL;

  if( TextHcsValue(nw) == NULL ) {
    if( TextValue(nw) == NULL ) {
      nw->text_out.string_length = 0;
      nw->text_out.alloc_size = sizeof(hchar_t) * ALLOC_SIZE;
      TextHcsValue(nw) = (hchar_t *)XtMalloc( nw->text_out.alloc_size );
      TextHcsValue(nw)[0] = '\0';
    }
    else
      ValidateString( nw, TextValue(nw) );
  }
  else {
    hcs = TextHcsValue(nw);
    nw->text_out.string_length = hcslen( hcs );
    nw->text_out.alloc_size = sizeof(hchar_t) *
      ( nw->text_out.string_length + ALLOC_SIZE );
    TextHcsValue(nw) = hcscpy( (hchar_t *)XtMalloc( nw->text_out.alloc_size ),
			      hcs );
    TextValue(nw) = NULL;
  }

  if( TextMaxLength(nw) < MAXINT ) {
    if( nw->text_out.string_length > TextMaxLength(nw) ) {
      nw->text_out.string_length = TextMaxLength(nw);
      TextHcsValue(nw)[nw->text_out.string_length] = '\0';
    }
  }

  nw->text_out.cursor_height = nw->text_out.cursor_width = 0;
  nw->text_out.cursor = None;

  nw->text_out.highlight.number = nw->text_out.highlight.maximum = 1;
  nw->text_out.highlight.list
    = (HighlightRec *)XtMalloc( (unsigned)sizeof(HighlightRec) );
  nw->text_out.highlight.list[0].position = 0;
  nw->text_out.highlight.list[0].mode = HawHIGHLIGHT_SECONDARY_SELECTED;

  nw->text_out.timer_id = '\0';

  nw->text_out.cursor_width = 5;

  TextCursorPosition(nw) = 0;
}


static void
Realize( Widget widget, Mask *vmask, XSetWindowAttributes *attr )
{
  HawTextOutWidget		pw = (HawTextOutWidget)widget;

  pw->text_out.event_mask = attr->event_mask;

  attr->border_pixel = pw->text_out.foreground;
  *vmask |= CWBitGravity;
  attr->bit_gravity = NorthWestGravity;
  XtCreateWindow( widget, (unsigned)InputOutput, (Visual *)CopyFromParent,
		  *vmask, attr );

  LoadGCs( pw, pw->core.background_pixel, pw->text_out.foreground );
  MakeCursors( pw );
}

static void Resize(Widget w)
{
  HawTextOutWidget tw = (HawTextOutWidget)w;

  if( tw->text_out.old_width < tw->core.width  &&
      tw->text_out.right_margin > 0 )
    XClearArea( XtDisplay(w), XtWindow(w),
	tw->core.width - tw->text_out.right_margin, 0, 0, 0,
	False );
  if( tw->text_out.old_height < tw->core.height  &&
      tw->text_out.bottom_margin > 0 )
    XClearArea( XtDisplay(w), XtWindow(w),
	0, tw->core.height - tw->text_out.bottom_margin, 0, 0,
	False );

  tw->text_out.old_width = tw->core.width;
  tw->text_out.old_height = tw->core.height;

  tw->text_out.columns = tw->core.width / tw->text_out.max_extent;

  RedisplayText( (HawTextOutWidget) w, TextCursorPosition(tw),
		 tw->text_out.string_length );
}


static void Redisplay( Widget w, XEvent *event, Region region )
{
  HawTextOutWidget pw = (HawTextOutWidget)w;

  if( !pw->core.sensitive )
    return;
  RedisplayText( pw, 0, pw->text_out.string_length );
}


static void
Enter( Widget w, XEvent *event, String *params, Cardinal *num_params )
{
  HawTextOutWidget tw = (HawTextOutWidget)w;

  tw->text_out.enter = True;
}

static void
Leave( Widget w, XEvent *event, String *params, Cardinal *num_params )
{
  HawTextOutWidget	tw = (HawTextOutWidget)w;

  tw->text_out.enter = False;
}

static XtActionsRec	actionsList[] = {
  { "Enter",		Enter	},
  { "Leave",		Leave	}};

#define Offset(field)	XtOffsetOf(HawTextOutRec, field)
static XtResource	local_resources[] =
{
  {
    XtNvalue, XtCValue,
    XtRString, sizeof(String), Offset(text_out.value),
    XtRString, NULL },
  {
    HawNhcsValue, HawCHcsValue,
    HawRHcsValue, sizeof(hchar_t *), Offset(text_out.hvalue),
    XtRString, NULL },
  {
    HawNcursorPosition, HawCCursorPosition,
    XtRPosition, sizeof(Position), Offset(text_out.cursor_position),
    XtRImmediate, (XtPointer)0 },
  {
    XtNfontSet, XtCFontSet,
    XtRFontSet, sizeof(XFontSet), Offset(text_out.font_set),
    XtRString, XtDefaultFontSet },
  {
    XtNforeground, XtCForeground,
    XtRPixel, sizeof(Pixel), Offset(text_out.foreground),
    XtRString, XtDefaultForeground },
  {
    HawNcolumns, HawCColumns,
    XtRShort, sizeof(short), Offset(text_out.columns),
    XtRImmediate, (XtPointer)0 },
  {
    HawNmaxLength, HawCMaxLength,
    XtRInt, sizeof(int), Offset(text_out.max_length),
    XtRImmediate, (XtPointer)MAXINT },
  {
    HawNverifyBell, HawCVerifyBell,
    XtRBoolean, sizeof(Boolean), Offset(text_out.verify_bell),
    XtRImmediate, (XtPointer)True },
  {
    HawNbottomMargin, XtCMargin,
    XtRPosition, sizeof(Position), Offset(text_out.bottom_margin),
    XtRImmediate, (XtPointer)2 },
  {
    HawNtopMargin, XtCMargin,
    XtRPosition, sizeof(Position), Offset(text_out.top_margin),
    XtRImmediate, (XtPointer)2 },
  {
    HawNleftMargin, XtCMargin,
    XtRPosition, sizeof(Position), Offset(text_out.left_margin),
    XtRImmediate, (XtPointer)2 },
  {
    HawNrightMargin, XtCMargin,
    XtRPosition, sizeof(Position), Offset(text_out.right_margin),
    XtRImmediate, (XtPointer)2 }
};
#undef Offset

static char defaultTranslations[] = "\
	<EnterWindow>:	Enter()\n\
	<LeaveWindow>:	Leave()";

HawTextOutClassRec	hawTextOutClassRec = 
{
  { /* core_class record     */
    /* superclass	     */ (WidgetClass) &widgetClassRec,
    /* class_name	     */ "HawTextOut",
    /* widget_size	     */ sizeof(HawTextOutRec),
    /* class_initialize	     */ ClassInitialize,
    /* class_part_initialize */ NULL,
    /* class_inited          */ False,
    /* initialize	     */ Initialize,
    /* initialize_hook       */ NULL,
    /* realize		     */ Realize,
    /* actions               */ actionsList,
    /* num_actions    	     */ XtNumber(actionsList),
    /* resources	     */ local_resources,
    /* num_resources         */ XtNumber(local_resources),
    /* xrm_class	     */ NULLQUARK,
    /* compress_motion       */ True,
    /* compress_exposure     */ XtExposeCompressMaximal,
    /* compress_enterleave   */ True,
    /* visible_interest      */ False,
    /* destroy               */ Destroy,
    /* resize                */ Resize,
    /* expose                */ Redisplay,
    /* set_values	     */ SetValues,
    /* set_values_hook       */ NULL,
    /* set_values_almost     */ XtInheritSetValuesAlmost,
    /* get_values_hook       */ GetValuesHook,
    /* accept_focus	     */ NULL,
    /* version               */ XtVersion,
    /* callback private      */ NULL,
    /* tm_table              */ defaultTranslations,
    /* query_geometry        */ XtInheritQueryGeometry,
    /* display_accelerator   */ NULL,
    /* extension             */ NULL
  },
  { /* simple_class record   */
    /* cahnge_sensitive      */	XtInheritChangeSensitive
  },
  { /* text_out_class record */
    /* extension	     */ NULL
  },

};

WidgetClass	hawTextOutClass = (WidgetClass)&hawTextOutClassRec;

Widget HawCreateTextOut( Widget parent, char *name,
    ArgList arglist, Cardinal argcount )
{
  return( XtCreateManagedWidget( name, hawTextOutClass,
      parent, arglist, argcount ) );
}

void
HawTextOutSetHighlight( Widget w, Position left, Position right,
    HawHighlightMode mode )
{
  HawTextOutWidget	pw = (HawTextOutWidget)w;

  if( left >= right || right <= 0 )	return;

  if( left < 0 )	left = 0;

  if( right > pw->text_out.string_length )
    right = pw->text_out.string_length;

  TextOutAreaSetHighlight( pw, left, right, mode );

  RedisplayText( pw, left, right );
}

Boolean	
HawTextOutReplace( Widget w, Position from_pos, Position to_pos, char *value )
{
  HawTextOutWidget	pw = (HawTextOutWidget)w;
  Boolean		result;
  int			len;
  hchar_t		*hcs;

  if( from_pos < 0 )
    from_pos = 0;

  if( to_pos >= pw->text_out.string_length )
    to_pos = pw->text_out.string_length;

  if( value == NULL )
    value = "";
  len = strlen( value );
  hcs = (hchar_t *)XtMalloc( sizeof(hchar_t) * ( len + 1 ) );
  len = mbstohcs( hcs, value, len );

  result = ReplaceText( pw, NULL, from_pos, to_pos, hcs, len );
  XtFree( (char *)hcs );
  return( result );
}

Boolean
HawTextOutReplaceHcs( Widget w, Position from_pos, Position to_pos,
    hchar_t *value )
{
  HawTextOutWidget	pw = (HawTextOutWidget)w;
  Boolean		result;
  hchar_t		*hcs;

  if( from_pos < 0 )
    from_pos = 0;

  if( to_pos >= pw->text_out.string_length )
    to_pos = pw->text_out.string_length;

  if( value == NULL ) {
    hcs = (hchar_t *)XtMalloc( sizeof(hchar_t) );
    hcs[0] = '\0';
  }
  else
    hcs = value;

  result = ReplaceText( pw, NULL, from_pos, to_pos, hcs, hcslen(hcs) );
  if( value == NULL )
    XtFree( (char *)hcs );
  return( result );
}

void
HawTextOutInsert( Widget w, Position position, char *value )
{
  HawTextOutReplace( w, position, position, value );
}

void
HawTextOutInsertHcs( Widget w, Position position, hchar_t *value )
{
  HawTextOutReplaceHcs( w, position, position, value );
}

void
HawTextOutSetString( Widget w, char *value, HawHighlightMode mode )
{
  HawTextOutWidget	pw = (HawTextOutWidget)w;

  if( value == NULL )
    value = "";
  ValidateString( pw, value );
  pw->text_out.highlight.number = 1;
  pw->text_out.highlight.list[0].position = 0;
  pw->text_out.highlight.list[0].mode = mode;

  HawTextOutSetCursorPosition( (Widget)pw, pw->text_out.string_length );
  RedisplayText( pw, 0, pw->text_out.string_length );
}

void
HawTextOutSetStringHcs( Widget w, hchar_t *value, HawHighlightMode mode )
{
  HawTextOutWidget	pw = (HawTextOutWidget)w;

  XtFree( (char *)TextHcsValue(pw) );
  if( value == NULL ) {
    pw->text_out.string_length = 0;
    pw->text_out.alloc_size = sizeof(hchar_t) * ALLOC_SIZE;
    TextHcsValue(pw) = (hchar_t *)XtMalloc( pw->text_out.alloc_size );
    TextHcsValue(pw)[0] = '\0';
  }
  else {
    pw->text_out.string_length = hcslen( value );
    pw->text_out.alloc_size = sizeof(hchar_t) *
      ( pw->text_out.string_length + ALLOC_SIZE );
    TextHcsValue(pw) =
      hcscpy( (hchar_t *)XtMalloc( pw->text_out.alloc_size ), value );
  }

  pw->text_out.highlight.number = 1;
  pw->text_out.highlight.list[0].position = 0;
  pw->text_out.highlight.list[0].mode = mode;

  HawTextOutSetCursorPosition( w, pw->text_out.string_length );
  RedisplayText( pw, 0, pw->text_out.string_length );
}

void HawTextOutSetCursorPosition( Widget w, Position position )
{
  SetCursorPosition( (HawTextOutWidget)w, position );
}

char *HawTextOutGetString( Widget w )
{
  int	len = hcslen(TextHcsValue(w)) * MB_CUR_MAX;
  char	*mbs = XtMalloc(len + 1);

  if( mbs == NULL )
    return( NULL );
  len = hcstombs( mbs, TextHcsValue(w), len );
  return( (len != -1) ? mbs : NULL );
}

hchar_t	*HawTextOutGetStringHcs( Widget w )
{
  int		len = hcslen(TextHcsValue(w));
  hchar_t	*hcs = (hchar_t *)XtMalloc((len + 1) * MB_CUR_MAX);

  if( hcs == NULL )
    return( NULL );
  hcsncpy( hcs, TextHcsValue(w), len );
  hcs[len] = '\0';
  return( hcs );
}

Position HawTextOutGetLastPosition( Widget w )
{
  return( hcslen(TextHcsValue(w)) );
}
