/* $Id: XuiCvtr.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include	<stdlib.h>
#include	<ctype.h>
#include	<X11/keysymdef.h>
#include	<Haw/Haw.h>
#include	<Haw/HawStrDefs.h>
#include	"hanim.h"
#include	"Ximp.h"
#include	"resource.h"

static Boolean StringsAreEqual(char *s1, char *s2)
{
    register char c;

    if (!*s1 && *s2)
        return False;
    while ((c = *s1) != '\0') {
        if (isupper(c))
            c = tolower(c);
        if (c != *s2)
            return(False);
        s1++;
        s2++;
    }
    return(True);
}

static void CvtStringToLong(XrmValue *args,
    Cardinal *num_args, XrmValue *from_val, XrmValue *to_val)
{
    static long l;

    if (*num_args != 0)
        XtWarningMsg("wrongParameters", "cvtStringToLong", "XtToolkitError",
	    "String to Long conversion needs no extra arguments",
	    (String *) NULL, (Cardinal *)NULL);
    l = strtoul((char *)from_val->addr, (char **)NULL, 0);

    to_val->size = sizeof(l);
    to_val->addr = (XtPointer) & l;
}

static void CvtStringToCharacterCode(XrmValue *args,
    Cardinal *num_args, XrmValue *from_val, XrmValue *to_val)
{
    static CharacterCode code;
    char *in_str = (char *)(from_val->addr);

    to_val->size = sizeof(CharacterCode);
    to_val->addr = (XtPointer) & code;

    if (StringsAreEqual(in_str, "kssm"))
        code = _KSSM;
    else if (StringsAreEqual(in_str, "euckr"))
        code = _EUCKR;
    else if (StringsAreEqual(in_str, "ksc5601"))
        code = _KSC5601;
    else if (StringsAreEqual(in_str, "ucs2"))
        code = _UNICODE;
    else {
        code = _EUCKR;
        XtStringConversionWarning((char *)from_val->addr, HawRCharacterCode);
    }
}

static void CvtStringToInputLanguage(XrmValue *args,
    Cardinal *num_args, XrmValue *from_val, XrmValue *to_val)
{
    static InputLanguage lang;
    char	*in_str = (char *)(from_val->addr);

    to_val->size = sizeof(InputLanguage);
    to_val->addr = (XtPointer) & lang;

    if (StringsAreEqual(in_str, "hangul")
         || StringsAreEqual(in_str, "korean"))
        lang = HANGUL;
    else if (StringsAreEqual(in_str, "english"))
        lang = ENGLISH;
    else if (StringsAreEqual(in_str, "japanese"))
        lang = JAPANESE;
    else if (StringsAreEqual(in_str, "cyrillic")
              || StringsAreEqual(in_str, "russia"))
        lang = CYRILLIC;
    else {
        lang = 0;
    }
}

static void CvtStringToLangKeyboard(XrmValue *args,
    Cardinal *num_args, XrmValue *from_val, XrmValue *to_val)
{
    static LangKeyboard kbd;
    char *in_str = (char *)(from_val->addr);

    to_val->size = sizeof(LangKeyboard);
    to_val->addr = (XtPointer) & kbd;

    if (StringsAreEqual(in_str, "3")
	|| StringsAreEqual(in_str, "3bulsik"))
        kbd.hangul = HANGUL3;
    else if (StringsAreEqual(in_str, "2")
	|| StringsAreEqual(in_str, "2bulsik"))
        kbd.hangul = HANGUL2;
    else if (StringsAreEqual(in_str, "qwerty"))
        kbd.english = QWERTY;
    else if (StringsAreEqual(in_str, "dvorak"))
        kbd.english = DVORAK;
#if 0 /* LATER */
    else if (StringsAreEqual(in_str, "enghiragana"))
        kbd.japanese = HIRAGANA_ROMAN;
    else if (StringsAreEqual(in_str, "engkatakana"))
        kbd.japanese = KATAKANA_ROMAN;
    else if (StringsAreEqual(in_str, "hiragana"))
        kbd.japanese = HIRAGANA;
    else if (StringsAreEqual(in_str, "katakana"))
        kbd.japanese = KATAKANA;
    else if (StringsAreEqual(in_str, "hanroman"))
        kbd.hangul = HAN_ROMAN;
    else if (StringsAreEqual(in_str, "hanhiragana"))
        kbd.japanese = HIRAGANA_HANGUL;
    else if (StringsAreEqual(in_str, "hankatakana"))
        kbd.japanese = KATAKANA_HANGUL;
#endif    /* LATER */
    else {
        switch (resources.lang) {
        case ENGLISH:
            kbd.english = QWERTY; break;
#if 0 /* LATER */
        case JAPANESE:
            kbd.japanese = HIRAGANA_ROMAN; break;
#endif    /* LATER */
        case HANGUL:
        default:
            kbd.hangul = HANGUL3; break;
        }
        XtStringConversionWarning((char *)from_val->addr, HawRLangKeyboard);
    }
}

#define	AllMask	(ControlMask | LockMask | ShiftMask | Mod1Mask)

static void CvtStringToProcessStartKeys(XrmValue *args,
    Cardinal *num_args, XrmValue *from_val, XrmValue *to_val)
{
    static XimpKey *keys = NULL, key;
    char	*in_str = (char *)(from_val->addr);
    char	event[256], tmp[256];
    char	*modifier, *detail;
    char	c, *ss;
    int	ii, count;
    Boolean	exclamation, tilde;
    static struct {
        char	*name;
        int	len;
        long	mask;
    } mod[] = {
        { "Ctrl", 4, ControlMask	},
        { "Lock", 4, LockMask	},
        { "Caps", 4, LockMask	},
        { "Shift", 5, ShiftMask	},
        { "Alt", 3, Mod1Mask	},
        { "Meta", 4, Mod1Mask	},
        { NULL, 0, 0	}};

    to_val->size = sizeof(XimpKey *);
    count = 0;

    keys = (XimpKey *)XtMalloc(sizeof(XimpKey));
    do {
        ss = event;
        while ((c = *in_str) != '\0') {
            in_str++;
            if (c == '\n')
                break;
            *ss++ = c;
        }
        *ss = '\0';
        strcpy(tmp, event);

        if ((ss = strstr(event, "<Key>")) != NULL)
            detail = ss + 5;
        else if ((ss = strstr(event, "<KeyDown>")) != NULL)
            detail = ss + 9;
        else {
            XtStringConversionWarning(tmp, "beginConversionKey");
            continue;
        }
        *ss = '\0';

        modifier = event;
        key.modifier = 0;
        key.modifier_mask = 0;
        key.keysym = XK_VoidSymbol;
        exclamation = False;
        do {
            while (*modifier == ' ' || *modifier == '\t')
                modifier++;
            if (*modifier == '\0')
                break;
            if (strncmp(modifier, "None", 4) == 0) {
                if (key.modifier_mask != 0 || exclamation) {
                    XtStringConversionWarning(tmp, "beginConversionKey");
                    goto error;
                }
                key.modifier_mask = AllMask;
                modifier += 4;
            } else {
                if (*modifier == '!') {
                    if (key.modifier_mask != 0 || exclamation) {
                        XtStringConversionWarning(tmp, "beginConversionKey");
                        goto error;
                    }
                    key.modifier_mask = AllMask;
                    exclamation = True;
                    modifier++;
                    continue;
                }
                if ((tilde = (*modifier == '~'))) {
                    modifier++;
                    while (*modifier == ' ' || *modifier == '\t')
                        modifier++;
                }
                for (ii = 0; mod[ii].name != NULL; ii++) {
                    if (strncmp(modifier, mod[ii].name, mod[ii].len) == 0) {
                        key.modifier_mask |= mod[ii].mask;
                        if (!tilde)
                            key.modifier |= mod[ii].mask;
                        modifier += mod[ii].len;
                        break;
                    }
                }
                if (mod[ii].name == NULL) {
                    XtStringConversionWarning(tmp, "beginConversionKey");
                    goto error;
                }
            }
        } while (*modifier != '\0');

        while (*detail == ' ' || *detail == '\t')
            detail++;
        for (ss = detail; isalnum(*ss) || *ss == '_'; ss++);
        *ss = '\0';
        if ((key.keysym = XStringToKeysym(detail)) == NoSymbol) {
            key.keysym = XK_VoidSymbol;
            XtStringConversionWarning(tmp, "beginConversionKey");
        } else {
            keys = (XimpKey *)XtRealloc((char *)keys,
		sizeof(XimpKey) * (count + 2));
            keys[count] = key;
            count++;
            keys[count].modifier = 0;
            keys[count].modifier_mask = 0;
            keys[count].keysym = XK_VoidSymbol;
        }
    error: ;
    } while (*in_str != '\0');
    to_val->addr = (XtPointer) & keys;
}

void AddConverter(XtAppContext app)
{
    XtAppAddConverter(app, XtRString, HawRCharacterCode,
	CvtStringToCharacterCode, NULL, 0);
    XtAppAddConverter(app, XtRString, HawRInputLanguage,
	CvtStringToInputLanguage, NULL, 0);
    XtAppAddConverter(app, XtRString, HawRLangKeyboard,
	CvtStringToLangKeyboard, NULL, 0);
    XtAppAddConverter(app, XtRString, HawRProcessStartKeys,
	CvtStringToProcessStartKeys, NULL, 0);
    XtAppAddConverter(app, XtRString, HawRFlag,
	CvtStringToLong, NULL, 0);
}
