/* $Id: XuiDialog.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include	<stdio.h>
#include	<stdlib.h>
#include	<assert.h>

#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Form.h>
#include	<Haw/Label.h>
#include	<Haw/Command.h>

#include 	<X11/Xatom.h>
#define		XK_MISCELLANY
#include	<X11/keysymdef.h>

#include	"hanim.h"
#include	"resource.h"

extern	Pixel light_color( Colormap cm, Pixel pixel );

//   
static Atom	wm_delete_window = None;

typedef struct _wmDelInfo {
    /* Destroy request info */
    void	*data;
    XtCallbackProc	func;
    /* next link pointer */
    struct _wmDelInfo	*next;
} WMDelInfo;

static WMDelInfo	*wmDelHead = NULL;

/*  Լ widget destroyɶ ȴ. */
static void WMDelCallback(Widget w, XtPointer client, XtPointer dummy)
{
    WMDelInfo *wmDel = (WMDelInfo *)client;
    WMDelInfo *node, **pp;

    for ( node = *(pp = &wmDelHead); node != NULL && node != wmDel;
          node = *(pp = &node->next) ) ;
    if ( node == NULL)
        return;

    /* delte current node int list */
    *pp = node->next;
    XFree(client);
}

static void WMDelHandler(Widget w, XtPointer client,
                         XEvent * event, Boolean *dummy)
{
    WMDelInfo	*wmDel = (WMDelInfo *)client;

    /* run registered callback function */
    if ( w ) {
        if ( event->type == ClientMessage &&
             event->xclient.data.l[0] == wm_delete_window) {
            wmDel->func(w, wmDel->data, NULL);
        }
    }
}

void AddWMDelCallback(Widget w, XtCallbackProc func, void *data)
{
    WMDelInfo	*wmDel;

    if ( wm_delete_window == None )
        wm_delete_window = XInternAtom(XtDisplay(w),
	    "WM_DELETE_WINDOW", FALSE);
    if ( wm_delete_window == None )
        return;

    /* Set up the shell destroy info */
    wmDel	= XtNew(WMDelInfo);
    wmDel->data = data;
    wmDel->func = func;

    XtAddRawEventHandler(w, 0, True, WMDelHandler, wmDel);
    XtAddCallback(w, XtNdestroyCallback, WMDelCallback, (XtPointer)wmDel);
    XSetWMProtocols(XtDisplay(w), XtWindow(w), &wm_delete_window, 1);

    wmDel->next = wmDelHead;
    wmDelHead = wmDel;
}

void CloseDialogCB(Widget w, XtPointer client, XtPointer call)
{
    Widget shell = w;

    while(!XtIsShell(shell))
	shell = XtParent(shell);
    if (!shell)
        return;
    XtPopdown(shell);
    XtDestroyWidget(shell);
}

void CloseDialogHandler(Widget w, XtPointer client,
    XEvent * event, Boolean *dummy)
{
    if (event->type == KeyPress &&
	XK_Escape == XLookupKeysym((XKeyEvent *)event, 0))
        CloseDialogCB(w, client, NULL);
}

void Message(char *message, int w, int h)
{
    Widget	dshell, dform, dlabel, dclose;
    Screen	*screen;
    Pixel	bg, lbg;

    XtVaGetValues( TopLevel, XtNbackground, &bg,
	XtNscreen, &screen, NULL );
    lbg = light_color( None, bg );

    dshell = XtVaCreatePopupShell("Message",
	topLevelShellWidgetClass, TopLevel,
	XtNmappedWhenManaged, False,
	XtNwidth, w, XtNheight, h,
	XtNminWidth, w, XtNminHeight, h,
	XtNmaxWidth, w, XtNmaxHeight, h,
	NULL );
    dform = XtVaCreateManagedWidget("DForm", formWidgetClass, dshell,
	NULL);
    dlabel = XtVaCreateManagedWidget("DLabel",
	hLabelWidgetClass, dform,
	XtNlabel, message,
	XtNborderWidth, 0,
	XtNvertDistance, 10,
	XtNhorizDistance, 20,
	XtNtop, XawChainTop,
	XtNbottom, XawChainTop,
	NULL);
    dclose = XtVaCreateManagedWidget("Close",
	hCommandWidgetClass, dform,
	XtNbackground, lbg,
	XtNfromVert, dlabel,
	XtNvertDistance, 10,
	XtNhorizDistance, 150,
	XtNleft, XawChainLeft,
	XtNright, XawChainLeft,
	NULL);
    if ( resources.close_label[0] )
        XtVaSetValues(dclose, XtNlabel, resources.close_label, NULL);
    XtVaSetValues(dclose, XtNwidth, 100, NULL);
    XtManageChild(dshell);

    XtAddCallback(dclose, XtNcallback, CloseDialogCB, dshell);
    XtAddEventHandler(dform, KeyPressMask, True,
	CloseDialogHandler, NULL);
    AddWMDelCallback(dshell, CloseDialogCB, NULL);

    /* draw dialog in screnn center */
	XtVaSetValues(dshell, XtNx, (WidthOfScreen(screen) - w) / 2,
			XtNy, (HeightOfScreen(screen) - h ) / 2, NULL);

    XtPopup(dshell, XtGrabExclusive);
}

Boolean yes_no_answer = False;
Boolean yes_no_done = False;

static void SetBooleanAnswer(Widget w, XtPointer client, XtPointer call)
{
    unsigned long n;

    n = (unsigned long) client;
    yes_no_answer = (n) ? True : False;
    yes_no_done = True;

	CloseDialogCB(w, 0, call);
}

static void YesNoEventHandler(Widget w, XtPointer client,
    XEvent * event, Boolean *dummy)
{
    if ( !XtIsRealized(w) )
        return;
    if ( event->type == KeyPress &&
         XK_Escape == XLookupKeysym((XKeyEvent *)event, 0) )
        SetBooleanAnswer(w, (XtPointer) 0, NULL);
}

Boolean AskBoolean(char *question, char *ok_label, char *no_label)
{
    Widget qshell;
    Widget qform, qlabel, qok, qno;
    Pixel bg, lbg;
    Screen *screen;
    Dimension x, y, width, height;

    XtVaGetValues(TopLevel, XtNbackground, &bg, NULL);
    lbg = light_color(None, bg);

    qshell = XtVaCreatePopupShell("YES/NO",
	topLevelShellWidgetClass, TopLevel,
	XtNwidth, 410,
	XtNmappedWhenManaged, False,
	NULL );
    qform = XtVaCreateManagedWidget("QForm",
	formWidgetClass, qshell,
	NULL);
    qlabel = XtVaCreateManagedWidget("QLabel",
	hLabelWidgetClass, qform,
	XtNlabel, question,
	XtNborderWidth, 0,
	XtNvertDistance, 10,
	XtNhorizDistance, 10,
	XtNtop, XawChainTop,
	XtNbottom, XawChainTop,
	XtNleft, XawChainLeft,
	NULL);
    qok = XtVaCreateManagedWidget("OkLabel",
	hCommandWidgetClass, qform,
	XtNlabel, ok_label,
	XtNbackground, lbg,
	XtNwidth, 100,
	XtNfromVert, qlabel,
	XtNvertDistance, 10,
	XtNhorizDistance, 50,
	XtNleft, XawChainLeft,
	XtNright, XawChainLeft,
	NULL);
    qno = XtVaCreateManagedWidget("NoLabel",
	hCommandWidgetClass, qform,
	XtNlabel, no_label,
	XtNbackground, lbg,
	XtNfromHoriz, qok,
	XtNhorizDistance, 100,
	XtNwidth, 100,
	XtNfromVert, qlabel,
	XtNvertDistance, 10,
	XtNleft, XawChainLeft,
	XtNright, XawChainLeft,
	NULL);

    XtAddCallback(qok, XtNcallback, SetBooleanAnswer, (XtPointer) 1);
    XtAddCallback(qno, XtNcallback, SetBooleanAnswer, (XtPointer) 0);

    XtManageChild(qshell);

    /* draw dialog in screnn center */
    XtVaGetValues(qno, XtNx, &x, XtNy, &y,
                  XtNwidth, &width, XtNheight, &height,
                  XtNscreen, &screen,
                  NULL);
    width = 4 * width;
    height = y + height + 10;
    XtVaSetValues( qshell, XtNwidth, width, XtNheight, height,
                   XtNminWidth, width, XtNminHeight, height,
                   XtNmaxWidth, width, XtNmaxHeight, height,
                   NULL );
    x = (WidthOfScreen(screen) - width) / 2;
    y = (HeightOfScreen(screen) - height) / 2;
    XtMoveWidget(qshell, x, y);
    XtPopup(qshell, XtGrabExclusive);

    XtAddEventHandler(qform, KeyPressMask, True,
                      YesNoEventHandler, qshell);
    AddWMDelCallback(qshell, SetBooleanAnswer, (XtPointer)0);

    yes_no_done = False;
    while ( yes_no_done == False ) {
		XtAppProcessEvent(app, XtIMAll);
	}

    return yes_no_answer;
}
