/* $Id: XuiInkey.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#define		XK_KOREAN
#include	<X11/keysym.h>
#include	<X11/Xatom.h>
#include	<X11/Xlib.h>
#include	<X11/Xlibint.h>
#include	<X11/Xutil.h>
#include	<Haw/TextOut.h>
#include	<Haw/Preedit.h>

#include	"hanim.h"
#include	"resource.h"
#include	"preference.h"
#include	"romandic.h"
#include	"hcodec.h"
#include	"hstring.h"
#include	"buffer.h"
#include	"XuiRootWin.h"

extern Widget	AddTerm, AddTermText[2];

extern int check_action_args(String *, int, char *);
extern Boolean SendCompoundText(Display *, Window, hchar_t *);

static inline Boolean IsHangulKey(KeySym keysym)
{
#ifdef XK_Hangul
    return ((keysym >= XK_Hangul_Kiyeog) && 
	(keysym <= XK_Korean_Won));
#else
    return 0;
#endif
}

static void alphanumeric(Buffer *b, HawTextVerifyPtr verify)
{
    unsigned char c;
    hchar_t *hh = verify->text->hptr;

    save_buffer(b);
    c = verify->text->ptr[0];

    if (!insert_buffer(b, c, '\0', b->conversion_mode)) {
        verify->doit = False;
        return;
    }

    /* unsigned short -> cyrillicconv, eng12conv -> hchar */
    if ( b->conversion_mode == CYRILLIC )
        hh[0] = cyrillicconv(c);
    else {
        if ( b->conversion_mode == ENGLISH_DVORAK )
            c = dvorakconv(c);
        /*  ݰ ȯ */
        hh[0] = (prefer.english12 == 1) ? c : eng12conv(c);
    }
    hh[1] = 0;
    if (!add_hangul(b, hh)) {
        verify->doit = False;
        return;
    }
    b->conversion_flag[b->point++] = True;
    if ( b->point == b->end_point )
        b->end_point++;
    b->cursor = verify->currInsert + 1;

    verify->text->length = 1;
    if (b->xim_style & (XIMPreeditPosition | XIMPreeditArea)) {
        if (b->state & EDIT)
            verify->text->mode = HawHIGHLIGHT_SELECTED;
        else
            verify->text->mode = HawHIGHLIGHT_SECONDARY_SELECTED;
    }
}

static void japaneseconv( Buffer *b, HawTextVerifyPtr verify )
{
    unsigned char c;
    hchar_t *hh = verify->text->hptr;

    save_buffer( b );
    c = verify->text->ptr[0];
    if (!insert_buffer( b, c, '\0', b->conversion_mode)) {
        verify->doit = False;
        return;
    }
    /* unsigned short -> cyrillicconv, eng12conv -> hchar */
    if (b->conversion_mode == JAPANESE_HIRAGANA)
        hh[0] = hiraganaconv(c);
    else
        hh[0] = katakanaconv(c);
    hh[1] = '\0';
    if (!add_hangul(b, hh)) {
        verify->doit = False;
        return;
    }
    b->conversion_flag[b->point++] = True;
    if (b->point == b->end_point)
	b->end_point++;
    b->cursor = verify->currInsert + 1;

    verify->text->length = 1;
    if (b->xim_style & (XIMPreeditPosition | XIMPreeditArea)) {
        if ( b->state & EDIT )
            verify->text->mode = HawHIGHLIGHT_SELECTED;
        else
            verify->text->mode = HawHIGHLIGHT_SECONDARY_SELECTED;
    }
}

static Boolean split_key(Buffer *b, RomanTable *rt)
{
    unsigned int	p = b->point++;
    unsigned int	ep = b->end_point;
    unsigned int	length;
    int	ii;

    b->keyC[ep + 1] = b->keyC[ep];
    if ( b->point == ep || rt || rt->str != NULL )
        ep = b->end_point++;

    b->conversion_flag[p] = True;
    length = (rt->str == NULL) ? 0 : strlen( (char *)rt->str );
    if ( length == 0 &&
	(b->keyC[p + 1] - b->keyC[p] == strlen((char *)rt->key)) )
        return( False );
    for ( ii = ep; ii >= 0 && ii > p; ii-- ) {
        b->keyC[ii + 1] = b->keyC[ii] + length;
        b->hangulC[ii + 1] = b->hangulC[ii];
        b->mode[ii + 1] = b->mode[ii];
        b->conversion_flag[ii + 1] = b->conversion_flag[ii];
    }
    b->mode[p + 1] = b->mode[p];
    b->keyC[p + 1] = b->keyC[p] + strlen( (char *)rt->key );
    p = b->point;
    b->hangulC[p] = b->hangulC[p + 1];
    b->conversion_flag[p] = False;

    if ( length != 0 ) {
        for ( ii = b->keyC[b->end_point]; ii >= 0 && ii >= b->keyC[p]; ii-- ) {
            b->key[ii + length] = b->key[ii];
            b->str[ii + length] = b->str[ii];
        }
        strncpy( (char *)&b->key[b->keyC[p]], (char *)rt->str, length );
        strncpy( (char *)&b->str[b->keyC[p]], (char *)rt->str, length );
    }
    return( True );
}

static void kanaconv(Buffer *b, HawTextVerifyPtr verify)
{
    unsigned char c;
    unsigned int point;
    char ss[32];
    hchar_t hh[32], xx[32];
    RomanTable *rt;
    int	ii, jj;
    ConversionMode mode = b->conversion_mode;

    save_buffer( b );
    c = verify->text->ptr[0];
    if ( !insert_buffer( b, c, '\0', mode ) ) {
        verify->doit = False;
        return;
    }

    point = b->point;
    b->conversion_flag[point] = False;
    if ( (ii = b->keyC[point + 1] - b->keyC[point]) > 30 ) {
        verify->doit = False;
        return;
    }

    strncpy(ss, (char *)&b->key[b->keyC[point]],
	(ii = b->keyC[point + 1] - b->keyC[point]));
    ss[ii] = '\0';
    if ((rt = jp_forward_search(ss)) == NULL) {
        if (jp_forward_sub_search(ss) != NULL) {
            /* ̺ ̱ Ȱ ڿ ã ,
            key ϴ ڿ ã ̴. ,  
            ʹ ٽ յ ɼ ִ ̴. */
            verify->text->hptr[0] = (prefer.english12 == 1) ? c : eng12conv(c);
            verify->text->hptr[1] = '\0';
            if (!add_hangul(b, verify->text->hptr)) {
                verify->doit = False;
                return;
            }
            verify->text->length = 1;
            b->cursor = verify->currInsert + 1;
        } else {
            /* key ϴ ڿ ȯ ̺  ̴.
             ̾ٸ  Է   ȿ
            . ó 纯ȯѴ.*/
            verify->startPos = b->hangulC[point];
            verify->endPos = b->hangulC[point + 1];
            b->cursor = b->hangulC[point];

            /* rt ʱ ã */
            rt = jp_backward_sub_search( (unsigned char *)ss );
            hh[0] = '\0';

            while ( rt != NULL ) {
                if (mode == JAPANESE_ENG_HIRAGANA ||
		    mode == JAPANESE_HIRAGANA)
                    hcscpy(xx, rt->hangul);
                else
                    hiras2katas(rt->hangul, xx);
                hcscat(hh, xx);
                if (!replace_hangul(b, xx)) {
                    verify->doit = False;
                    return;
                }
                if (!split_key(b, rt))
                    break;

                point = b->point;
                strncpy(ss, (char *)&b->key[b->keyC[point]],
		    (ii = b->keyC[point + 1] - b->keyC[point]));
                ss[ii] = '\0';
                rt = jp_backward_sub_search(ss);
            }

            if ( rt == NULL ) {
		unsigned char ch;
		
                /*   ڱ ȯ  ϰ,
                    Ʈ ٽ Ѵ. */
                for (ii = 0, jj = b->keyC[point], xx[0] = '\0';
		     jj < b->keyC[point + 1]; jj++, ii++) {
		    ch = b->key[jj];
		    xx[ii] = (prefer.english12 == 1) ? ch : eng12conv(ch);;
                }
                xx[ii] = '\0';  /*   ձ */
                hcscat( hh, xx );

                if (!replace_hangul(b, xx)) {
                    verify->doit = False;
                    return;
                }
            }
            verify->text->length = hcslen(hh);
            hcscpy(verify->text->hptr, hh);
            b->cursor += verify->text->length;
        }
    } else {
        /*  key Ȱ ڿ ִ ̴. */
        if ( rt->next != NULL &&
             strncmp( ss, (char *)rt->next->key, strlen(ss) ) == 0 ) {
            verify->text->hptr[0] = (prefer.english12 == 1) ? c : eng12conv(c);
            verify->text->hptr[1] = '\0';
            if ( !add_hangul( b, verify->text->hptr ) ) {
                verify->doit = False;
                return;
            }
            verify->text->length = 1;
            b->cursor = verify->currInsert + 1;
        } else {
            verify->startPos = b->hangulC[point];
            verify->endPos = b->hangulC[point + 1];
            if ( mode == JAPANESE_ENG_HIRAGANA ||
		mode == JAPANESE_HIRAGANA)
                hcscpy( hh, rt->hangul );
            else
                hiras2katas( rt->hangul, hh );
            verify->text->length = hcslen( hh );
            b->cursor = b->hangulC[point] + verify->text->length;

            if ( !replace_hangul( b, hh ) ) {
                verify->doit = False;
                return;
            }
            b->conversion_flag[point] = True;

            if ( split_key( b, rt ) ) {
                for ( ii = 0; (c = rt->str[ii]) != '\0'; ii++ ) {
                    xx[ii] = (prefer.english12 == 1) ? c : eng12conv(c);
                }
                xx[ii] = '\0';
                hcscat( hh, xx );
                if ( !add_hangul( b, xx ) ) {
                    verify->doit = False;
                    return;
                }
                b->cursor += hcslen( xx );
                verify->text->length += hcslen( xx );
            }
            hcscpy( verify->text->hptr, hh );
        }
    }
    if (b->xim_style & (XIMPreeditPosition | XIMPreeditArea)) {
        if (b->state & EDIT)
            verify->text->mode = HawHIGHLIGHT_SELECTED;
        else
            verify->text->mode = HawHIGHLIGHT_SECONDARY_SELECTED;
    }
}

static void hangulconv(Buffer *b, HawTextVerifyPtr verify)
{
    unsigned char c;
    hchar_t hh[22], xx[2];
    register int hangul2_backward = 0, point = b->point;
    unsigned short code1 = 0, code2 = 0;
    ConversionMode mode = b->conversion_mode;

    save_buffer( b );
    c = verify->text->ptr[0];

    /* Ÿ ʱȭ */
    hh[0] = 0;
    b->conversion_flag[point] = False;

#ifdef DEBUG
    if ( resources.debug & 0x80 ) {
        printf("========== before get hangul  =========\n");
        print_buffer(b);
    }
#endif
    switch (get_hangul_code(&b->hautomata, c, &code1, &code2)) {
    case code_hanst:
#ifdef DEBUG
        if ( resources.debug & 0x80 ) {
            printf("code_hanst: amstate = %d\n", b->hautomata.AMstate);
            printf("code1 = %2x, code2 = %2x\n", code1, code2);
        }
#endif
        hh[0] = code1; hh[1] = 0;
        if ( !insert_buffer(b, c, '\0', mode) || !add_hangul(b, hh) ) {
            verify->doit = False;
            return;
        }
        b->conversion_flag[b->point] = False;  /* for safety */
        /* Ŀ Ѵ */
        b->cursor = verify->currInsert + 1;
        break;

    case code_hanin:
        /* ̹ ԷµǾ ִ    ٲ۴. */
        hh[0] = code1; hh[1] = 0;
        if ( !insert_buffer( b, c, '\0', mode ) ||
             !replace_hangul( b, hh ) ) {
            verify->doit = False;
            return;
        }
        /*    ġ ״ д */
        verify->startPos = b->hangulC[point];
        verify->endPos = b->hangulC[point + 1];
        break;

    case code_hanend:
#ifdef DEBUG
        if ( resources.debug & 0x80 ) {
            printf("code_hanend: amstate = %d\n", b->hautomata.AMstate);
            printf("code1 = %2x, code2 = %2x\n", code1, code2);
        }
#endif
        /* ϼǾ   ó: ι  ʿϴ */
        hh[0] = code1; hh[1] = 0;
        if ( b->hangul[b->hangulC[point]] != hh[0] ) /*   */
            hangul2_backward = 1;
        if ( !replace_hangul( b, hh ) ) {
            verify->doit = False;
            return;
        }
        complete_hangul( b );

        verify->startPos = b->hangulC[point];
        verify->endPos = b->hangulC[point + 1];

        /* space  յǴ  */
        xx[0] = code2; xx[1] = 0;
        if (!insert_buffer( b, c, '\0', mode ) || !add_hangul(b, xx)) {
            verify->doit = False;
            return;
        }
        if ( hangul2_backward )
            b->keyC[b->point] -= 1;
        hcscat( hh, xx );
        verify->endPos = b->hangulC[point + 1];
        b->conversion_flag[b->point] = False;  /* for safety */
        b->cursor = verify->currInsert + 1;

        if ( ! b->hautomata.AMstate ) {
            b->conversion_flag[b->point++] = True;
            if ( b->point == b->end_point ) {
                b->end_point++;
                b->keyC[b->end_point] = b->keyC[b->point];
            }
        }
        break;

    case code_hanbs:
        break;

    case code_ascii:  /* 3 Bulsik needs the following code */
#ifdef DEBUG
        if ( resources.debug & 0x80 )
            printf("code_ascii : code1 = %d, code2 = %d\n", code1, code2);
#endif
        hh[0] = code1; hh[1] = 0;
        if ( !insert_buffer(b, c, '\0', mode) || !add_hangul(b, hh) ) {
            verify->doit = False;
            return;
        }
        b->conversion_flag[b->point++] = True;
        if ( b->point == b->end_point )
            b->end_point++;
        b->cursor = verify->currInsert + 1;
        break;

    default:  /* code_nohan || code_delete */
        break;
    }

    hcscpy(verify->text->hptr, hh);
    verify->text->length = hcslen(hh);

    if ( b->xim_style & (XIMPreeditPosition | XIMPreeditArea) ) {
        if ( b->state & EDIT )
            verify->text->mode = HawHIGHLIGHT_SELECTED;
        else
            verify->text->mode = HawHIGHLIGHT_SECONDARY_SELECTED;
    }
}

static void inkey_lookup(Buffer *b, Widget w, HawTextVerifyPtr verify)
{
    KeySym keysym;
    int ii, count;
    char str[32];
    unsigned char c;
    XKeyEvent *kevent = (XKeyEvent *)verify->event;
    ConversionMode mode;

    /* ignore CapLock in Hangul input mode */
    if (b->conversion_mode == HANGUL_2BULSIK ||
	b->conversion_mode == HANGUL_2BULSIK ||
	b->conversion_mode == JAPANESE_HAN_HIRAGANA ||
	b->conversion_mode == JAPANESE_HAN_KATAKANA)
	kevent->state &= ~LockMask;

    if (!(count = XLookupString(kevent, str, 8, &keysym, NULL)))
    {
        /*  ѱȭǾ Է  ƴ϶ ѱ۷
            ̴.  ̷   
           Ѵ.  ΰ  ΰ?
           Ͽư  ڼ  2 byte Ѵٰ ϰ
           α׷ Ѵ. */
        if (IsHangulKey(keysym) == False) {
	    verify->doit = False;
            return;
	}
        count = 2;
    }

    verify->text->mode = HawHIGHLIGHT_NORMAL;
    verify->text->hptr[0] = 0;
    verify->text->length = count;
    for (ii = 0; ii < count; ++ii)
		verify->text->ptr[ii] = (prefer.is_xkb_dvorak && !IsKeypadKey(keysym))
			? qwertyconv(str[ii]) : str[ii];
    verify->text->ptr[count] = 0;
    c = verify->text->ptr[0];

    /* do not convert if keypad key */
    if (IsKeypadKey(keysym))
	mode = ENGLISH_QWERTY;
    else
	mode = b->conversion_mode;

#if 0 
    /* ǿ ѱ   */
    if (c >= 0xa1 && c < 0xe0) {
	if (HANGUL != (mode & LANGMASK))
	    return;
	hangulconv( b, verify );
	if (verify->doit)
	    verify->newInsert = b->cursor;
	else
	    load_buffer( b );
	return;
    }
#endif
    if (b->state & ADD) {
        if (c >= 0x20 && c < 0x7f)
            hangulconv(b, verify);
        if (verify->doit)
            verify->newInsert = b->cursor;
        else
            load_buffer(b);
    } else if (b->state & (PREEDIT | EDIT)) {
	/* c MSB 0̰, ڰ ƴ  */
	switch (mode) {
	case ENGLISH_QWERTY:
	case ENGLISH_DVORAK:
	case CYRILLIC:
	    alphanumeric(b, verify);
	    break;
	case JAPANESE_HIRAGANA:
	case JAPANESE_KATAKANA:
	    japaneseconv(b, verify);
	    break;
	case JAPANESE_ENG_HIRAGANA:
	case JAPANESE_ENG_KATAKANA:
	    kanaconv(b, verify);
	    break;
	case HANGUL_3BULSIK:
	case HANGUL_2BULSIK:
	    hangulconv(b, verify);
	    break;
#if 0 /* LATER */
	case JAPANESE_HAN_HIRAGANA:
	case JAPANESE_HAN_KATAKANA:
	case HANGUL_HAN_ROMAN:
	    break;
#endif  /* LATER */
	}
	if (verify->doit)
	    verify->newInsert = b->cursor;
	else
	    load_buffer(b);
    } else {
        verify->doit = False;
    }
}

void inkey(Widget w, XEvent *event, String *params, Cardinal *n)
{
    Buffer *b;
    HawTextVerifyStruct	verify;
    HawTextBlockRec text;

    if (!event || (event->type != KeyPress &&
	event->type != KeyRelease))
        return;
    b = set_buffer( w, event->type );
    if (b == NULL || !check_action_args(params, *n, "inkey"))
        return;
    if (b->state & (ERROR | CODE | HANJA | DELETE))
        return;
    /* root Է Ÿ */
    if ((b->xim_style & XIMPreeditNothing) && resources.ghost_mode)
        GhostShow();

    /* prepare to inkey_lookup() */
    verify.event = event;
    verify.doit = True;
    verify.currInsert = verify.newInsert =
	verify.startPos = verify.endPos = b->cursor;
    verify.text = &text;

    /* ⿡  ڵ ȯѴ. */
    inkey_lookup(b, w, &verify);
    if (!verify.doit || (b->state & HANJA))
        return;
    if (verify.text->hptr[0] == '\0') {
        XBell(XtDisplay(w), 0);
        return;
    }

    if (b->state & ADD) {
        HawTextOutReplaceHcs(AddTermText[1], verify.startPos,
	    verify.endPos, text.hptr);
        HawTextOutSetHighlight(AddTermText[1], verify.startPos,
	    verify.endPos, text.mode);
        HawTextOutSetCursorPosition(AddTermText[1], verify.newInsert);
        if (b->cursor == 1)
            XtVaSetValues(AddTerm, XtNsensitive, True, NULL);
    }
    else if (b->state & EDIT) {
		// zzz
        if (b->xim_style & XIMPreeditNothing) {
            HawTextOutReplaceHcs(PEText, verify.startPos,
		verify.endPos, text.hptr );
            HawTextOutSetHighlight(PEText, verify.currInsert,
		verify.newInsert, text.mode);
            HawTextOutSetCursorPosition(PEText, verify.newInsert);
        } else {
            HawPreeditReplaceHcs( b->preedit_widget,
		verify.startPos + text.length,
		verify.endPos + text.length,
		text.hptr );
            HawPreeditSetHighlight( b->preedit_widget,
		verify.currInsert + text.length,
		verify.newInsert + text.length,
		text.mode );
        }
    } else if ( b->state & PREEDIT ) {
        if ( b->xim_style & XIMPreeditNothing ) {
            HawTextOutReplaceHcs(b->preedit_widget, verify.startPos,
		verify.endPos, text.hptr);
            HawTextOutSetHighlight(b->preedit_widget, verify.currInsert,
		verify.newInsert, text.mode);
            HawTextOutSetCursorPosition(b->preedit_widget,
		verify.newInsert);
        } else {
#if 1 // xxx
            if ((b->conversion_mode & LANGMASK) == ENGLISH) {
                if ( SendCompoundText(event->xany.display,
		    event->xkey.subwindow, b->hangul) ) {
                    register_buffer(b);
                    buffer_initialize(b);
                }
            } else {
#endif
                HawPreeditReplaceHcs( b->preedit_widget, verify.startPos,
		    verify.endPos, text.hptr );
                HawPreeditSetHighlight( b->preedit_widget, verify.startPos,
		    verify.endPos, text.mode );
#if 1 // xxx
	    }
#endif
	}
    }
}

