/* $Id: XuiKeycap.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <Haw/Command.h>

#include "hanim.h"
#include "hangul.h"
#include "resource.h"
#include "buffer.h"
#include "XuiDialog.h"
#include "hcodec.h"

extern Pixel light_color( Colormap, Pixel );

/*   */
#define NROWS	4

static int numCols[4] = { 14, 12, 11, 10 };

static char *keyboard[4] =
    {
        "`1234567890-=\\",
        "qwertyuiop[]",
        "asdfghjkl;'",
        "zxcvbnm,./",
    };

static char *sKeyboard[4] =
    {
        "~!@#$%^&*()_+|",
        "QWERTYUIOP{}",
        "ASDFGHJKL:\"",
        "ZXCVBNM<>?",
    };

static void make_keypad_string(char *str, int ii, int jj, ConversionMode mode)
{
    hchar_t hh, hhs, dummy;
    unsigned short code;
    struct hangul_automata_state hfrm;

    switch ( mode ) {
    case HANGUL3:
    case HANGUL2:
        hfrm.han_kbd_kind = (mode == HANGUL3) ? 3 : 2;
        hfrm.han_kbd_state = 1;
        clear_hangul_automata(&hfrm);
        get_hangul_code(&hfrm, sKeyboard[jj][ii], &hh, &dummy);

        clear_hangul_automata(&hfrm);
        get_hangul_code(&hfrm, keyboard[jj][ii], &hhs, &dummy);
        break;
    case ENGLISH_DVORAK:
        hh = dvorakconv(sKeyboard[jj][ii]);
        hhs = dvorakconv(keyboard[jj][ii]);
        break;
    case JAPANESE_HIRAGANA:
        hh = hiraganaconv(sKeyboard[jj][ii]);
        hhs = hiraganaconv(keyboard[jj][ii]);
        break;
    case JAPANESE_KATAKANA:
        hh = katakanaconv(sKeyboard[jj][ii]);
        hhs = katakanaconv(keyboard[jj][ii]);
        break;
    case CYRILLIC:
        hh = cyrillicconv(sKeyboard[jj][ii]);
        hhs = cyrillicconv(keyboard[jj][ii]);
        break;
    case ENGLISH_QWERTY:
    case JAPANESE_ENG_HIRAGANA:
    case JAPANESE_ENG_KATAKANA:
        hh = sKeyboard[jj][ii];
        hhs = keyboard[jj][ii];
        break;
    }

    code = hc2ksc(hh);
    str[0] = code >> 8;
    str[1] = code & 0x00ff;
    code = keyboard[jj][ii];
    str[2] = code >> 8;
    str[3] = code & 0x00ff;
    str[4] = '\n';
    code = hc2ksc(hhs);
    str[5] = code >> 8;
    str[6] = code & 0x00ff;

    for ( ii = 0; ii < 8; ii++ )
        if ( str[ii] == 0 )
            str[ii] = ' ';
    str[7] = 0;
}

void keycap( Widget w, XEvent *event, String *params, Cardinal *nn )
{
    Buffer *b;
    Widget	kshell;
    Widget	kform, close;
    unsigned char	keystr[9];
    Pixel	bg, lbg;
    Arg	args[20];
    int	n, ii, jj;
    Dimension	width, height;
    ConversionMode mode;

    if (!event || (b = set_buffer( w, event->type )) == NULL)
        mode = resources.hangul_keyboard;
    else
        mode = b->conversion_mode;

    XtVaGetValues(TopLevel, XtNbackground, &bg, NULL);
    lbg = light_color((Colormap)NULL, bg);

    n = 0;
    width = 14 * 50 - 4 + 6 * 2;
    height = 4 * 45 - 4 + 6 * 2;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    XtSetArg( args[n], XtNiconic, False ); n++;
    XtSetArg( args[n], XtNgeometry, "+80+80" ); n++;
    XtSetArg( args[n], XtNmappedWhenManaged, False ); n++;
    XtSetArg( args[n], XtNwidth, width ); n++;
    XtSetArg( args[n], XtNheight, height ); n++;
    kshell = XtCreatePopupShell( "Keycaps",
                                 topLevelShellWidgetClass, TopLevel, args, n );

    n = 0;
    XtSetArg( args[n], XtNbackground, bg ); n++;
    kform = XtCreateManagedWidget( "KeycapForm",
                                   formWidgetClass, kshell, args, n );

    for ( jj = 0; jj < NROWS; jj++ ) {
        int xoff = (14 - numCols[jj]) * 25 + 5;
        for ( ii = 0; ii < numCols[jj]; ii++ ) {
            n = 0;
            XtSetArg( args[n], XtNbackground, lbg ); n++;
            XtSetArg( args[n], XtNhorizDistance, ii*50 + xoff); n++;
            XtSetArg( args[n], XtNvertDistance, jj*45 + 5 ); n++;
            XtSetArg( args[n], XtNwidth, 46 ); n++;
            XtSetArg( args[n], XtNheight, 41 ); n++;
            XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
            XtSetArg( args[n], XtNright, XawChainLeft ); n++;

            make_keypad_string(keystr, ii, jj, mode );
            XtCreateManagedWidget(keystr, hLabelWidgetClass, kform,
                                  args, n );
        }
    }
    /* ݱ  */
    n = 0;
    close = XtVaCreateManagedWidget( "close",
                                     hCommandWidgetClass, kform,
                                     XtNlabel, resources.close_label,
                                     XtNbackground, lbg,
                                     XtNwidth, 80,
                                     XtNhorizDistance, width - 85,
                                     XtNvertDistance, height - 28,
                                     NULL);
    XtAddCallback(close, XtNcallback, CloseDialogCB, NULL);

    /* add callback */
    XtRealizeWidget(kshell);
    XtAddEventHandler(kform, KeyPressMask, False,
	CloseDialogHandler, NULL);
    XtAddCallback(kshell, XtNdestroyCallback, CloseDialogCB, NULL);
    AddWMDelCallback(kshell, CloseDialogCB, kshell);

    XtVaGetValues(kshell, XtNwidth, &width, XtNheight, &height, NULL);
    XtVaSetValues(kshell, XtNminWidth, width, XtNmaxWidth, width,
                  XtNminHeight, height, XtNmaxHeight, height, NULL);
    XtMapWidget(kshell);
}
