/* $Id: XuiLang.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include	<assert.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/keysym.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Box.h>
#include	<Haw/Command.h>

#include	"hanim.h"
#include	"buffer.h"
#include	"resource.h"

extern void hangul3(Widget, XEvent *, String *, Cardinal *);
extern void hangul2(Widget, XEvent *, String *, Cardinal *);
extern void english_qwerty(Widget, XEvent *, String *, Cardinal *);
extern void english_dvorak(Widget, XEvent *, String *, Cardinal *);
extern void hiragana_roman(Widget, XEvent *, String *, Cardinal *);
extern void katakana_roman(Widget, XEvent *, String *, Cardinal *);
extern void hiragana(Widget, XEvent *, String *, Cardinal *);
extern void katakana(Widget, XEvent *, String *, Cardinal *);
extern void cyrillic(Widget, XEvent *, String *, Cardinal *);
extern int check_action_args(String *, int, char *);

#define		MAXLANG	9

static Widget	kshell = NULL, kbox, klang[MAXLANG];
static int	current_item = 0;

static void invert_lang_item(int lang)
{
    XGCValues	gcv;
    GC	gc;
    Dimension	width, height;
    Pixel	fg, bg;

    assert( 0 <= lang && lang < MAXLANG );
    XtVaGetValues(klang[lang], XtNwidth, &width, XtNheight, &height,
                  XtNforeground, &fg, XtNbackground, &bg, 0);
    gcv.foreground = fg ^ bg;
    gcv.function	= GXxor;
    gcv.line_width = 2;
    gc = XtGetGC(TopLevel, GCForeground | GCFunction | GCLineWidth, &gcv);
    XDrawRectangle(XtDisplay(klang[lang]), XtWindow(klang[lang]),
                   gc, 1, 1, width - 2, height - 2);
    XtReleaseGC(TopLevel, gc);
}

static void place_select_lang(Widget w, XtPointer client, XtPointer call)
{
    Window	root, child;
    unsigned int	mask;
    int	x, y, wx, wy;
    Dimension	width, height;

    XQueryPointer(XtDisplay(w), XtWindow(w), &root, &child,
                  &x, &y, &wx, &wy, &mask);
    XtVaGetValues(kshell, XtNwidth, &width, XtNheight, &height, NULL);
    if ( DisplayWidth(XtDisplay(w), 0) < x + width )
        x -= width;
    if ( DisplayHeight(XtDisplay(w), 0) < y + height )
        y -= height;
    XtVaSetValues(w, XtNx, x, XtNy, y, NULL );
}

static void end_select_lang(void)
{
    if ( kshell == 0 )
        return;

    XUngrabKeyboard( XtDisplay(kbox), CurrentTime );
    XUngrabPointer( XtDisplay(kbox), CurrentTime );
    XtPopdown(kshell);
    XtDestroyWidget(kshell);
    kshell = NULL;
}

static void set_select_lang(Widget w, XtPointer client, XtPointer call)
{
    XEvent *ev = (XEvent *)client;
    int	val, nn = 0;

    end_select_lang();
    if (ev == NULL) return;
    for (val = 0; val < MAXLANG; val++)
        if (w == klang[val])
            break;
    if (val == MAXLANG)
        return;
    switch (val) {
    case 0: hangul3(w, ev, NULL, &nn); break;
    case 1: hangul2(w, ev, NULL, &nn); break;
    case 2: english_qwerty(w, ev, NULL, &nn); break;
    case 3: english_dvorak(w, ev, NULL, &nn); break;
    case 4: hiragana_roman(w, ev, NULL, &nn); break;
    case 5: katakana_roman(w, ev, NULL, &nn); break;
    case 6: hiragana(w, ev, NULL, &nn); break;
    case 7: katakana(w, ev, NULL, &nn); break;
    case 8: cyrillic(w, ev, NULL, &nn); break;
    default: break;
    }

	/* change keyboard toggle value to selected one */
	if (val == 0)
		resources.hangul_keyboard = HANGUL3;
	else if (val == 1)
		resources.hangul_keyboard = HANGUL2;
	else if (val == 2)
		resources.english_keyboard = ENGLISH_QWERTY;
	else if (val == 3)
		resources.english_keyboard = ENGLISH_DVORAK;
}

static void select_lang_handler(Widget w, XtPointer client,
                                XEvent *xev, Boolean *dummy)
{
    KeySym	keysym;
    Dimension	width, height;

    XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, 0);

    switch ( xev->type ) {
    case MapNotify:
        XGrabKeyboard(XtDisplay(w), XtWindow(w), False,
                      GrabModeAsync, GrabModeAsync, CurrentTime);
        XGrabPointer(XtDisplay(w), XtWindow(w), True,
                     ButtonPressMask, GrabModeAsync, GrabModeAsync,
                     None, None, CurrentTime);
        invert_lang_item(current_item);
        break;
    case KeyPress:
        keysym = XLookupKeysym((XKeyEvent *)xev, 0);
        switch ( keysym ) {
        case XK_Escape:
            end_select_lang();
            break;
case XK_0: case XK_1: case XK_2: case XK_3: case XK_4:
case XK_5: case XK_6: case XK_7: case XK_8: case XK_9:
            current_item = (keysym - XK_0) % MAXLANG;
            set_select_lang(klang[current_item], (XtPointer)xev, NULL);
            break;
        case XK_space:
        case XK_Return:
            set_select_lang(klang[current_item], (XtPointer)xev, NULL);
            break;
        case XK_j:
        case XK_n:
        case XK_Tab:
        case XK_Down:
        case XK_Right:
            invert_lang_item(current_item);
            current_item = (current_item + 1) % MAXLANG;
            invert_lang_item(current_item);
            break;
        case XK_k:
        case XK_p:
        case XK_Up:
        case XK_Left:
            invert_lang_item(current_item);
            current_item = (current_item - 1 + MAXLANG) % MAXLANG;
            invert_lang_item(current_item);
            break;
        case XK_Home:
        case XK_Prior:
            invert_lang_item(current_item);
            current_item = 0;
            invert_lang_item(current_item);
            break;
        case XK_End:
        case XK_Next:
            invert_lang_item(current_item);
            current_item = MAXLANG - 1;
            invert_lang_item(current_item);
            break;
        }
        break;
    case ButtonPress:
        if ( 0 < xev->xbutton.x && xev->xbutton.x < width &&
             0 < xev->xbutton.y && xev->xbutton.y < height ) {
            invert_lang_item(current_item);
            current_item = xev->xbutton.y * MAXLANG / height;
            invert_lang_item(current_item);
            set_select_lang( klang[current_item], (XtPointer)xev, NULL);
        }
        end_select_lang();
        break;
    case MotionNotify:
        if ( 0 < xev->xbutton.x && xev->xbutton.x < width &&
             0 < xev->xbutton.y && xev->xbutton.y < height ) {
            invert_lang_item(current_item);
            current_item = xev->xbutton.y * MAXLANG / height;
            invert_lang_item(current_item);
        }
        break;
    }
}

void select_lang(Widget w, XEvent *xev, String *params, Cardinal *n)
{
    Buffer	*b;
    Arg	args[2];

    /* buffer  */
    if ( (b = set_buffer( w, xev->type )) == 0 ||
         (b->state & ERROR) || b->shell_widget ||
         !check_action_args( params, *n, "select-lang" ) )
        return;
    if ( COMB_STATE(b->hautomata) ) {
        complete_hangul( b );
        clear_hangul_automata( &b->hautomata );
    }
    /*   */
    switch ( b->conversion_mode ) {
    case HANGUL_3BULSIK: current_item = 0; break;
    case HANGUL_2BULSIK: current_item = 1; break;
    case ENGLISH_QWERTY: current_item = 2; break;
    case ENGLISH_DVORAK: current_item = 3; break;
    case JAPANESE_ENG_HIRAGANA: current_item = 4; break;
    case JAPANESE_ENG_KATAKANA: current_item = 5; break;
    case JAPANESE_HIRAGANA: current_item = 6; break;
    case JAPANESE_KATAKANA: current_item = 7; break;
    case CYRILLIC: current_item = 8; break;
    default: current_item = 0; break;
    }

    kshell = XtVaCreatePopupShell("KeyboardSelect",
                                  overrideShellWidgetClass, TopLevel,
                                  XtNmappedWhenManaged, False,
                                  NULL);
    kbox = XtVaCreateManagedWidget("KeyboardBox",
                                   boxWidgetClass, kshell,
                                   XtNhSpace, 0, XtNvSpace, 0,
                                   NULL);
    XtSetArg(args[0], XtNborderWidth, 0);
    XtSetArg(args[1], XtNwidth, 135);

    klang[0] = XtCreateManagedWidget(resources.hangul_3bulsik,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[1] = XtCreateManagedWidget(resources.hangul_2bulsik,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[2] = XtCreateManagedWidget(resources.english_qwerty,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[3] = XtCreateManagedWidget(resources.english_dvorak,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[4] = XtCreateManagedWidget(resources.japanese_eng_hiragana,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[5] = XtCreateManagedWidget(resources.japanese_eng_katakana,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[6] = XtCreateManagedWidget(resources.japanese_hiragana,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[7] = XtCreateManagedWidget(resources.japanese_katakana,
                                     hLabelWidgetClass, kbox, args, 2);
    klang[8] = XtCreateManagedWidget(resources.cyrillic_mode,
                                     hLabelWidgetClass, kbox, args, 2);
    XtAddEventHandler(kbox, StructureNotifyMask | PointerMotionMask |
                      ButtonPressMask | KeyPressMask,
                      False, select_lang_handler, NULL);
    /* popup and pulldown menu position */
    if ( xev->type == KeyPress )
        XtAddCallback(kshell, XtNpopupCallback, place_select_lang, b);
    else
        XtAddCallback(kshell, XtNpopupCallback, place_select_lang, NULL);

    XtManageChild(kshell);
    XtPopup(kshell, XtGrabExclusive);
}

