/* $Id: XuiPrefer.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include <stdio.h>
#include <errno.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <Haw/Label.h>
#include <Haw/Command.h>

#include "hanim.h"
#include "util.h"
#include "preference.h"
#include "XuiDialog.h"
#include "bitmaps/check.xbm"

#define HANIMRC_SIGNATURE "# hanIM Config File"
#ifndef HANIMRC
#  define HANIMRC "hanIMrc"
#endif

extern Boolean is_xkb_dvorak();

Preference prefer = {
    COMMIT_WORD,	/* commit method */
    1,			/* bangak/jungak */
    STATIC_CONTROL,	/* event flow control */
    0,			/* english keyboard */
};

// preference ui data

typedef struct prefer_entry_struct {
    Widget widget;
    char *name;
    int state;
} PreferEntryStruct, *PreferEntry;

static PreferEntryStruct commit_data[] = {
    { 0, "Char", False },
    { 0, "Word", True },
    { 0, "Line", False },
    { 0, 0, 0 }
};

static PreferEntryStruct eng12_data[] = {
    { 0, "One", True },
    { 0, "Two", False },
    { 0, 0, 0 }
};

static PreferEntryStruct flow_control_data[] = {
    { 0, "Static", True },
    { 0, "Dynamic", False },
    { 0, 0, 0 }
};

/* return False if there is no name/value pair in file */
static Boolean read_rc_entry(char *name, char *value, char *def)
{
    FILE *fp;
    char filename[BUFSIZ];
    int found = 0;
    char *line, *optstart;
    int line_length;
    int name_length;
    size_t line_allocated;

    strcpy(value, def);

    sprintf(filename, "%s/%s/%s", home_directory, HANIMUSRDIR, HANIMRC);
    fp = fopen(filename, "r");
    if (fp == NULL)
	return;

    /* Check each line to see if we have this entry already. */
    line = NULL;
    line_allocated = 0;
    name_length = strlen(name);

    while ((line_length = getline(&line, &line_allocated, fp)) >= 0)
    {
	/* skip over comment lines */
	if (line[0] == '#')
	    continue;
	/* stop if we match the current command */
	if (!STRNCASECMP(line, name, name_length)
	    && isspace((unsigned char) *(line + name_length)))
	{
	    found = 1;
	    break;
	}
    }
    if (fclose(fp) < 0)
	fprintf(stderr, "Error[%d]: cannot close %s", errno, filename);

    if (found) {
	/* skip over command in the options line */
	for (optstart = strtok(line + name_length, "\t \n");
	     optstart; optstart = strtok (NULL, "\t \n"))
	{
	    strcpy(value, optstart);
	}
    }
    if (line != NULL)
	free(line);

    return (Boolean)found;
}

static void write_hanim_rc()
{
    char dirname[BUFSIZ], filename[BUFSIZ];
    FILE *fp;

    sprintf(dirname, "%s/%s", home_directory, HANIMUSRDIR);
    sprintf(filename, "%s/%s", dirname, HANIMRC);

    if (!isdir(dirname)) {
	if (access(dirname, 0) == 0)
	    unlink(dirname);
	mkdir(dirname, 0777);
    }

    fp = fopen(filename, "w");

    if (fp != NULL) {
	fprintf(fp, "%s\n", HANIMRC_SIGNATURE);
	
	// commit method
	fprintf(fp, "Commit Method = ");
	if (prefer.commit_method == COMMIT_CHAR)
	    fprintf(fp, "%s\n", "Char");
	else if (prefer.commit_method == COMMIT_LINE)
	    fprintf(fp, "%s\n", "Line");
	else
	    fprintf(fp, "%s\n", "Word");

	// event flow control
	fprintf(fp, "Event Flow Control = ");
	if (prefer.event_flow == DYNAMIC_CONTROL)
	    fprintf(fp, "%s\n", "Dynamic");
	else
	    fprintf(fp, "%s\n", "Static");

	fclose(fp);
    }
}

void read_hanim_rc()
{
    char value[256];
    PreferEntry ptr;

    // commit method
    read_rc_entry("Commit Method", value, "Word");
    if (!STRCASECMP(value, "Char"))
	prefer.commit_method = COMMIT_CHAR;
    else if (!STRCASECMP(value, "Line"))
	prefer.commit_method = COMMIT_LINE;
    else
	prefer.commit_method = COMMIT_WORD;

    ptr = commit_data;
    while (ptr->name) {
	ptr->state = !STRCASECMP(value, ptr->name);
	ptr++;
    }

    // event flow control
    read_rc_entry("Event Flow Control", value, "Static");
    if (!STRCASECMP(value, "Dynamic"))
	prefer.event_flow = DYNAMIC_CONTROL;
    else
	prefer.event_flow = STATIC_CONTROL;

    ptr = flow_control_data;
    while (ptr->name) {
	ptr->state = !STRCASECMP(value, ptr->name);
	ptr++;
    }

    prefer.is_xkb_dvorak = is_xkb_dvorak();
}

// preference XUI data

#define ENTRY_WIDTH	300
#define BUTTON_START	30
#define BUTTON_TOTAL	180

static void toggleCB(Widget w, XtPointer client, XtPointer call);


// X Window Dialog

static void widget_set_toggle(Widget w, Boolean value)
{
    static Pixmap check_pix = 0;
    
    if (!check_pix)
        check_pix = XCreateBitmapFromData(XtDisplay(TopLevel),
	    XtWindow(TopLevel), check_bits, check_width, check_height);
    if (value)
	XtVaSetValues(w, XtNleftBitmap, check_pix, NULL);
    else
	XtVaSetValues(w, XtNleftBitmap, 0, NULL);
}

static void toggleCB(Widget w, XtPointer client, XtPointer call)
{
    PreferEntry ptr = (PreferEntry)client;
    
    while (ptr->name) {
	ptr->state = (ptr->widget == w) ? True : False;
	widget_set_toggle(ptr->widget, ptr->state);
	ptr++;
    }
}

// event flow control

static void create_toggle(char *title, Widget form, PreferEntry entry)
{
    Widget label;
    PreferEntry ptr;
    int nitems, width, hskip, xpos;
    Pixel lbg = light_color(None, default_background);

    label = XtVaCreateManagedWidget(title,
	hLabelWidgetClass, form, XtNborderWidth, 0,
	XtNleft, XawChainLeft, XtNright, XawChainLeft,
	XtNbackground, lbg,
	NULL);
    
    for (nitems = 0, ptr = entry; ptr->name; ptr++)
	nitems++;
    width = BUTTON_TOTAL / nitems;
    hskip = (ENTRY_WIDTH - BUTTON_TOTAL - BUTTON_START * 2) /
	(nitems - 1);
    xpos = BUTTON_START;

    ptr = entry;
    while (ptr->name) {
	ptr->widget = XtVaCreateManagedWidget(ptr->name,
	    hCommandWidgetClass, form,
	    XtNbackground, lbg,
	    XtNwidth, width,
	    XtNfromVert, label,
	    XtNvertDistance, 10, XtNhorizDistance, xpos,
	    XtNleft, XawChainLeft, XtNright, XawChainLeft,
	    NULL);
	XtAddCallback(ptr->widget, XtNcallback, toggleCB, entry);

	xpos += width + hskip;
	ptr++;
    }
}

static void set_entry_state(PreferEntry entry)
{
    while (entry->widget) {
	widget_set_toggle(entry->widget, entry->state);
	entry++;
    }
}

static void applyPrefer(Widget w, XtPointer client, XtPointer call)
{
    Preference prefer2;
    PreferEntry ptr;

    /* store preference data */
    memcpy(&prefer2, &prefer, sizeof(Preference));

    /* commit method */
    ptr = commit_data;
    while (ptr->state == False)	ptr++;
    if (!STRCASECMP(ptr->name, "Char"))
	prefer2.commit_method = COMMIT_CHAR;
    else if (!STRCASECMP(ptr->name, "Line"))
	prefer2.commit_method = COMMIT_LINE;
    else
	prefer2.commit_method = COMMIT_WORD;
    
    /* event flow control */
    ptr = flow_control_data;
    while (ptr->state == False)	ptr++;
    if (!STRCASECMP(ptr->name, "Dynamic"))
	prefer2.event_flow = DYNAMIC_CONTROL;
    else
    	prefer2.event_flow = STATIC_CONTROL;
    
    /* save preference */
    if (memcmp(&prefer, &prefer2, sizeof(Preference))) {
	memcpy(&prefer, &prefer2, sizeof(Preference));
	write_hanim_rc();
    }
}

void preference(void)
{
    Widget pshell, form, entryform;
    Widget apply, cancel;
    Dimension width, height;
    Pixel lbg = light_color( None, default_background );

    /* Dialog Shell */
    pshell = XtVaCreatePopupShell("Preference",
	transientShellWidgetClass, TopLevel,
	XtNwidth, ENTRY_WIDTH + 20,
	XtNmappedWhenManaged, False,
	0 );

    /* Form which contains push buttons */
    form = XtCreateManagedWidget("Form", formWidgetClass,
	pshell, NULL, 0);

    /* create commit method */
    entryform = XtVaCreateManagedWidget("CommitMethod",
	formWidgetClass, form,
	XtNhorizDistance, 5,
	XtNtop, XawChainTop, XtNbottom, XawChainTop,
	XtNleft, XawChainLeft,XtNright, XawChainRight,
	XtNbackground, lbg,
	0);
    create_toggle("CommitMethod", entryform, commit_data);

    /* create eng12 control */
    entryform = XtVaCreateManagedWidget("English12",
	formWidgetClass, form,
	XtNhorizDistance, 5, XtNresizable, True,
	XtNfromVert, entryform,
	XtNtop, XawChainTop, XtNbottom, XawChainTop,
	XtNleft, XawChainLeft,XtNright, XawChainRight,
	XtNvertDistance, 5, XtNbackground, lbg,
	0);
    create_toggle("English12", entryform, eng12_data);

    /* create event flow control */
    entryform = XtVaCreateManagedWidget("EventFlowControl",
	formWidgetClass, form,
	XtNfromVert, entryform,
	XtNtop, XawChainTop, XtNbottom, XawChainTop,
	XtNleft, XawChainLeft,XtNright, XawChainRight,
	XtNhorizDistance, 5, XtNresizable, True,
	XtNvertDistance, 5, XtNbackground, lbg,
	0);
    create_toggle("EventFlowControl", entryform, flow_control_data);

    /* create apply/cancel button */
    apply = XtVaCreateManagedWidget("Ok",
	hCommandWidgetClass, form,
	XtNbackground, lbg,
	XtNwidth, 90,
	XtNfromVert, entryform,
	XtNvertDistance, 10,
	XtNhorizDistance, 10,
	XtNleft, XawChainLeft,
	XtNright, XawChainLeft,
	NULL);
    cancel = XtVaCreateManagedWidget("No",
	hCommandWidgetClass, form,
	XtNbackground, lbg,
	XtNwidth, 90,
	XtNfromVert, entryform,
	XtNfromHoriz, apply,
	XtNvertDistance, 10,
	XtNleft, XawChainLeft,
	XtNright, XawChainLeft,
	NULL);

    XtAddCallback(apply, XtNcallback, applyPrefer, NULL);
    XtAddCallback(apply, XtNcallback, CloseDialogCB, NULL);
    XtAddCallback(cancel, XtNcallback, CloseDialogCB, NULL);

    XtRealizeWidget(pshell);
    XtVaGetValues(pshell, XtNwidth, &width, XtNheight, &height, NULL);
    XtVaSetValues(pshell, XtNminWidth, width, XtNminHeight, height,
	XtNmaxWidth, width, XtNmaxHeight, height,
	NULL);

    XtAddEventHandler(form, KeyPressMask, False, CloseDialogHandler, 0);
    AddWMDelCallback(pshell, CloseDialogCB, NULL);

    /* set toggle state */
    set_entry_state(commit_data);
    set_entry_state(eng12_data);
    set_entry_state(flow_control_data);

    XtPopup(pshell, XtGrabExclusive);
}
