/* $Id: XuiRootWin.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include	<stdlib.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Form.h>
#include	<X11/Xaw/Box.h>
#include	<Haw/HanList.h>
#include	<Haw/Command.h>
#include	<Haw/TextOut.h>

#include	"hanim.h"
#include	"resource.h"
#include	"buffer.h"

extern void	codetable(Widget, XEvent *, String *, Cardinal *);   /* code.c	*/
extern void	addterm(Widget, XEvent *, String *, Cardinal *);     /* addterm.c	*/
extern void	hanja(Widget, XEvent *, String *, Cardinal *);
extern void	select_lang();   		/* langmenu.c	*/
extern void	help();

extern Pixel	light_color(Colormap, Pixel pixel);
extern void	set_conversion_mode(Buffer *b, ConversionMode mode);

static int	preedit_nothing_count = 0;
static int	status_nothing_count = 0;
static int	ghost_show = 1;

Widget	PEShell, PELabel, PEText;

static void
function_activate( Widget w, XtPointer client, XtPointer call )
{
    HawListCallback	lc = (HawListCallback)call;
    Cardinal	n = 0;
    int	item = lc->item;

    switch ( item ) {
    case 0: hanja( w, lc->event, NULL, &n );
        break;
    case 1: codetable( w, lc->event, NULL, &n );
        break;
    case 2: select_lang( w, lc->event, NULL, &n );
        break;
    case 3: help( w, lc->event, NULL, &n );
        break;
    }
}

void PEWindowMap(unsigned long style)
{
    if ( style & XIMPreeditNothing ) {
        preedit_nothing_count++;
        if ( preedit_nothing_count == 1 ) {
            buffer_initialize( &sys_buffer );
            XtVaSetValues( PEText, HawNdisplayCaret, True, NULL );
            HawTextOutSetString( PEText, "", HawHIGHLIGHT_NORMAL );
        }
    }
    if ( style & XIMStatusNothing ) {
        status_nothing_count++;
        set_conversion_mode(buffer, buffer->conversion_mode);
    }
    if ( style == (XIMPreeditNothing | XIMStatusNothing) ) {
        XMapRaised( XtDisplay(PEShell), XtWindow(PEShell) );
        ghost_show = 1;
    }
}

void PEWindowUnmap( unsigned long style )
{
    if ( style & XIMPreeditNothing ) {
        preedit_nothing_count--;
        if ( preedit_nothing_count == 0 ) {
            HawTextOutSetString( PEText, NULL, HawHIGHLIGHT_NORMAL );
            XtVaSetValues( PEText, HawNdisplayCaret, False, NULL );
        }
    }
    if ( style & XIMStatusNothing ) {
        status_nothing_count--;
        if ( status_nothing_count == 0 )
            set_conversion_mode(buffer, NONE);
    }
    XtUnmapWidget( PEShell );
}

void PEMoveWidget( Widget w )
{
    int	x, y;
    Position	newx, newy;
    unsigned int	width, height;
    Display	*display = XtDisplay(w);
    int	screenno = DefaultScreen(display);
    Window	child;
    XSizeHints	hints;
    long	supplied;
    Dimension	theight, pwidth, pheight;

    XtVaGetValues( PEShell, XtNheight, &theight, NULL );
    XGetWMNormalHints( XtDisplay(w), XtWindow(w), &hints, &supplied );
    XtVaGetValues( w, XtNwidth, &pwidth, XtNheight, &pheight, NULL );
    hints.flags |= PWinGravity;

    XTranslateCoordinates( display, XtWindow(PEShell),
                           DefaultRootWindow(display), 0, (int)theight,
                           &x, &y, &child );
    newx = x;
    newy = y;
    hints.win_gravity = NorthWestGravity;

    width = DisplayWidth( display, screenno );
    height = DisplayHeight( display, screenno );
    if ( x + pwidth > width ) {
        newx -= x + pwidth - width;
        hints.win_gravity = NorthEastGravity;
    }
    if ( y + pheight > height ) {
        newy -= pheight + theight;
        hints.win_gravity = (hints.win_gravity == NorthWestGravity) ?
                            SouthWestGravity : SouthEastGravity;
    }
    XSetWMNormalHints( display, XtWindow(w), &hints );
    XtVaSetValues( w, XtNx, newx, XtNy, newy, NULL );
    XtMapWidget( w );
}

void GhostShow()
{
    if ( ghost_show )
        return;
    ghost_show = 1;
    XMapRaised( XtDisplay(PEShell), XtWindow(PEShell) );
}

void GhostHide()
{
    if ( ! ghost_show )
        return;
    ghost_show = 0;
    XtUnmapWidget( PEShell ); ;
}

#define MAX_BUTTON	4

void hanim_rootwin_init( void )
{
    Widget	PEForm, PEButtonBox;
    Arg	args[20];
    int	n, ii, length;
    unsigned long	lbg;
    Position	xl, yl, xb, yb, xt, yt;
    Dimension	width, border_width, wt, wb;
    Dimension	height, hl, hb, ht;
    char *button_items;

    PEShell = XtVaCreatePopupShell("RootWin",
                                   overrideShellWidgetClass, TopLevel,
                                   // topLevelShellWidgetClass, TopLevel,
                                   // XtNoverrideRedirect, True,
                                   XtNmappedWhenManaged, False,
                                   NULL);
    lbg = light_color( None, default_background );

    /* Create Form */
    PEForm = XtCreateManagedWidget( "PEForm", formWidgetClass,
                                    PEShell, NULL, 0 );

    buffer_initialize( &sys_buffer );
    init_hangul_automata( &sys_buffer.hautomata );

    sys_buffer.prev = NULL;
    sys_buffer.next = NULL;
    sys_buffer.connect = 0;
    sys_buffer.xim_style = XIMPreeditNothing | XIMStatusNothing;
    sys_buffer.conversion_mode = resources.initial_mode;
    sys_buffer.ksc5601 = 0xa1a2;

    /* Label */
    n = 0;
    XtSetArg( args[n], XtNhorizDistance, 5 ); n++;
    XtSetArg( args[n], XtNlabel, resources.hanIM ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainLeft ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    PELabel = XtCreateWidget( "PELabel", hLabelWidgetClass, PEForm, args, n );
    XtManageChild( PELabel );
    sys_buffer.label_widget = PELabel;

    /* Create Text Widget on IM Server  */
    n = 0;
    XtSetArg( args[n], XtNwidth, 600 ); n++;
    XtSetArg( args[n], XtNbackground, lbg ); n++;
    XtSetArg( args[n], XtNhorizDistance, 5 ); n++;
    XtSetArg( args[n], XtNvertDistance, 5 ); n++;
    XtSetArg( args[n], XtNfromHoriz, PELabel ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainRight ); n++;
    XtSetArg( args[n], XtNtop, XawChainTop ); n++;
    XtSetArg( args[n], XtNbottom, XawChainTop ); n++;
    XtSetArg( args[n], XtNresizable, True ); n++;
    XtSetArg( args[n], HawNdisplayCaret, False ); n++;
    PEText = HawCreateTextOut( PEForm, "PEText", args, n );
    XtManageChild( PEText );
    sys_buffer.preedit_widget = PEText;

    /* Button  List ׸ ѷ Box . */
    n = 0;
    XtSetArg( args[n], XtNvertDistance, 5 ); n++;
    XtSetArg( args[n], XtNfromHoriz, PELabel ); n++;
    XtSetArg( args[n], XtNfromVert, PEText ); n++;
    XtSetArg( args[n], XtNleft, XawChainLeft ); n++;
    XtSetArg( args[n], XtNright, XawChainLeft ); n++;
    XtSetArg( args[n], XtNbottom, XawChainBottom ); n++;
    XtSetArg( args[n], XtNtop, XawChainBottom ); n++;
    XtSetArg( args[n], XtNborderWidth, 0 ); n++;
    XtSetArg( args[n], HawNitemBackground, lbg ); n++;
    XtSetArg( args[n], HawNitemCount, MAX_BUTTON ); n++;
    for ( ii = length = 0; ii < MAX_BUTTON; ii++ )
        if ( length < strlen(resources.rootwin_label[ii]) )
            length = strlen(resources.rootwin_label[ii]);
    button_items = XtMalloc( length * MAX_BUTTON + 1 );
    memset( button_items, 0, length * MAX_BUTTON + 1 );
    for ( ii = 0; ii < MAX_BUTTON; ii++ )
        strcpy( button_items + ii * length, resources.rootwin_label[ii] );
    XtSetArg( args[n], HawNitemLength, length * MAX_BUTTON ); n++;
    XtSetArg( args[n], HawNitems, button_items ); n++;
    XtSetArg( args[n], HawNnumRows, MAX_BUTTON ); n++;
    XtSetArg( args[n], HawNvisibleHorizontalItemCount, MAX_BUTTON ); n++;
    XtSetArg( args[n], HawNvisibleVerticalItemCount, 1 ); n++;
    PEButtonBox = HawCreateList( PEForm, "PEButtonBox", args, n );
    XtManageChild( PEButtonBox );
    XtAddCallback( PEButtonBox, HawNunsetCallback,
                   (XtCallbackProc)function_activate,
                   (XtPointer)&sys_buffer );

    /* Create Windows */
    XtUninstallTranslations( PEText );
    XtVaSetValues( PEText, XtNaccelerators, resources.base_accelerators,
                   NULL );
    XtInstallAccelerators( PEText, PEText );
    if ( resources.accelerators ) {
        XtOverrideTranslations( PEText, resources.accelerators );
        XtInstallAccelerators( PEText, PEText );
    }

    XtRealizeWidget( PEShell );

    /* Widget values reset */
    XtVaGetValues( PEText, XtNx, &xt, XtNy, &yt,
                   XtNwidth, &wt, XtNheight, &ht,
                   XtNborderWidth, &border_width, NULL );
    XtVaGetValues( PELabel, XtNx, &xl, XtNy, &yl, XtNheight, &hl, NULL );
    XtVaGetValues( PEButtonBox, XtNx, &xb, XtNy, &yb,
                   XtNwidth, &wb, XtNheight, &hb, NULL );

    height = hl;
    if ( height < hb )	height = hb;
    if ( height < ht )	height = ht;
    XtVaSetValues(PELabel, XtNx, xl,
                  XtNy, yl + (height - hl) / 2, NULL);
    XtVaSetValues(PELabel, XtNx, xb,
                  XtNy, yb + (height - hb) / 2, NULL);

    XtVaGetValues( PEShell, XtNwidth, &width, XtNheight, &height, NULL );
    XtVaSetValues( PEShell, XtNminWidth, width, XtNmaxWidth, width,
                   XtNminHeight, height, XtNmaxHeight, height, NULL );

    XtAddGrab( PEShell, False, False );
}
