/* $Id: buffer.c,v 1.1 2000/02/16 20:44:20 frog Exp $ */

#include	<assert.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<X11/Xlib.h>
#include	<X11/Xlocale.h>
#include	<Haw/Preedit.h>
#include	<Haw/TextOut.h>

#include	"hangul.h"
#include	"hcodec.h"
#include	"hstring.h"

#include	"hanim.h"
#include	"resource.h"
#include	"buffer.h"

static int	count = 0;
static Buffer	*stack[5];
Buffer	sys_buffer;
Buffer	*buffer = &sys_buffer;

static Boolean	flag = False;
static unsigned int	point;
static unsigned int	end_point[2];
static unsigned char	key[2][MAX_KEY];
static unsigned char	str[2][MAX_KEY];
static unsigned int	keyC[2][MAX_KEY];
static hchar_t	hangul[2][MAX_HANGUL];
static unsigned int	hangulC[2][MAX_HANGUL];
static ConversionMode	mode[2][MAX_KEY];
static Boolean	conversion_flag[2][MAX_KEY];
static unsigned int	cursor;

#ifdef DEBUG
void print_buffer(Buffer *b)
{
    int	ii, p = b->point;
    char	str[MAX_HANGUL * 2];

    ii = hcstombs(str, b->hangul, MAX_HANGUL * 2);
    str[ii] = 0;
    printf("p = %d, ep = %d, cursor = %d\n",
           b->point, b->end_point, b->cursor); ;
    printf("keyC[p] = %d, keyC[p+1] = %d\n", b->keyC[p], b->keyC[p + 1]);
    printf("hangulC[p] = %d, hangulC[p+1] = %d\n",
           b->hangulC[p], b->hangulC[p + 1]);
    printf("mode[p] = %d, conversion_flag[p] = %d\n",
           b->mode[p], b->conversion_flag[p]);
    printf("buffer = %s\n", str);
    printf("key = %s\n", buffer->key);
}

void print_buffer_list(void)
{
    Buffer *b = buffer;

    while (b) {
        printf("[prev] = %x, buffer = %x, [next] = %x\n",
               b->prev, b, b->next);
        b = b->next;
    }
}
#endif

/** @name  ó Լ */
//@{

///
/**  Լ  ũ Ʈ ߿ b  
    ̰ ǥ . */
void sort_buffer(Buffer *b)
{
    if (!b->prev)
        return;
    b->prev->next = b->next;
    if (b->next)
        b->next->prev = b->prev;
    b->prev = NULL;
    buffer->prev = b;
    b->next = buffer;
    buffer = b;
}


/// set buffer
/** μ  widget type   Ҹ ã
   ̰  ó . */
Buffer *set_buffer(Widget widget, int type)
{
    Buffer *b;

    if (buffer->state & (ERROR | ADD | DELETE))
        return(buffer);

    if ((type == ButtonPress) || (type == ButtonRelease)) {
        for (b = buffer; b; b = b->next) {
            if ((b->xim_style & XIMStatusNothing) && b->connect) {
                sort_buffer(b);
                return(buffer);
            }
        }
    }
    for (b = buffer; b; b = b->next) {
        if (widget == b->preedit_widget) {
            // we may check connection number of sys_buffer.
            sort_buffer(b);
            return(buffer);
        }
    }
    return(buffer);
}

/// buffer initialize
/** ۸ ʱȭѴ.  Ҵ ޸𸮰 ִٸ
   ̰͵ freeѴ */
void buffer_initialize(Buffer *b)
{
    int	ii;

    b->point = 0;
    b->end_point = 1;
    for (ii = 0; ii < MAX_KEY; ii++) {
        b->key[ii] = '\0';
        b->keyC[ii] = 0;
        b->str[ii] = '\0';
    }
    for (ii = 0; ii < MAX_HANGUL; ii++) {
        b->hangul[ii] = '\0';
        b->hangulC[ii] = 0;
    }
    b->mode[0] = b->conversion_mode;
    b->conversion_flag[0] = False;
    b->cursor = 0;
    b->state = PREEDIT;

    /*  ѱ    ʴ´. */
    clear_hangul_automata(&b->hautomata);
}

/// buffer allocation
/** ο  Ҹ Ҵϰ ʱȭѴ.
   ׸,  Ҹ ũ Ʈ   д. */
Buffer *buffer_alloc(void)
{
    Buffer *b;

    if ((b = (Buffer *)XtMalloc(sizeof(Buffer))) == NULL)
        return(NULL);
    memset(b, 0, sizeof(Buffer));
    buffer_initialize(b);

    /* init_hangul_automata() ѱ   ʱ· Ѵ. */
    init_hangul_automata(&b->hautomata);

    b->ksc5601 = 0xa1a2;
    b->conversion_mode = resources.initial_mode;
    b->next = NULL;
    for ( b->prev = buffer; b->prev->next;
          b->prev = b->prev->next );
    b->prev->next = b;
    return(b);
}

/// buffer free
/** ۿ Ҵ ޸𸮸 Ѵ. */
void buffer_free(Buffer *b)
{
    if (b->prev) {
        b->prev->next = b->next;
        if (b->next)
            b->next->prev = b->prev;
    } else {
        buffer = buffer->next;
        buffer->prev = 0;
    }
    XFree(b);
}

/// push buffer
/** ý  Ʈʹ   b 
   ÿ (push) ξٰ, (pop)  ִ.
       ϴٰ ϴ   ·
   ǵư  Ѵ. */
void push_buffer(Buffer *b)
{
    if (count == 5) {
        fputs("Irregular push_buffer access.", stderr);
        exit(1);
    }
    if ((stack[count] = (Buffer *)XtMalloc(sizeof(Buffer))) == NULL) {
        fputs("not enough memory", stderr);
        exit(1);
    }
    *stack[count++] = *b;
}

/// pop buffer
/** ǪǾ ϴ ۵      ۷ Ҵ. */
void pop_buffer(Buffer *b)
{
    if (count == 0) {
        fputs("Irregular pop_buffer access.", stderr);
        exit(1);
    }
    *b = *stack[--count];
    XFree(stack[count]);
}

///
/** ۸ Ѵ. */
void register_buffer(Buffer *b)
{
    int	ii;

    flag = True;
    end_point[0] = b->end_point;
    for (ii = 0; ii < MAX_KEY; ii++) {
        key[0][ii] = b->key[ii];
        str[0][ii] = b->str[ii];
        keyC[0][ii] = b->keyC[ii];
        mode[0][ii] = b->mode[ii];
        conversion_flag[0][ii] = b->conversion_flag[ii];
    }
    for (ii = 0; ii < MAX_HANGUL; ii++) {
        hangul[0][ii] = b->hangul[ii];
        hangulC[0][ii] = b->hangulC[ii];
    }
}

/// save buffer
/** ۸ Ѵ. */
void save_buffer(Buffer *b)
{
    int	ii;

    point = b->point;
    end_point[1] = b->end_point;
    for (ii = 0; ii < MAX_KEY; ii++) {
        key[1][ii] = b->key[ii];
        str[1][ii] = b->str[ii];
        keyC[1][ii] = b->keyC[ii];
        mode[1][ii] = b->mode[ii];
        conversion_flag[1][ii] = b->conversion_flag[ii];
    }
    for (ii = 0; ii < MAX_HANGUL; ii++) {
        hangul[1][ii] = b->hangul[ii];
        hangulC[1][ii] = b->hangulC[ii];
    }
    cursor = b->cursor;
}

///
/** ۸ εѴ. */
void load_buffer(Buffer *b)
{
    int	ii;

    b->point = point;
    b->end_point = end_point[1];
    for (ii = 0; ii < MAX_KEY; ii++) {
        b->key[ii] = key[1][ii];
        b->str[ii] = str[1][ii];
        b->keyC[ii] = keyC[1][ii];
        b->mode[ii] = mode[1][ii];
        b->conversion_flag[ii] = conversion_flag[1][ii];
    }
    for (ii = 0; ii < MAX_HANGUL; ii++) {
        b->hangul[ii] = hangul[1][ii];
        b->hangulC[ii] = hangulC[1][ii];
    }
    b->cursor = cursor;
}

/// ۿ ڿ 
/**  b ۼ key Է  str Ѵ. */
Boolean insert_buffer( Buffer *b, unsigned char key,
    unsigned char str, ConversionMode mode )
{
    unsigned int	p = b->point;
    unsigned int	ep = b->end_point;
    int	ii;

    if ((ep == MAX_KEY - 1) || (b->keyC[ep] == (MAX_KEY - 1))) {
        load_buffer(b);
        return(False);
    }

    if (b->keyC[p] == b->keyC[p + 1]) {
        if (p + 1 != ep) {
            fputs("buffer error", stderr);
            done(0);
        }
        b->key[b->keyC[p]] = key;
        b->str[b->keyC[p]] = str;
        b->keyC[ep + 1] = ++b->keyC[ep];
        b->hangulC[ep + 1] = b->hangulC[ep];
        b->mode[p] = mode;
    } else if (b->conversion_flag[p] ||
	(b->keyC[p] != b->keyC[p + 1] && b->hangulC[p] == b->cursor) ) {
        for (ii = b->keyC[ep] - 1; ii >= 0 && ii >= b->keyC[p]; ii--) {
            b->key[ii + 1] = b->key[ii];
            b->str[ii + 1] = b->str[ii];
        }
        b->key[b->keyC[p]] = key;
        b->str[b->keyC[p]] = str;

        for (ii = ep; ii >= 0 && ii >= p; ii--) {
            b->keyC[ii + 1] = b->keyC[ii] + 1;
            b->hangulC[ii + 1] = b->hangulC[ii];
            b->mode[ii + 1] = b->mode[ii];
            b->conversion_flag[ii + 1] = b->conversion_flag[ii];
        }
        b->mode[p] = mode;
        b->end_point++;
    } else if (b->mode[p] == mode) {
        for (ii = b->keyC[ep] - 1; ii >= 0 && ii >= b->keyC[p + 1]; ii--) {
            b->key[ii + 1] = b->key[ii];
            b->str[ii + 1] = b->str[ii];
        }
        b->key[b->keyC[p + 1]] = key;
        b->str[b->keyC[p + 1]] = str;

        for (ii = ep + 1; ii > p; ii--)
            b->keyC[ii]++;
    } else {
        p = ++b->point;
        for (ii = b->keyC[ep] - 1; ii >= 0 && ii >= b->keyC[p]; ii--) {
            b->key[ii + 1] = b->key[ii];
            b->str[ii + 1] = b->str[ii];
        }
        b->key[b->keyC[p]] = key;
        b->str[b->keyC[p]] = str;

        for (ii = ep; ii >= 0 && ii >= p; ii--) {
            b->keyC[ii + 1] = b->keyC[ii] + 1;
            b->hangulC[ii + 1] = b->hangulC[ii];
            b->mode[ii + 1] = b->mode[ii];
            b->conversion_flag[ii + 1] = b->conversion_flag[ii];
        }
        b->mode[p] = mode;
        b->end_point++;
    }
    return(True);
}

/// ۿ  
/** ѱ   ̶ ڼ  ϸ ̹ Է
       Ѵ. */
void delete_character_buffer(Buffer *b, Widget w,
    hchar_t *hangul, ConversionMode mode)
{
    unsigned int p, ep;
    hchar_t ss[MAX_HANGUL];
    unsigned int len;
    int	ii;
    int	start, end;

    p = b->point;
    ep = b->end_point;
    if (hangul != NULL) {
        if (b->xim_style & XIMPreeditNothing) {
            start = b->hangulC[p];
            end = b->hangulC[p + 1];
            HawTextOutSetCursorPosition(w, start);
#if 0
            hcscpy(ss, &b->hangul[end]);
            hcscpy(&b->hangul[start], hangul);
            hcscat(b->hangul, ss);
#else
            hcscpy(ss, b->hangul + end);
            hcscpy(b->hangul + start, hangul);
            hcscpy(b->hangul + start + hcslen(hangul), ss);
#endif
            len = hcslen(hangul) - (b->hangulC[p + 1] - b->hangulC[p]);
            for (ii = p + 1; ii <= ep; ii++)
                b->hangulC[ii] += len;
            HawTextOutReplaceHcs(w, start, end, hangul);
        } else if (b->xim_style & (XIMPreeditPosition | XIMPreeditArea)) {
            if (!(b->state & EDIT)) {
                HawPreeditReplaceHcs(w, b->hangulC[p], b->hangulC[p + 1], hangul);
            }
            /* ȯ ڿ   ʿϴ. */
            hcscpy(ss, &b->hangul[b->hangulC[p + 1]]);
            hcscpy(&b->hangul[b->hangulC[p]], hangul);
            hcscat(b->hangul, ss);
            len = hcslen(hangul) - (b->hangulC[p + 1] - b->hangulC[p]);
            for (ii = p + 1; ii <= ep; ii++)
                b->hangulC[ii] += len;
        }
    } else {
        if (b->xim_style & XIMPreeditNothing) {
            start = b->hangulC[p + 1] - 1;
            end = b->hangulC[p + 1];
            hcscpy(ss, &b->hangul[b->hangulC[p + 1]]);
            hcscpy(&b->hangul[b->hangulC[p + 1]-1], ss);
            for (ii = p + 1; ii <= ep; ii++)
                b->hangulC[ii]--;
            HawTextOutReplace(w, start, end, NULL);
        } else if (b->xim_style & (XIMPreeditPosition | XIMPreeditArea)) {
            if (!(b->state & EDIT)) {
                HawPreeditReplace(w, b->hangulC[p + 1] - 1, b->hangulC[p + 1], NULL);
            }
#if 0
            else {
                for (ii = 0, start = 0; ii < b->henkan->point; ii++)
                    start += b->henkan->bun[ii].destlen;
                HawPreeditReplace( w, b->hangulC[p + 1] - 1 + start,
                                   b->hangulC[p + 1] + start, NULL );
            }
#endif
            hcscpy(ss, &b->hangul[b->hangulC[p + 1]]);
            hcscpy(&b->hangul[b->hangulC[p + 1]-1], ss);
            for (ii = p + 1; ii <= ep; ii++)
                b->hangulC[ii]--;
        }
    }

    for (ii = b->keyC[p + 1]; ii <= b->keyC[ep]; ii++) {
        b->key[ii - 1] = b->key[ii];
        b->str[ii - 1] = b->str[ii];
    }
    for (ii = p + 1; ii <= ep; ii++)
        b->keyC[ii]--;

    b->conversion_flag[p] = False;
    b->cursor = b->hangulC[p + 1];

#ifdef DEBUG
    if (resources.debug & 0x08) {
        printf("====== delete_char_buffer ======\n");
        print_buffer(b);
    }
#endif
}

///  ܾ  ƾ
/**  Է  ѱ̶ ܾ  Ѵ. */
void delete_word_buffer(Buffer *b, Widget w)
{
    unsigned int p = b->point;
    unsigned int ep = b->end_point;
    unsigned char ss[MAX_KEY];
    hchar_t hh[MAX_HANGUL];
    unsigned int lk, ly;
    int	ii;

    assert(b->keyC[p + 1] <= b->keyC[ep]);

    if ( b->xim_style & XIMPreeditNothing ||
         (b->state & ADD) ) {
        HawTextOutSetCursorPosition(w, b->hangulC[p]);
        HawTextOutReplace(w, b->hangulC[p], b->hangulC[p + 1], NULL);
    } else if (b->xim_style & (XIMPreeditPosition | XIMPreeditArea)) {
        if (!(b->state & EDIT)) {
            HawPreeditReplace(b->preedit_widget, b->hangulC[p],
		b->hangulC[p + 1], NULL );
        }
        /* ȯ ڿ   ʿϴ. */
    }
    b->cursor = b->hangulC[p];

    strcpy((char *)ss, (char *)&b->key[b->keyC[p + 1]]);
    strcpy((char *)&b->key[b->keyC[p]], (char *)ss);

    if ((lk = b->keyC[ep] - b->keyC[p + 1]) != 0) {
        memcpy(ss, &b->str[b->keyC[p + 1]], lk);
        memcpy(&b->str[b->keyC[p]], ss, lk);
    } else
        b->str[b->keyC[p]] = '\0';

    hcscpy(hh, &b->hangul[b->hangulC[p + 1]]);
    hcscpy(&b->hangul[b->hangulC[p]], hh);

    lk = b->keyC[p + 1] - b->keyC[p];
    ly = b->hangulC[p + 1] - b->hangulC[p];
    for (ii = p; ii < ep; ii++) {
        b->keyC[ii] = b->keyC[ii + 1] - lk;
        b->hangulC[ii] = b->hangulC[ii + 1] - ly;
        b->mode[ii] = b->mode[ii + 1];
        b->conversion_flag[ii] = b->conversion_flag[ii + 1];
    }
    b->keyC[ep] -= lk;
    b->hangulC[ep] -= ly;

    if (ep > 2 && b->keyC[ep - 2] == b->keyC[ep])
        b->end_point--;
}

/// ѱ ġȯ ƾ
/** 丶Ÿ ѱ Է ̶ ̹ ۿ  ִ ڿ
  ġȯϿ  찡 ִ.  ȯ  Լ ؾ ϴ
  κ̴. */
Boolean replace_hangul(Buffer *b, hchar_t *hangul)
{
    unsigned int p = b->point;
    unsigned int ep = b->end_point;
    hchar_t hh[MAX_HANGUL];
    int	length;
    int	ii;

    hcscpy(hh, &b->hangul[b->hangulC[p + 1]]);
    hcscpy(&b->hangul[b->hangulC[p]], hangul);
    hcscat(b->hangul, hh);

    length = (b->hangulC[p + 1] - b->hangulC[p]) - hcslen(hangul);
    if ((b->hangulC[ep] - length) >= MAX_HANGUL) {
        load_buffer(b);
        return(False);
    }
    for (ii = p + 1; ii <= ep; ii++)
        b->hangulC[ii] -= length;
    b->hangulC[ep + 1] = b->hangulC[ep];
#ifdef DEBUG
    if (resources.debug & 0x08) {
        printf("========= replace_hangul =========\n");
        print_buffer(b);
    }
#endif
    return(True);
}

/// ѱ ߰ ƾ
/**  Է  ۿ ѱ ߰Ѵ. */
Boolean add_hangul(Buffer *b, hchar_t *hangul)
{
    unsigned int	p = b->point;
    unsigned int	ep = b->end_point;
    hchar_t	hh[MAX_HANGUL];
    int	length;
    int	ii;

    if ((b->hangulC[ep] + hcslen(hangul)) >= MAX_HANGUL) {
        load_buffer(b);
        return(False);
    }

    hcscpy(hh, &b->hangul[b->hangulC[p + 1]]);
    hcscpy(&b->hangul[b->hangulC[p + 1]], hangul);
    hcscat(b->hangul, hh);

    length = hcslen(hangul);
    for (ii = p + 1; ii <= ep; ii++)
        b->hangulC[ii] += length;
    b->hangulC[ep + 1] = b->hangulC[ep];
#ifdef DEBUG
    if (resources.debug & 0x08) {
        printf("========= add_hangul =========\n");
        print_buffer(b);
    }
#endif
    return(True);
}

/// ѱ Է ϼ κ
/**   Է̶ ̰ ϼŰ
   ͷ ̵Ѵ. */
Boolean complete_hangul(Buffer *b)
{
    unsigned int	p = b->point;
    unsigned int	ep = b->end_point;

    if (b->conversion_flag[p])
        return False;
    b->conversion_flag[b->point++] = True;

#if 0
    b->keyC[b->point] -= 1;
    b->mode[b->point] = mode;
    if (b->point == b->end_point) {
        b->end_point++;
        /* insert_buffer() ٽ 1 ϰ ȴ */
        b->keyC[b->end_point] = b->keyC[b->point] + 1;
    }
#else
    if (b->point == ep) {
        b->end_point++;
        b->keyC[b->end_point] = b->keyC[b->point] + 1;
    }
#endif
#ifdef DEBUG
    if (resources.debug & 0x08) {
        printf("========= complete_hangul =========\n");
        print_buffer(b);
    }
#endif
    return(True);
}

//@}
