/* $Id: hanim.c,v 1.28 2000/01/11 01:28:45 frog Exp $ */

/*
 * hanIM: X Window System Version 11
 *	  Input Method Server
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<signal.h>
#ifndef WIN32
# include	<sys/param.h>
# include	<unistd.h>
# include	<pwd.h>
#endif    /* !WIN32 */
#include	<X11/Xlib.h>
#include	<X11/Xlocale.h>

#include	"hanim.h"
#include	"version.h"
#include	"hcodec.h"
#include	"hstring.h"
#include	"romandic.h"
#include	"XuiIcon.h"
#include	"XuiRootWin.h"
#include	"protowrap.h"
#include	"Ximcp.h"
#include	"Ximp.h"
#include	"resource.h"
#include	"preference.h"

#ifdef SVR4
#include	<sys/utsname.h>
#include	<sys/systeminfo.h>
#define sigmask(n)	((unsigned long)1 << ((n) - 1))
#endif    /* SVR4 */

extern Widget TopLevel;
extern XtAppContext hanim_window_init(int, char **);

char	*program_name;
char	**command_vector;
char	home_directory[BUFSIZ];
char	sys_directory[BUFSIZ];

LcCtype	user_locale;
char	*locale_name;
Endian	endian;
static Boolean init_done = False;

XtAppContext	app;

static void usage(void)
{
    puts("Hangul IM Server");
    printf( "%s %d.%d.%d\n", program_name,
            MAJOR_VERSION, MINOR_VERSION, BUGFIX_VERSION );
    puts("Copyright (c) 1997-1998 Seo, Young-Jin & Mizi Research Inc.");
    puts("\nusage:");
    printf("        %s [-options ...] &", program_name);
    puts("\nwhere options include:");
    puts("    -help                        print out this message");
    puts("    -display displayname         X server to contact");
    puts("    -geometry geom               size (in characters) and position");
    puts("    -bg color                    background color");
    puts("    -fg color                    foreground color");
    puts("    -bd color                    border color");
    puts("    -bw number                   border width in pixels");
    puts("    -iconic                      start iconic");
    puts("    -name string                 client instance, icon, and title strings");
    puts("    -reverse                     reverse of foreground color");
    puts("    -/+rv                        turn on/off reverse video");
    puts("    -/+synchronous               turn on/off synchronous mode");
    puts("    -title string                title string");
    puts("    -no_auto_replace             turn off auto replace mode");
#if defined(USE_XIMCP)
    puts("    -/+ximcp                     turn on/off xim common protocol");
#endif
#ifdef USE_XIMP
    puts("    -ximp version                ximp version");
#endif
    puts("    -xrm resourcestring          additional resource specifications");
    exit(1);
}

static void initialize(int argc, char *argv[])
{
    int	ii;
    char *ss, *tt;
#ifndef WIN32
    struct passwd *pwd;
#endif    /* !WIN32 */

    /* set program name */
    if ((ss = strrchr(argv[0], DIRSEP)) == 0)
        program_name = argv[0];
    else
        program_name = ss + 1;

    /* get user locale */
    locale_name = setlocale(LC_ALL, "");
    if (locale_name == 0 || (strncmp(locale_name, "ko", 2) &&
	strncmp(locale_name, "Ko", 2))) {
        fprintf( stderr, "%s error: locale could not be set.\n",
	    program_name );
        exit(1);
    }

    user_locale = LC_CTYPE_EUC;
    /* set user name and home directory */
    tt = getenv("HOME");
#ifndef WIN32
    ss = getlogin();
    setpwent();
    pwd = ss ? getpwnam(ss) : getpwuid(getuid());
    if (pwd && !tt)
        tt = pwd->pw_dir;
    if (tt) strcpy( home_directory, tt);
    endpwent();
#else /* UNIX */
    if (tt) strcpy(home_directory, tt);
    else strcpy(home_directory, "c:\\");
#endif

    /* set hanim library directory */
    if (access("/usr/X11R6/lib/X11/hanIM", 0) == 0)
        strcpy(sys_directory, "/usr/X11R6/lib/X11/hanIM");
    else if (access("/usr/openwin/lib/X11/hanIM", 0) == 0)
        strcpy(sys_directory, "/usr/openwin/lib/X11/hanIM");
    else
        strcpy(sys_directory, "/usr/lib/X11/hanIM");

    /* argument previous check */
    for (ii = 1; ii < argc; ++ii)
        if (!strcmp("-help", argv[ii]))
            usage();

    /* endian */
    ii = 1;
    endian = *(char *) & ii ? LITTLEENDIAN : BIGENDIAN;

    /* copy command vector */
    command_vector = (char **)XtMalloc(sizeof(char **) * (argc + 1));
    for (ii = 0; ii < argc; ii++)
        command_vector[ii] = argv[ii];
    command_vector[argc] = NULL;
}

void done(int dummy)
{
    if (init_done) {
#ifdef USE_XIMP
        ximp_all_end();
#endif
        XSync(XtDisplay(TopLevel), False);
        XtCloseDisplay(XtDisplay(TopLevel));
    }
    fprintf(stderr, "\n\r%s done.\n\r", program_name);
    exit(0);
}

#ifndef WIN32
static void suspend(int sig)
{
#ifdef USE_XIMP
    ximp_all_end();
#endif

    XSync(XtDisplay(TopLevel), False);

    signal(SIGTSTP, SIG_DFL);
#ifndef	SVR4
    sigsetmask(sigblock( 0 ) & ~sigmask(SIGTSTP));
    sigblock(sigmask(SIGTSTP));
    signal(SIGTSTP, suspend);
#else
    sighold(SIGTSTP);
    sigset(SIGTSTP, suspend);
    sigrelse(SIGTSTP);
#endif
}
#endif    /* !WIN32 */

#ifndef	DEBUG
static void makecore(int sig)
{
    if (init_done) {
#ifdef USE_XIMP
        ximp_all_end();
#endif
        XSync(XtDisplay(TopLevel), False);
    }
    fprintf(stderr, "\n\r%s terminate.\n\r", program_name);
    abort();
}
#endif    /* DEBUG */

static void signal_initialize()
{
#ifndef WIN32
# ifdef SVR4
    sigignore(SIGTSTP);   			/* CTRL z */
# else
    signal(SIGTSTP, SIG_IGN);   		/* CTRL z */
# endif    /* SVR4 */
#endif    /* !WIN32 */

#ifdef SVR4
    sigset(SIGTERM, done);
    sigset(SIGINT, done);
#ifndef	DEBUG
    sigset(SIGSEGV, makecore);
    sigset(SIGFPE, makecore);
#endif   	/* DEBUG */
#else	/* !SVR4 */
    signal(SIGTERM, done);
    signal(SIGINT, done);
#ifndef	DEBUG
    signal(SIGSEGV, makecore);
    signal(SIGFPE, makecore);
#endif   	/* DEBUG */
#endif   	/* SVR4 */
}

static Boolean event_pre_hook(XEvent *event)
{
    switch (event->type) {
    case KeyPress:
#ifdef USE_XIMP
        if (XimpFilterEvent( (XKeyEvent *)event ))
            return(True);
#endif
        /*   ָ   ȵȴ */
        event->xkey.subwindow = event->xany.window;
        if (PreeditChangeKeyDirection( event ))
            break;
        break;

    case DestroyNotify:
        if (ChangeDestroyNotify( event ))
            return(True);
        break;

    case Expose:
        if (event->xexpose.count == 0)
            if (ChangeExpose( event ))
                break;
        break;

    case MapNotify:
    case UnmapNotify:
        if (IconFilterEvent( event ))
            return(True);
        break;

    case ClientMessage:
        ChangeClientMessage(event);
        break;
    default:
        break;
    }
    return(False);
}

static void event_post_hook(XEvent *event)
{}

int main(int argc, char *argv[])
{
    XEvent event;

    initialize(argc, argv);
    // LATER  read_hanim_rc();

    app = hanim_window_init(argc, argv);

    signal_initialize();
    init_roman_tables();

#if defined(USE_XIMCP)
    ximcp_initialize();
#endif
#ifdef USE_XIMP
    ximp_initialize();
#endif

    init_done = True;
    XBell(XtDisplay(TopLevel), 0);
#ifndef WIN32
    signal(SIGTSTP, suspend);   		/* CTRL z */
#else /* WIN32 */
    XIconifyWindow(XtDisplay(TopLevel), XtWindow(TopLevel),
                   DefaultScreen(XtDisplay(TopLevel)));
#endif    /* !WIN32 */
    XFlush(XtDisplay(TopLevel));

    /* loop */
    for (; ; ) {
        XtAppNextEvent(app, &event);
        if (event_pre_hook( &event ))
            continue;
        XtDispatchEvent(&event);
        event_post_hook(&event);
    }
    XtDestroyApplicationContext(app);
    return 0;
}

#if defined(DEBUG) && defined(linux)
/* Linux ׽Ʈ Ǿ. */
int get_memory_usage()
{
    FILE	*fp;
    char	buf[256];
    int	size = 0;
    sprintf(buf, "/proc/%d/status", getpid());
    if ((fp = fopen(buf, "r")) == NULL)
        return - 1;
    while (fgets(buf, 255, fp))
        if (strncmp(buf, "VmSize:", 7) == 0) {
            sscanf(buf + 7, "%d", &size);
            break;
        }

    fclose(fp);
    return size;
}
#endif
