/* $Id: hanim.h,v 1.30 2000/01/11 01:28:45 frog Exp $ */

#ifndef _HANIM_H_
#define	_HANIM_H_

#include <X11/Intrinsic.h>

#include "hcodec.h"
#include "hangul.h"

#define CLASS_NAME	"HanIM"
#define HANIMUSRDIR	".hanIM"

#define MAX_KEY		512
#define MAX_HANGUL	256
#define MAX_HANJA	256

#ifndef	MIN
# define MIN( a, b ) (((a) < (b)) ? (a) : (b))
#endif

typedef enum {
    XIM_COMMON_PROTOCOL,
    XIM_PROTOCOL
} Protocol;

typedef enum {
    BIGENDIAN	= 0x42,
    LITTLEENDIAN	= 0x6c
} Endian;

/** LcCtype CharacterCode   ,
    ϱ    Ͽ Ѵ.
   LcCtype ڰ ȯ溯  ̰,
   CharacterCode ø̼ ޱ⸦ ϴ ڵ̴. */

typedef enum {
    LC_CTYPE_EUC,
    LC_CTYPE_KSSM,
    LC_CTYPE_UNICODE
} LcCtype;

typedef enum {
    _EUCKR,
    _KSSM,
    _UNICODE,
    _UTF8,
    _KSC5601
} CharacterCode;

#define	LANGMASK	0xFF00

typedef enum {
    HANGUL	= 0x0100,
    ENGLISH	= 0x0200,
    JAPANESE	= 0x0400,
    CYRILLIC	= 0x0800
} InputLanguage;

typedef enum {
    HANGUL3	= 0x0001 | HANGUL,
    HANGUL2	= 0x0002 | HANGUL
#if 0 /* LATER */
              HAN_ROMAN	= 0x0004 | HANGUL,
    /* no combination of consonant and voewl chars */
    NC_HANGUL3	= 0x0081 | HANGUL,
    NC_HANGUL2	= 0x0082 | HANGUL,
    NC_HAN_ROMAN	= 0x0084 | HANGUL
#endif    /* LATER */
} HangulKeyboard;

typedef enum {
    QWERTY	= 0x0001 | ENGLISH,
    DVORAK	= 0x0002 | ENGLISH
} EnglishKeyboard;

typedef enum {
    HIRAGANA	= 0x0001 | JAPANESE,
    KATAKANA	= 0x0002 | JAPANESE,
    HIRAGANA_ROMAN	= 0x0004 | JAPANESE,
    KATAKANA_ROMAN	= 0x0008 | JAPANESE,
    HIRAGANA_HANGUL	= 0x0010 | JAPANESE,
    KATAKANA_HANGUL	= 0x0020 | JAPANESE
} JapaneseKeyboard;

/* ͸ ϱ  ̷ Ѵ. LATER  */
typedef union {
    HangulKeyboard	hangul;
    EnglishKeyboard	english;
} LangKeyboard;

/*  Ǵ ʿ  δ.  LATER */
typedef enum {
    HANGUL_3BULSIK	= HANGUL3,
    HANGUL_2BULSIK	= HANGUL2,
#if 0 /* LATER */
    HANGUL_HAN_ROMAN	= HAN_ROMAN,
#endif    /* LATER */

    ENGLISH_QWERTY	= QWERTY,
    ENGLISH_DVORAK	= DVORAK,

    JAPANESE_HIRAGANA	= HIRAGANA,
    JAPANESE_KATAKANA	= KATAKANA,
    JAPANESE_ENG_HIRAGANA	= HIRAGANA_ROMAN,
    JAPANESE_ENG_KATAKANA	= KATAKANA_ROMAN,
    JAPANESE_HAN_HIRAGANA	= HIRAGANA_HANGUL,
    JAPANESE_HAN_KATAKANA	= KATAKANA_HANGUL
    /* CYRILLIC, */
} ConversionMode;

/** @name State 
 
   Է  ¸ Ÿ  Ѵ.
   ¿  ٸ ó ؾ ϹǷ ſ ߿ϴ.
   ڿ (HanIM) action() μ ϴ
  ̱⵵ ϴ.
  */
typedef enum {
    NONE	= 0x0000,
    /// ʱȭ 
    INIT	= 0x0001,
    ///     丶Ÿ  Է ̴.
    PREEDIT	= 0x0002,
    ///  ȯ 
    HANJA	= 0x0004,
    /// 뱸   ϱ    ̴.
    EDIT	= 0x0008,
    ///    ߰ϰų ̴.
    ADD	= 0x0010,
    ///    ׸  ̴.
    DELETE	= 0x0020,
    /// ڵǥ Է ̴.
    CODE	= 0x0080,
    /// Է   ߻ ̴.
    ERROR	= 0x0100
} State;

/*@name  ÷ */
enum {
    DBG_ACTION	= 0x0001,
    DBG_XIMP	= 0x0002,
    DBG_XIM	= 0x0004,
    DBG_EDIT	= 0x0010,
    DBG_ALL	= 0xffff
};

typedef enum {
    COMMIT_CHAR, COMMIT_WORD, COMMIT_LINE
} CommitMethod;

enum {
    STATIC_CONTROL, DYNAMIC_CONTROL
};

/** @name Ÿ global variables */
//@{
extern char	*program_name;
extern char	**command_vector;
extern char	home_directory[];
extern char	sys_directory[];
extern LcCtype	user_locale;
///  hanIM   ȣƮ Ʈ 
extern Endian	endian;
/// XtAppContext - X    ó  Ѵ.
extern XtAppContext	app;
/// TopLevel ǰ
extern Widget	TopLevel;
/// default    
extern Pixel	default_foreground, default_background;
/// default ۲
extern XFontSet	gFontSet;
//@}

#ifdef WIN32
# define DIRSEP '\\'
# define STRCASECMP stricmp
# define STRNCASECMP strnicmp
#else
# define DIRSEP '/'
# define STRCASECMP strcasecmp
# define STRNCASECMP strncasecmp
#endif /* WIN32 */

#endif /* _HANIM_H_ */
