/* $Id: hanjadic.c,v 1.9 2000/01/11 01:28:45 frog Exp $ */

#include	<stdlib.h>
#include	<assert.h>

#include	"hanim.h"
#include	"hcodec.h"
#include	"hstring.h"

typedef struct {
    hchar_t	hh;
    int	offset;
} Hangul2Hanja;

Hangul2Hanja HanjaTbl[] = {
    { 0x8861, 0 },    /*  */
    { 0x8862, 29 },    /*  */
    { 0x8865, 40 },    /*  */
    { 0x8869, 64 },    /*  */
    { 0x8871, 74 },    /*  */
    { 0x8873, 94 },    /*  */
    { 0x8877, 100 },    /*  */
    { 0x8881, 124 },    /*  */
    { 0x8882, 144 },    /*  */
    { 0x8897, 146 },    /*  */
    { 0x88A2, 150 },    /*  */
    { 0x88E1, 151 },    /*  */
    { 0x88E5, 168 },    /*  */
    { 0x88E9, 180 },    /*  */
    { 0x88F1, 184 },    /*  */
    { 0x88F3, 191 },    /*  */
    { 0x8941, 194 },    /*  */
    { 0x8962, 197 },    /*  */
    { 0x8965, 204 },    /*  */
    { 0x8969, 215 },    /*  */
    { 0x8971, 221 },    /*  */
    { 0x8977, 227 },    /*  */
    { 0x8981, 272 },    /*  */
    { 0x89A1, 296 },    /*  */
    { 0x89A2, 335 },    /*  */
    { 0x89A5, 342 },    /*  */
    { 0x89A9, 352 },    /*  */
    { 0x89B7, 355 },    /*  */
    { 0x89B8, 371 },    /*  ksonly */
    { 0x89C1, 372 },    /*  */
    { 0x89C2, 384 },    /*  */
    { 0x89C5, 388 },    /*  */
    { 0x89C9, 405 },    /*  */
    { 0x89D7, 409 },    /*  */
    { 0x89E1, 422 },    /*  */
    { 0x8A41, 425 },    /*  */
    { 0x8A57, 434 },    /*  */
    { 0x8A61, 438 },    /*  */
    { 0x8A81, 463 },    /*  */
    { 0x8A82, 517 },    /*  */
    { 0x8A85, 523 },    /*  */
    { 0x8A89, 529 },    /*  */
    { 0x8A97, 533 },    /*  */
    { 0x8AA5, 539 },    /*  */
    { 0x8AA9, 549 },    /*  */
    { 0x8AC1, 554 },    /*  */
    { 0x8AE1, 560 },    /*  */
    { 0x8B41, 566 },    /*  */
    { 0x8B45, 581 },    /*  */
    { 0x8B62, 589 },    /*  */
    { 0x8B65, 596 },    /*  */
    { 0x8B71, 612 },    /*  */
    { 0x8B73, 626 },    /*  */
    { 0x8B77, 633 },    /*  */
    { 0x8BA1, 637 },    /*  */
    { 0x8BA5, 701 },    /*  */
    { 0x8BA9, 702 },    /*  */
    { 0x8BB1, 706 },    /*  ksonly */
    { 0x8FA2, 707 },    /*  ksonly */
    { 0x9061, 708 },    /*  */
    { 0x9062, 723 },    /*  */
    { 0x9065, 731 },    /*  */
    { 0x9069, 740 },    /*  */
    { 0x9071, 742 },    /*  */
    { 0x9073, 751 },    /*  */
    { 0x9077, 756 },    /*  */
    { 0x9081, 763 },    /*  */
    { 0x9097, 769 },    /*  ksonly */
    { 0x9161, 770 },    /*  */
    { 0x9165, 771 },    /*  */
    { 0x9171, 774 },    /*  */
    { 0x9177, 778 },    /*  */
    { 0x91A1, 780 },    /*  */
    { 0x91A2, 798 },    /*  */
    { 0x91A5, 804 },    /*  ksonly */
    { 0x91B7, 805 },    /*  */
    { 0x9241, 812 },    /*  */
    { 0x9261, 818 },    /*  */
    { 0x9281, 819 },    /*  */
    { 0x9285, 827 },    /*  */
    { 0x9289, 828 },    /*  */
    { 0x9341, 829 },    /*  */
    { 0x9362, 831 },    /*  ksonly */
    { 0x9371, 833 },    /*  ksonly */
    { 0x9377, 834 },    /*  ksonly */
    { 0x93A1, 840 },    /*  */
    { 0x93A2, 842 },    /*  */
    { 0x9461, 844 },    /*  */
    { 0x9465, 846 },    /*  */
    { 0x9469, 866 },    /*  */
    { 0x9471, 871 },    /*  */
    { 0x9473, 888 },    /*  */
    { 0x9477, 893 },    /*  */
    { 0x9481, 904 },    /*  */
    { 0x9482, 920 },    /*  ksonly */
    { 0x94E2, 921 },    /*  */
    { 0x95A1, 923 },    /*  */
    { 0x95A2, 963 },    /*  */
    { 0x95A5, 973 },    /*  */
    { 0x95A9, 983 },    /*  */
    { 0x95B7, 985 },    /*  */
    { 0x9681, 1002 },    /*  */
    { 0x9685, 1013 },    /*  */
    { 0x9762, 1019 },    /*  */
    { 0x9777, 1020 },    /*  */
    { 0x9C61, 1029 },    /*  */
    { 0x9C62, 1038 },    /*  */
    { 0x9C65, 1047 },    /*  */
    { 0x9C69, 1056 },    /*  */
    { 0x9C71, 1058 },    /*  */
    { 0x9C73, 1068 },    /*  */
    { 0x9C77, 1071 },    /*  */
    { 0x9C81, 1079 },    /*  */
    { 0x9CA2, 1084 },    /*  */
    { 0x9CB7, 1086 },    /*  */
    { 0x9D61, 1099 },    /*  */
    { 0x9D62, 1117 },    /*  */
    { 0x9D65, 1124 },    /*  */
    { 0x9D69, 1136 },    /*  */
    { 0x9D71, 1142 },    /*  */
    { 0x9D73, 1147 },    /*  */
    { 0x9D77, 1148 },    /*  */
    { 0x9D81, 1166 },    /*  */
    { 0x9DA1, 1171 },    /*  */
    { 0x9DA2, 1189 },    /*  */
    { 0x9DA5, 1196 },    /*  */
    { 0x9DB7, 1197 },    /*  */
    { 0x9E41, 1204 },    /*  */
    { 0x9E61, 1212 },    /*  */
    { 0x9E77, 1224 },    /*  */
    { 0x9E81, 1225 },    /*  */
    { 0x9F41, 1238 },    /*  */
    { 0x9F42, 1252 },    /*  */
    { 0x9F45, 1255 },    /*  */
    { 0x9F49, 1261 },    /*  */
    { 0x9F57, 1265 },    /*  */
    { 0x9F62, 1266 },    /*  */
    { 0x9F71, 1268 },    /*  */
    { 0x9F77, 1269 },    /*  */
    { 0x9FA1, 1275 },    /*  */
    { 0x9FA5, 1301 },    /*  */
    { 0x9FB1, 1310 },    /*  */
    { 0x9FB3, 1315 },    /*  */
    { 0xA061, 1319 },    /*  */
    { 0xA062, 1327 },    /*  */
    { 0xA065, 1333 },    /*  */
    { 0xA069, 1352 },    /*  */
    { 0xA077, 1359 },    /*  */
    { 0xA081, 1371 },    /*  */
    { 0xA082, 1385 },    /*  */
    { 0xA097, 1390 },    /*  */
    { 0xA162, 1396 },    /*  */
    { 0xA165, 1398 },    /*  */
    { 0xA169, 1409 },    /*  */
    { 0xA177, 1411 },    /*  */
    { 0xA1A1, 1427 },    /*  */
    { 0xA1A2, 1451 },    /*  */
    { 0xA1A9, 1458 },    /*  */
    { 0xA1B7, 1460 },    /*  */
    { 0xA261, 1463 },    /*  */
    { 0xA281, 1475 },    /*  */
    { 0xA282, 1497 },    /*  */
    { 0xA285, 1499 },    /*  */
    { 0xA289, 1511 },    /*  */
    { 0xA3A1, 1514 },    /*  */
    { 0xA3A5, 1533 },    /*  */
    { 0xA3A9, 1546 },    /*  */
    { 0xA462, 1549 },    /*  */
    { 0xA465, 1568 },    /*  */
    { 0xA469, 1593 },    /*  */
    { 0xA477, 1604 },    /*  */
    { 0xA481, 1632 },    /*  */
    { 0xA482, 1652 },    /*  */
    { 0xA4E5, 1660 },    /*  */
    { 0xA4E9, 1670 },    /*  */
    { 0xA4F1, 1674 },    /*  */
    { 0xA4F3, 1683 },    /*  */
    { 0xA562, 1685 },    /*  */
    { 0xA565, 1696 },    /*  */
    { 0xA569, 1703 },    /*  */
    { 0xA577, 1707 },    /*  */
    { 0xA5A1, 1724 },    /*  */
    { 0xA5A2, 1740 },    /*  */
    { 0xA5A5, 1757 },    /*  */
    { 0xA5A9, 1758 },    /*  ksonly */
    { 0xA5B7, 1759 },    /*  */
    { 0xA681, 1775 },    /*  */
    { 0xA682, 1818 },    /*  */
    { 0xA685, 1819 },    /*  */
    { 0xA689, 1838 },    /*  */
    { 0xA697, 1843 },    /*  */
    { 0xA7A1, 1849 },    /*  */
    { 0xA7A5, 1892 },    /*  */
    { 0xA7B7, 1906 },    /*  */
    { 0xAC61, 1910 },    /*  */
    { 0xAC62, 1970 },    /*  */
    { 0xAC65, 1974 },    /*  */
    { 0xAC69, 1986 },    /*  */
    { 0xAC71, 1991 },    /*  */
    { 0xAC73, 1999 },    /*  */
    { 0xAC77, 2003 },    /*  */
    { 0xAC81, 2034 },    /*  */
    { 0xAC82, 2037 },    /*  */
    { 0xAC97, 2042 },    /*  */
    { 0xACE1, 2047 },    /*  */
    { 0xACE2, 2077 },    /*  */
    { 0xACE5, 2092 },    /*  */
    { 0xACE9, 2124 },    /*  */
    { 0xACF1, 2137 },    /*  */
    { 0xACF3, 2145 },    /*  */
    { 0xACF7, 2149 },    /*  */
    { 0xAD41, 2167 },    /*  */
    { 0xADA1, 2176 },    /*  */
    { 0xADA2, 2213 },    /*  */
    { 0xADA5, 2222 },    /*  */
    { 0xADA9, 2228 },    /*  */
    { 0xADB7, 2229 },    /*  */
    { 0xADE1, 2237 },    /*  */
    { 0xAE41, 2242 },    /*  */
    { 0xAE81, 2244 },    /*  */
    { 0xAE82, 2305 },    /*  */
    { 0xAE85, 2317 },    /*  */
    { 0xAE89, 2344 },    /*  */
    { 0xAE97, 2348 },    /*  */
    { 0xAF69, 2351 },    /*  */
    { 0xAF73, 2354 },    /*  */
    { 0xAF77, 2359 },    /*  */
    { 0xAFA1, 2369 },    /*  */
    { 0xAFA2, 2397 },    /*  */
    { 0xAFA5, 2412 },    /*  */
    { 0xAFA9, 2436 },    /*  */
    { 0xAFB1, 2440 },    /*  */
    { 0xAFB3, 2450 },    /*  */
    { 0xB077, 2453 },    /*  */
    { 0xB3A1, 2454 },    /*  */
    { 0xB461, 2455 },    /*  */
    { 0xB462, 2473 },    /*  */
    { 0xB465, 2487 },    /*  */
    { 0xB469, 2497 },    /*  */
    { 0xB471, 2501 },    /*  */
    { 0xB473, 2509 },    /*  */
    { 0xB477, 2513 },    /*  */
    { 0xB481, 2520 },    /*  */
    { 0xB482, 2531 },    /*  */
    { 0xB497, 2538 },    /*  */
    { 0xB4A1, 2542 },    /*  */
    { 0xB4A2, 2553 },    /*  */
    { 0xB4B7, 2562 },    /*  */
    { 0xB4E1, 2593 },    /*  */
    { 0xB4E2, 2603 },    /*  */
    { 0xB4E5, 2608 },    /*  */
    { 0xB4E9, 2614 },    /*  */
    { 0xB4F1, 2616 },    /*  */
    { 0xB4F3, 2622 },    /*  */
    { 0xB545, 2624 },    /*  ksonly */
    { 0xB561, 2625 },    /*  */
    { 0xB562, 2649 },    /*  */
    { 0xB565, 2662 },    /*  */
    { 0xB569, 2705 },    /*  */
    { 0xB571, 2715 },    /*  */
    { 0xB573, 2730 },    /*  */
    { 0xB577, 2734 },    /*  */
    { 0xB581, 2774 },    /*  */
    { 0xB5A1, 2798 },    /*  */
    { 0xB5A2, 2828 },    /*  */
    { 0xB5A5, 2833 },    /*  */
    { 0xB5A9, 2839 },    /*  */
    { 0xB5B7, 2840 },    /*  */
    { 0xB5C1, 2849 },    /*  */
    { 0xB5C5, 2857 },    /*  */
    { 0xB5C9, 2875 },    /*  ksonly */
    { 0xB5D7, 2876 },    /*  */
    { 0xB5E1, 2881 },    /*  */
    { 0xB641, 2885 },    /*  */
    { 0xB661, 2890 },    /*  */
    { 0xB662, 2928 },    /*  */
    { 0xB677, 2934 },    /*  */
    { 0xB681, 2958 },    /*  */
    { 0xB682, 2990 },    /*  */
    { 0xB685, 2999 },    /*  */
    { 0xB689, 3012 },    /*  */
    { 0xB697, 3015 },    /*  ksonly */
    { 0xB6A5, 3017 },    /*  */
    { 0xB6A9, 3044 },    /*  */
    { 0xB6E1, 3047 },    /*  */
    { 0xB741, 3072 },    /*  */
    { 0xB742, 3128 },    /*  */
    { 0xB745, 3135 },    /*  */
    { 0xB749, 3148 },    /*  */
    { 0xB757, 3153 },    /*  */
    { 0xB765, 3158 },    /*  */
    { 0xB769, 3165 },    /*  */
    { 0xB771, 3166 },    /*  */
    { 0xB773, 3172 },    /*  */
    { 0xB777, 3175 },    /*  */
    { 0xB781, 3179 },    /*  */
    { 0xB7A1, 3198 },    /*  */
    { 0xB7A2, 3236 },    /*  */
    { 0xB7A5, 3244 },    /*  */
    { 0xB7A9, 3268 },    /*  */
    { 0xB7B1, 3277 },    /*  */
    { 0xB7B3, 3288 },    /*  */
    { 0xB7B7, 3293 },    /*  */
    { 0xB861, 3297 },    /*  */
    { 0xB862, 3323 },    /*  */
    { 0xB865, 3336 },    /*  */
    { 0xB871, 3341 },    /*  */
    { 0xB873, 3347 },    /*  */
    { 0xB877, 3348 },    /*  */
    { 0xB881, 3385 },    /*  */
    { 0xB897, 3402 },    /*  */
    { 0xB8E1, 3406 },    /*  */
    { 0xB8E2, 3434 },    /*  */
    { 0xB8E5, 3459 },    /*  */
    { 0xB8E9, 3500 },    /*  */
    { 0xB8F1, 3508 },    /*  */
    { 0xB8F3, 3517 },    /*  */
    { 0xB8F7, 3520 },    /*  */
    { 0xB941, 3575 },    /*  */
    { 0xB9A1, 3598 },    /*  */
    { 0xB9A2, 3644 },    /*  */
    { 0xB9A5, 3648 },    /*  */
    { 0xB9A9, 3650 },    /*  */
    { 0xB9B7, 3653 },    /*  */
    { 0xB9C1, 3670 },    /*  */
    { 0xBA41, 3675 },    /*  */
    { 0xBA81, 3676 },    /*  */
    { 0xBA82, 3716 },    /*  */
    { 0xBA85, 3718 },    /*  */
    { 0xBA89, 3737 },    /*  */
    { 0xBA97, 3738 },    /*  */
    { 0xBB62, 3742 },    /*  */
    { 0xBB69, 3743 },    /*  */
    { 0xBB73, 3744 },    /*  */
    { 0xBB77, 3747 },    /*  */
    { 0xBBA1, 3758 },    /*  */
    { 0xBBA2, 3792 },    /*  */
    { 0xBBA5, 3797 },    /*  */
    { 0xBBA9, 3832 },    /*  */
    { 0xBBB1, 3847 },    /*  */
    { 0xBBB3, 3849 },    /*  */
    { 0xBBB7, 3856 },    /* ¡ */
    { 0xC061, 3859 },    /*  */
    { 0xC062, 3874 },    /*  */
    { 0xC065, 3881 },    /*  */
    { 0xC069, 3896 },    /*  */
    { 0xC071, 3901 },    /*  */
    { 0xC077, 3911 },    /* â */
    { 0xC081, 3933 },    /* ä */
    { 0xC082, 3945 },    /* å */
    { 0xC0E1, 3949 },    /* ó */
    { 0xC0E2, 3953 },    /* ô */
    { 0xC0E5, 3968 },    /* õ */
    { 0xC0E9, 3987 },    /* ö */
    { 0xC0F1, 3997 },    /* ÷ */
    { 0xC0F3, 4007 },    /* ø */
    { 0xC0F7, 4017 },    /* û */
    { 0xC141, 4025 },    /* ü */
    { 0xC1A1, 4035 },    /*  */
    { 0xC1A2, 4062 },    /*  */
    { 0xC1A5, 4068 },    /*  */
    { 0xC1B7, 4072 },    /*  */
    { 0xC1C9, 4082 },    /*  */
    { 0xC241, 4083 },    /*  */
    { 0xC281, 4086 },    /*  */
    { 0xC282, 4109 },    /*  */
    { 0xC285, 4121 },    /*  */
    { 0xC289, 4124 },    /*  */
    { 0xC297, 4127 },    /*  */
    { 0xC2C1, 4133 },    /*  */
    { 0xC2E1, 4137 },    /*  */
    { 0xC362, 4151 },    /*  */
    { 0xC377, 4156 },    /*  */
    { 0xC3A1, 4157 },    /* ġ */
    { 0xC3A2, 4181 },    /* Ģ */
    { 0xC3A5, 4184 },    /* ģ */
    { 0xC3A9, 4185 },    /* ĥ */
    { 0xC3B1, 4188 },    /* ħ */
    { 0xC3B3, 4197 },    /* Ĩ */
    { 0xC3B7, 4198 },    /* Ī */
    { 0xC5E1, 4200 },    /*  */
    { 0xC861, 4201 },    /* Ÿ */
    { 0xC862, 4215 },    /* Ź */
    { 0xC865, 4231 },    /* ź */
    { 0xC869, 4241 },    /* Ż */
    { 0xC871, 4243 },    /* Ž */
    { 0xC873, 4247 },    /* ž */
    { 0xC877, 4250 },    /*  */
    { 0xC881, 4255 },    /*  */
    { 0xC882, 4269 },    /*  */
    { 0xC897, 4272 },    /*  */
    { 0xC8E1, 4273 },    /*  ksonly */
    { 0xC9A1, 4274 },    /*  */
    { 0xC9B7, 4278 },    /*  */
    { 0xCA41, 4285 },    /*  */
    { 0xCA81, 4291 },    /*  */
    { 0xCB62, 4297 },    /* Ư */
    { 0xCB71, 4299 },    /* ƴ ksonly */
    { 0xCC61, 4300 },    /*  */
    { 0xCC65, 4316 },    /*  */
    { 0xCC69, 4325 },    /*  */
    { 0xCC81, 4328 },    /*  */
    { 0xCC97, 4339 },    /*  */
    { 0xCCA2, 4343 },    /*  ksonly */
    { 0xCD65, 4344 },    /*  */
    { 0xCD71, 4354 },    /*  */
    { 0xCD77, 4355 },    /*  */
    { 0xCD81, 4360 },    /*  */
    { 0xCDA1, 4370 },    /*  */
    { 0xCDA2, 4398 },    /*  */
    { 0xCE61, 4404 },    /* ǥ */
    { 0xCE91, 4418 },    /* ǰ */
    { 0xCE97, 4420 },    /* ǳ */
    { 0xCFA1, 4425 },    /*  */
    { 0xCFA9, 4432 },    /*  */
    { 0xCFB3, 4442 },    /*  */
    { 0xD061, 4444 },    /*  */
    { 0xD062, 4458 },    /*  */
    { 0xD065, 4463 },    /*  */
    { 0xD069, 4477 },    /*  */
    { 0xD071, 4479 },    /*  */
    { 0xD073, 4491 },    /*  */
    { 0xD077, 4498 },    /*  */
    { 0xD081, 4515 },    /*  */
    { 0xD082, 4533 },    /*  */
    { 0xD097, 4535 },    /*  */
    { 0xD0B7, 4540 },    /*  */
    { 0xD0E1, 4549 },    /*  */
    { 0xD0E5, 4553 },    /*  */
    { 0xD0E9, 4557 },    /*  */
    { 0xD0F1, 4558 },    /*  */
    { 0xD162, 4560 },    /*  */
    { 0xD165, 4564 },    /*  */
    { 0xD169, 4585 },    /*  */
    { 0xD171, 4589 },    /*  */
    { 0xD173, 4590 },    /*  */
    { 0xD177, 4602 },    /*  */
    { 0xD181, 4622 },    /*  */
    { 0xD1A1, 4631 },    /* ȣ */
    { 0xD1A2, 4672 },    /* Ȥ */
    { 0xD1A5, 4675 },    /* ȥ */
    { 0xD1A9, 4681 },    /* Ȧ */
    { 0xD1B7, 4684 },    /* ȫ */
    { 0xD1C1, 4694 },    /* ȭ */
    { 0xD1C2, 4708 },    /* Ȯ */
    { 0xD1C5, 4714 },    /* ȯ */
    { 0xD1C9, 4731 },    /* Ȱ */
    { 0xD1D7, 4736 },    /* Ȳ */
    { 0xD241, 4760 },    /* ȸ */
    { 0xD242, 4780 },    /* ȹ */
    { 0xD257, 4782 },    /* Ⱦ */
    { 0xD261, 4785 },    /* ȿ */
    { 0xD281, 4798 },    /*  */
    { 0xD285, 4811 },    /*  */
    { 0xD297, 4821 },    /*  ksonly */
    { 0xD2A5, 4822 },    /*  */
    { 0xD2C1, 4826 },    /*  */
    { 0xD2E1, 4829 },    /*  */
    { 0xD341, 4837 },    /*  */
    { 0xD349, 4842 },    /*  */
    { 0xD357, 4845 },    /*  */
    { 0xD362, 4850 },    /*  */
    { 0xD365, 4851 },    /*  */
    { 0xD369, 4855 },    /*  */
    { 0xD371, 4859 },    /*  */
    { 0xD373, 4862 },    /*  */
    { 0xD377, 4866 },    /*  */
    { 0xD381, 4867 },    /*  */
    { 0xD3A9, 4887 },    /*  */
    { 0xFFFF, 4888 }
};

int hjdic_open(char * filename)
{
#if 0
    printf("hjdic_open\n");
#endif
    // return error values
    return 0;
}

void hjdic_close(void)
{
#if 0
    printf("hjdic_close\n");
#endif
}

int hjdic_gethanja(hchar_t *hangul, hchar_t *hanja, int size)
{
    int	ii, count;
    int	l, h, tlen, hlen;
    hchar_t	hh, hj;
    unsigned char	val1, val2;

#if 0
    printf("hjdic_gethanja\n");
#endif
    assert( hangul != NULL && hanja != NULL );
    hlen = hcslen(hangul);
    hh = hangul[0];

    tlen = sizeof(HanjaTbl) / sizeof(HanjaTbl[0]);
    for (l = 0, h = tlen - 1 ; l <= h; ) {
        ii = ( l + h ) / 2;
        if ( HanjaTbl[ii].hh > hh )
            h = ii - 1;
        else if ( HanjaTbl[ii].hh < hh )
            l = ii + 1;
        else
            break;
    }

    if ( HanjaTbl[ii].hh != hh )
        return 0;

    /* ó ѱ ΰ  Ŀ ڸ ä. */
    hanja[0] = hh;
    hj = HanjaTbl[ii].offset;
    count = HanjaTbl[ii + 1].offset - HanjaTbl[ii].offset;
    for ( ii = 0; ii < count; ii++, hj++) {
        val1 = hj / 188 + 0xe0;
        val2 = hj % 188;
        val2 = (val2 < 78) ? (val2 + 0x31) : (val2 + 0x91 - 78);
        hanja[ii + 1] = ((val1 << 8) + val2);
    }

    return count + 1;
}

int hjdic_addterm(hchar_t *hangul, hchar_t *hanja, int code)
{
#if 0
    printf("hjdic_addterm\n");
#endif
    return 0;
}

int hjdic_delterm(hchar_t *hangul, hchar_t *hanja, int code)
{
#if 0
    printf("hjdic_determ\n");
#endif
    return 0;
}

int hjdic_getdict(hchar_t *item)
{
#if 0
    printf("hjdic_getdict\n");
#endif
    return 0;
}

int hjdic_nextdict(hchar_t *item)
{
#if 0
    printf("hjdic_nextdict\n");
#endif
    return 0;
}
