/* Copyright (c) 1998 by Mizi Research Inc.  All Rights Reserved */

/* NAME $Id: navapi.c,v 1.7 2000/01/11 01:28:46 frog Exp $
 * PURPOSE
 *   
 * NOTES
 *   
 * HISTORY
 *        frog - Jul 31, 1998: Created.
 */

#include	<stdio.h>
#ifndef WIN32
# include	<unistd.h>
# include	<sys/wait.h>
#endif    /* !WIN32 */

#include	<X11/Xlib.h>
#include	<X11/Xatom.h>
#include	<X11/Intrinsic.h>
#include	<X11/Xmu/WinUtil.h>    /* for XmuClientWindow() */
#include	"hanim.h"

#define MOZILLA_VERSION_PROP	"_MOZILLA_VERSION"
#define MOZILLA_LOCK_PROP	"_MOZILLA_LOCK"
#define MOZILLA_COMMAND_PROP	"_MOZILLA_COMMAND"
#define MOZILLA_RESPONSE_PROP	"_MOZILLA_RESPONSE"

#ifndef WIN32

static Atom XA_MOZILLA_VERSION	= 0;
static Atom XA_MOZILLA_LOCK	= 0;
static Atom XA_MOZILLA_COMMAND	= 0;
static Atom XA_MOZILLA_RESPONSE	= 0;

static void mozilla_init_atoms(Display *dpy)
{
    if (! XA_MOZILLA_VERSION)
        XA_MOZILLA_VERSION = XInternAtom (dpy, MOZILLA_VERSION_PROP, False);
    if (! XA_MOZILLA_LOCK)
        XA_MOZILLA_LOCK = XInternAtom (dpy, MOZILLA_LOCK_PROP, False);
    if (! XA_MOZILLA_COMMAND)
        XA_MOZILLA_COMMAND = XInternAtom (dpy, MOZILLA_COMMAND_PROP, False);
    if (! XA_MOZILLA_RESPONSE)
        XA_MOZILLA_RESPONSE = XInternAtom (dpy, MOZILLA_RESPONSE_PROP, False);
}

static Window mozilla_find_window(Display *dpy)
{
    Window root = RootWindowOfScreen(DefaultScreenOfDisplay(dpy));
    Window root2, parent, *kids;
    int ii, nkids;

    if ( XA_MOZILLA_VERSION == 0 )
        return 0;
    if ( !XQueryTree(dpy, root, &root2, &parent, &kids, &nkids) ||
         !(kids && nkids) )
        return 0;

    for ( ii = nkids - 1; ii >= 0; --ii ) {
        Atom	type;
        int	format, status;
        unsigned long nitems, bytesafter;
        unsigned char *version = 0;
        Window win;

        win = XmuClientWindow( dpy, kids[ii] );
        status = XGetWindowProperty( dpy, win, XA_MOZILLA_VERSION,
                                     0, (65536 / sizeof (long)),
                                     False, XA_STRING,
                                     &type, &format, &nitems, &bytesafter,
                                     &version );
        if ( status == Success && version ) {
            XFree (version);
            return win;
        }
    }
    return 0;
}

static Boolean mozilla_command(Display *dpy, const char *command)
{
    Window	window;

    if ( (window = mozilla_find_window(dpy)) == 0 )
        return False;
    XChangeProperty(dpy, window, XA_MOZILLA_COMMAND, XA_STRING, 8,
                    PropModeReplace, (unsigned char *) command,
                    strlen(command));
    return True;
#if 0
    while ( 1 ) {
        XEvent event;
        XNextEvent (dpy, &event);
        if (event.xany.type == DestroyNotify &&
            event.xdestroywindow.window == window)
            return False;
        else if ( event.xany.type == PropertyNotify &&
                  event.xproperty.state == PropertyNewValue &&
                  event.xproperty.window == window &&
                  event.xproperty.atom == XA_MOZILLA_RESPONSE ) {
            return True;
        }
    }
#endif
}
#endif    /* WIN32 */

void open_url( const char *url )
{
#ifndef WIN32
    char	mocmd[256];
    pid_t	pid;
    int	status;

    mozilla_init_atoms( XtDisplay(TopLevel) );
    sprintf(mocmd, "OpenURL(%s)", url);
    if ( mozilla_command(XtDisplay(TopLevel), mocmd) )
        return;

    if ( (pid = fork()) == 0 ) {
        execlp("netscape", "netscape", url, NULL);
        execl("/usr/bin/netscape", "netscape", url, NULL);
        execl("/usr/bin/X11/netscape", "netscape", url, NULL);
        exit(3);
    }
    sleep(3);
    waitpid(pid, &status, WNOHANG);
    if ( status && WIFEXITED(status) && (WEXITSTATUS(status) == 3) )
        Message("netscape   ϴ.", 400, 70);
#endif
}
