/* $Id: romandic.c,v 1.16 2000/01/11 01:28:46 frog Exp $ */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"hanim.h"
#include	"hstring.h"
#include	"romandic.h"

/** @name θ   Լ
 *
 * θ Է  ϴ  ƴ,
 * buffer  ־ ϴ ͵ ƴϴ.
 * $ref_count$  ִٰ 0 Ǹ freeϵ Ѵ.
 */

/*@{*/

#define HANTBL	"hanroman.tbl"
#define JPTBL	"japanese.tbl"

typedef struct {
    RomanTable	*first;
    RomanTable	*last;
    int	ref_count;
} RomanList;

#if 0 /* LATER */
static RomanList	han_roman_tbl;
#endif    /* LATER */
static RomanList	jp_roman_tbl;

static RomanList *get_roman_table(ConversionMode);
static Boolean	roman_table_load(RomanList *, char *, LcCtype);

/// ѱ-θ, Ϻ-θ ̺ ʱȭѴ.
void init_roman_tables(void)
{
#if 0 /* LATER */
    han_roman_tbl.first = NULL;
    han_roman_tbl.last = NULL;
    han_roman_tbl.ref_count = 0;
#endif    /* LATER */
    jp_roman_tbl.first = NULL;
    jp_roman_tbl.last = NULL;
    jp_roman_tbl.ref_count = 0;

    /* LATER θ ȯǥ ̳ϰ ϵ  */
#if 0 /* LATER */
    get_roman_table(HAN_ROMAN);
#endif
    get_roman_table(HIRAGANA_ROMAN);
}

static RomanList *get_roman_table(ConversionMode mode)
{
    RomanList *roman_tbl;
    char filename[BUFSIZ], tblname[BUFSIZ];

    if ((LANGMASK & mode) == JAPANESE) {
        roman_tbl = &jp_roman_tbl;
        strcpy(tblname, JPTBL);
    }
#if 0 /* LATER */
    else if ((LANGMASK & mode) == HANGUL) {
        roman_tbl = &han_roman_tbl;
        strcpy(tblname, HANTBL);
    }
#endif    /* LATER */
    else {
        return NULL;
    }

    if (roman_tbl->ref_count != 0) {
        roman_tbl->ref_count++;
        return roman_tbl;
    } else {
	sprintf(filename, "%s/%s/%s", home_directory,
	    HANIMUSRDIR, tblname);
	if (roman_table_load(roman_tbl, filename, user_locale)) {
	    roman_tbl->ref_count = 1;
	    return roman_tbl;
	}
        sprintf(filename, "%s/%s", sys_directory, tblname);
        if (roman_table_load(roman_tbl, filename, LC_CTYPE_EUC)) {
            roman_tbl->ref_count = 1;
            return roman_tbl;
        }
    }
    /* LATER ̾˷α׷  */
    fprintf(stderr, "Warning:cannot open rule file %s\n\r",
	filename);
    return NULL;
}

static Boolean
roman_table_load(RomanList *roman_tbl, char *filename, LcCtype locale )
{
    FILE	*fp;
    unsigned char	line[BUFSIZ];
    unsigned char	*key;
    unsigned char	*ss;
    hchar_t	hangul[32];
    unsigned char	*str;
    RomanTable	*rt, *r;
    int	len, ii, jj, diff;

    if ((fp = fopen(filename, "r")) == NULL)
        return(False);
    while (fgets((char *)line, BUFSIZ, fp) != NULL) {
        if (line[0] == '#')
            continue;
        if (!(key = (unsigned char *)strtok((char *)line, " \t\n")))
            continue;
        if (!(ss = (unsigned char *)strtok(NULL, " \t\n")))
            continue;
        switch (locale) {
        case LC_CTYPE_KSSM:
            len = kssms2hcs(hangul, ss, strlen((char *)ss));
            break;
        case LC_CTYPE_EUC:
            len = kss2hcs(hangul, (char *)ss, strlen((char *)ss));
            break;
        default:
            break;
        }
        str = (unsigned char *)strtok(NULL, " \t\n");

        if (str != NULL) {
            for (ii = 0; *(str + ii) != 0; ii++)
                if (!isalpha(*(str + ii)))
                    goto next;

            jj = strlen((char *)key);
            if (ii > jj || strcmp((char *)&key[jj - ii], (char *)str))
                goto next;
        }

        rt = (RomanTable *)calloc(1, sizeof(RomanTable));
        rt->key = (unsigned char *)XtMalloc(strlen((char *)key) + 1);
        strcpy((char *)rt->key, (char *)key);
        rt->hangul = (hchar_t *)XtMalloc(sizeof(hchar_t) * (len + 1));
        hcscpy(rt->hangul, hangul);
        if (str != NULL) {
            rt->str = (unsigned char *)XtMalloc(strlen((char *)str) + 1);
            strcpy((char *)rt->str, (char *)str);
        }

        for (r = roman_tbl->first; r != NULL; r = r->next) {
            if ((diff = strcmp((char *)key, (char *)r->key)) == 0) {
                XFree(rt->key);
                XFree(rt->hangul);
                if (str != NULL)
                    XFree(rt->str);
                XFree(rt);
                break;
            }
            if (diff < 0) {
                if (r->previous == NULL) {
                    roman_tbl->first = rt;
                    rt->next = r;
                    r->previous = rt;
                } else {
                    r->previous->next = rt;
                    rt->next = r;
                    rt->previous = r->previous;
                    r->previous = rt;
                }
                break;
            }
            if (r->next == NULL) {
                roman_tbl->last = rt;
                r->next = rt;
                rt->previous = r;
                break;
            }
        }
        if (roman_tbl->first == NULL)
            roman_tbl->first = roman_tbl->last = rt;
    next: ;
    }
    return(roman_tbl->first != NULL);
}

/* ѱ θ ȯ Լ */
#if 0 /* LATER */
RomanTable *han_forward_search(char *key)
{
    int	diff;
    RomanTable	*rt;

    for (rt = han_roman_tbl.first; rt != NULL; rt = rt->next) {
        if ((diff = strcmp((char *)key, (char *)rt->key)) == 0)
            return(rt);
        if (diff < 0)
            break;
    }
    return(NULL);
}

RomanTable *han_forward_sub_search(char *key)
{
    int	diff;
    int	length = strlen((char *)key);
    RomanTable	*rt;

    for (rt = han_roman_tbl.first; rt != NULL; rt = rt->next) {
        if ((diff = strncmp((char *)key, (char *)rt->key, length)) == 0)
            return(rt);
        if (diff < 0)
            break;
    }
    return(NULL);
}

RomanTable *han_backward_sub_search(char *key)
{
    RomanTable	*rt;

    for (rt = han_roman_tbl.last; rt != NULL; rt = rt->previous) {
        if (strncmp((char *)key, (char *)rt->key,
	    strlen((char *)rt->key) == 0)
            return(rt);
        if (key[0] > rt->key[0])
            break;
    }
    return(NULL);
}
#endif    /* LATER */

/// Ϻ θ ̺ ڿ key ã´.
RomanTable *jp_forward_search(char *key)
{
    int	diff;
    RomanTable	*rt;

    for (rt = jp_roman_tbl.first; rt != NULL; rt = rt->next) {
        if ((diff = strcmp((char *)key, (char *)rt->key)) == 0)
            return(rt);
        if (diff < 0)
            break;
    }
    return(NULL);
}

/// Ϻ θ ̺ ڿ key ϴ κ ã´.
RomanTable *jp_forward_sub_search(char *key)
{
    int	diff;
    int	length = strlen((char *)key);
    RomanTable	*rt;

    for (rt = jp_roman_tbl.first; rt != NULL; rt = rt->next) {
        if ((diff = strncmp((char *)key, (char *)rt->key, length)) == 0)
            return(rt);
        if (diff < 0)
            break;
    }
    return(NULL);
}

/// Ϻ θ ̺ Ĺ̿ ڿ key ϴ κ ã´.
RomanTable *jp_backward_sub_search(char *key)
{
    RomanTable	*rt;

    for (rt = jp_roman_tbl.last; rt != NULL; rt = rt->previous) {
        if (strncmp((char *)key, (char *)rt->key,
	    strlen((char *)rt->key)) == 0)
            return(rt);
        if (key[0] > rt->key[0])
            break;
    }
    return(NULL);
}

/*@}*/
